// Copyright (c) 2015-2017 Nuxi, https://nuxi.nl/
//
// SPDX-License-Identifier: BSD-2-Clause

#include <locale.h>
#include <stdbool.h>
#include <testing.h>
#include <time.h>

struct testvector {
  time_t t;
  struct tm tm;
  int consistency;
#define UNI 0   // Can only be converted with localtime_l().
#define HINT 1  // Can be converted with mktime_l() and isdst >= 0.
#define BIDI 2  // Can be converted with mktime_l().
};

static void test_locale(const char *locale, const struct testvector *vectors,
                        size_t vectors_count) {
  locale_t locobj = newlocale(LC_TIMEZONE_MASK, locale, 0);
  ASSERT_NE(0, locobj);
  for (size_t i = 0; i < vectors_count; ++i) {
    const struct testvector *vector = &vectors[i];
    SCOPED_NOTE(vector->t, {
      // Test localtime_l().
      struct timespec ts = {.tv_sec = vector->t};
      struct tm tm;
      ASSERT_EQ(0, localtime_l(&ts, &tm, locobj));
      ASSERT_EQ(vector->tm.tm_gmtoff, tm.tm_gmtoff);
      ASSERT_EQ(vector->tm.tm_year, tm.tm_year);
      ASSERT_EQ(vector->tm.tm_mon, tm.tm_mon);
      ASSERT_EQ(vector->tm.tm_wday, tm.tm_wday);
      ASSERT_EQ(vector->tm.tm_mday, tm.tm_mday);
      ASSERT_EQ(vector->tm.tm_hour, tm.tm_hour);
      ASSERT_EQ(vector->tm.tm_min, tm.tm_min);
      ASSERT_EQ(vector->tm.tm_sec, tm.tm_sec);
      ASSERT_EQ(vector->tm.tm_isdst, tm.tm_isdst);
      ASSERT_STREQ(vector->tm.tm_zone, tm.tm_zone);

      // Test mktime_l() with tm_isdst >= 0.
      ASSERT_EQ(0, mktime_l(&tm, &ts, locobj));
      if (vector->consistency >= HINT) {
        ASSERT_EQ(vector->t, ts.tv_sec);
      } else {
        ASSERT_NE(vector->t, ts.tv_sec);
      }
      ASSERT_EQ(0, ts.tv_nsec);

      // Test mktime_l() with tm_isdst < 0.
      tm.tm_isdst = -1;
      ASSERT_EQ(0, mktime_l(&tm, &ts, locobj));
      if (vector->consistency >= BIDI) {
        ASSERT_EQ(vector->t, ts.tv_sec);
      } else {
        ASSERT_NE(vector->t, ts.tv_sec);
      }
      ASSERT_EQ(0, ts.tv_nsec);
    });
  }
  freelocale(locobj);
}

#define TEST_BEGIN(name, tzname)                 \
  TEST(localtime_l, name) {                      \
    static const char localename[] = "@" tzname; \
    struct testvector testvectors[] = {
#define TESTVECTOR(t, year, mon, wday, mday, hour, min, sec, isdst, gmtoff, \
                   zone, consistency)                                       \
  {                                                                         \
      t,                                                                    \
      {                                                                     \
          .tm_year = year - 1900,                                           \
          .tm_mon = mon - 1,                                                \
          .tm_wday = wday,                                                  \
          .tm_mday = mday,                                                  \
          .tm_hour = hour,                                                  \
          .tm_min = min,                                                    \
          .tm_sec = sec,                                                    \
          .tm_isdst = isdst,                                                \
          .tm_gmtoff = gmtoff,                                              \
          .tm_zone = zone,                                                  \
      },                                                                    \
      consistency,                                                          \
  },
#define TEST_END                                                   \
  }                                                                \
  ;                                                                \
  test_locale(localename, testvectors, __arraycount(testvectors)); \
  }

TEST_BEGIN(accra, "Africa/Accra")
TESTVECTOR(-1640995149, 1917, 12, 1, 31, 23, 59, 59, 0, -52, "LMT", BIDI)
TESTVECTOR(-1640995148, 1918, 1, 2, 1, 0, 0, 52, 0, 0, "GMT", BIDI)
TESTVECTOR(-1556841601, 1920, 8, 2, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1556841600, 1920, 9, 3, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1546388401, 1920, 12, 4, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1546388400, 1920, 12, 4, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1546387201, 1920, 12, 4, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1546387200, 1920, 12, 5, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1525305601, 1921, 8, 3, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1525305600, 1921, 9, 4, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1514852401, 1921, 12, 5, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1514852400, 1921, 12, 5, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1514851201, 1921, 12, 5, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1514851200, 1921, 12, 6, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1493769601, 1922, 8, 4, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1493769600, 1922, 9, 5, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1483316401, 1922, 12, 6, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1483316400, 1922, 12, 6, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1483315201, 1922, 12, 6, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1483315200, 1922, 12, 0, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1462233601, 1923, 8, 5, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1462233600, 1923, 9, 6, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1451780401, 1923, 12, 0, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1451780400, 1923, 12, 0, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1451779201, 1923, 12, 0, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1451779200, 1923, 12, 1, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1430611201, 1924, 8, 0, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1430611200, 1924, 9, 1, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1420158001, 1924, 12, 2, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1420158000, 1924, 12, 2, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1420156801, 1924, 12, 2, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1420156800, 1924, 12, 3, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1399075201, 1925, 8, 1, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1399075200, 1925, 9, 2, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1388622001, 1925, 12, 3, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1388622000, 1925, 12, 3, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1388620801, 1925, 12, 3, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1388620800, 1925, 12, 4, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1367539201, 1926, 8, 2, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1367539200, 1926, 9, 3, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1357086001, 1926, 12, 4, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1357086000, 1926, 12, 4, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1357084801, 1926, 12, 4, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1357084800, 1926, 12, 5, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1336003201, 1927, 8, 3, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1336003200, 1927, 9, 4, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1325550001, 1927, 12, 5, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1325550000, 1927, 12, 5, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1325548801, 1927, 12, 5, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1325548800, 1927, 12, 6, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1304380801, 1928, 8, 5, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1304380800, 1928, 9, 6, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1293927601, 1928, 12, 0, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1293927600, 1928, 12, 0, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1293926401, 1928, 12, 0, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1293926400, 1928, 12, 1, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1272844801, 1929, 8, 6, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1272844800, 1929, 9, 0, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1262391601, 1929, 12, 1, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1262391600, 1929, 12, 1, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1262390401, 1929, 12, 1, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1262390400, 1929, 12, 2, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1241308801, 1930, 8, 0, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1241308800, 1930, 9, 1, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1230855601, 1930, 12, 2, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1230855600, 1930, 12, 2, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1230854401, 1930, 12, 2, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1230854400, 1930, 12, 3, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1209772801, 1931, 8, 1, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1209772800, 1931, 9, 2, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1199319601, 1931, 12, 3, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1199319600, 1931, 12, 3, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1199318401, 1931, 12, 3, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1199318400, 1931, 12, 4, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1178150401, 1932, 8, 3, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1178150400, 1932, 9, 4, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1167697201, 1932, 12, 5, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1167697200, 1932, 12, 5, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1167696001, 1932, 12, 5, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1167696000, 1932, 12, 6, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1146614401, 1933, 8, 4, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1146614400, 1933, 9, 5, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1136161201, 1933, 12, 6, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1136161200, 1933, 12, 6, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1136160001, 1933, 12, 6, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1136160000, 1933, 12, 0, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1115078401, 1934, 8, 5, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1115078400, 1934, 9, 6, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1104625201, 1934, 12, 0, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1104625200, 1934, 12, 0, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1104624001, 1934, 12, 0, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1104624000, 1934, 12, 1, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1083542401, 1935, 8, 6, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1083542400, 1935, 9, 0, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1073089201, 1935, 12, 1, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1073089200, 1935, 12, 1, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1073088001, 1935, 12, 1, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1073088000, 1935, 12, 2, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1051920001, 1936, 8, 1, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1051920000, 1936, 9, 2, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1041466801, 1936, 12, 3, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1041466800, 1936, 12, 3, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1041465601, 1936, 12, 3, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1041465600, 1936, 12, 4, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1020384001, 1937, 8, 2, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1020384000, 1937, 9, 3, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1009930801, 1937, 12, 4, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-1009930800, 1937, 12, 4, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1009929601, 1937, 12, 4, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1009929600, 1937, 12, 5, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-988848001, 1938, 8, 3, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-988848000, 1938, 9, 4, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-978394801, 1938, 12, 5, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-978394800, 1938, 12, 5, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-978393601, 1938, 12, 5, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-978393600, 1938, 12, 6, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-957312001, 1939, 8, 4, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-957312000, 1939, 9, 5, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-946858801, 1939, 12, 6, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-946858800, 1939, 12, 6, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-946857601, 1939, 12, 6, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-946857600, 1939, 12, 0, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-925689601, 1940, 8, 6, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-925689600, 1940, 9, 0, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-915236401, 1940, 12, 1, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-915236400, 1940, 12, 1, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-915235201, 1940, 12, 1, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-915235200, 1940, 12, 2, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-894153601, 1941, 8, 0, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-894153600, 1941, 9, 1, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-883700401, 1941, 12, 2, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-883700400, 1941, 12, 2, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-883699201, 1941, 12, 2, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-883699200, 1941, 12, 3, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-862617601, 1942, 8, 1, 31, 23, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-862617600, 1942, 9, 2, 1, 0, 20, 0, 1, 1200, "+0020", BIDI)
TESTVECTOR(-852164401, 1942, 12, 3, 30, 23, 59, 59, 1, 1200, "+0020", BIDI)
TESTVECTOR(-852164400, 1942, 12, 3, 30, 23, 40, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-852163201, 1942, 12, 3, 30, 23, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-852163200, 1942, 12, 4, 31, 0, 0, 0, 0, 0, "GMT", BIDI)
TEST_END

TEST_BEGIN(amsterdam, "Europe/Amsterdam")
TESTVECTOR(-4260212373, 1834, 12, 3, 31, 23, 59, 59, 0, 1172, "LMT", BIDI)
TESTVECTOR(-4260212372, 1835, 1, 4, 1, 0, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1693700373, 1916, 4, 0, 30, 23, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1693700372, 1916, 5, 1, 1, 1, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1680484773, 1916, 9, 6, 30, 23, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1680484772, 1916, 9, 6, 30, 23, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1680481173, 1916, 9, 6, 30, 23, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1680481172, 1916, 10, 0, 1, 0, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1663453173, 1917, 4, 1, 16, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1663453172, 1917, 4, 1, 16, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1650147573, 1917, 9, 1, 17, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1650147572, 1917, 9, 1, 17, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1650143973, 1917, 9, 1, 17, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1650143972, 1917, 9, 1, 17, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1633213173, 1918, 4, 1, 1, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1633213172, 1918, 4, 1, 1, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1617488373, 1918, 9, 1, 30, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1617488372, 1918, 9, 1, 30, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1617484773, 1918, 9, 1, 30, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1617484772, 1918, 9, 1, 30, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1601158773, 1919, 4, 1, 7, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1601158772, 1919, 4, 1, 7, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1586038773, 1919, 9, 1, 29, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1586038772, 1919, 9, 1, 29, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1586035173, 1919, 9, 1, 29, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1586035172, 1919, 9, 1, 29, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1569709173, 1920, 4, 1, 5, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1569709172, 1920, 4, 1, 5, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1554589173, 1920, 9, 1, 27, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1554589172, 1920, 9, 1, 27, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1554585573, 1920, 9, 1, 27, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1554585572, 1920, 9, 1, 27, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1538259573, 1921, 4, 1, 4, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1538259572, 1921, 4, 1, 4, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1523139573, 1921, 9, 1, 26, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1523139572, 1921, 9, 1, 26, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1523135973, 1921, 9, 1, 26, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1523135972, 1921, 9, 1, 26, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1507501173, 1922, 3, 0, 26, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1507501172, 1922, 3, 0, 26, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1490566773, 1922, 10, 0, 8, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1490566772, 1922, 10, 0, 8, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1490563173, 1922, 10, 0, 8, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1490563172, 1922, 10, 0, 8, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1470176373, 1923, 6, 5, 1, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1470176372, 1923, 6, 5, 1, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1459117173, 1923, 10, 0, 7, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1459117172, 1923, 10, 0, 7, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1459113573, 1923, 10, 0, 7, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1459113572, 1923, 10, 0, 7, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1443997173, 1924, 3, 0, 30, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1443997172, 1924, 3, 0, 30, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1427667573, 1924, 10, 0, 5, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1427667572, 1924, 10, 0, 5, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1427663973, 1924, 10, 0, 5, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1427663972, 1924, 10, 0, 5, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1406672373, 1925, 6, 5, 5, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1406672372, 1925, 6, 5, 5, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1396217973, 1925, 10, 0, 4, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1396217972, 1925, 10, 0, 4, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1396214373, 1925, 10, 0, 4, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1396214372, 1925, 10, 0, 4, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1376950773, 1926, 5, 6, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1376950772, 1926, 5, 6, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1364768373, 1926, 10, 0, 3, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1364768372, 1926, 10, 0, 3, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1364764773, 1926, 10, 0, 3, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1364764772, 1926, 10, 0, 3, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1345414773, 1927, 5, 0, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1345414772, 1927, 5, 0, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1333318773, 1927, 10, 0, 2, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1333318772, 1927, 10, 0, 2, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1333315173, 1927, 10, 0, 2, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1333315172, 1927, 10, 0, 2, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1313792373, 1928, 5, 2, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1313792372, 1928, 5, 2, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1301264373, 1928, 10, 0, 7, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1301264372, 1928, 10, 0, 7, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1301260773, 1928, 10, 0, 7, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1301260772, 1928, 10, 0, 7, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1282256373, 1929, 5, 3, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1282256372, 1929, 5, 3, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1269814773, 1929, 10, 0, 6, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1269814772, 1929, 10, 0, 6, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1269811173, 1929, 10, 0, 6, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1269811172, 1929, 10, 0, 6, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1250720373, 1930, 5, 4, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1250720372, 1930, 5, 4, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1238365173, 1930, 10, 0, 5, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1238365172, 1930, 10, 0, 5, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1238361573, 1930, 10, 0, 5, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1238361572, 1930, 10, 0, 5, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1219184373, 1931, 5, 5, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1219184372, 1931, 5, 5, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1206915573, 1931, 10, 0, 4, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1206915572, 1931, 10, 0, 4, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1206911973, 1931, 10, 0, 4, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1206911972, 1931, 10, 0, 4, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1186957173, 1932, 5, 0, 22, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1186957172, 1932, 5, 0, 22, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1175465973, 1932, 10, 0, 2, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1175465972, 1932, 10, 0, 2, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1175462373, 1932, 10, 0, 2, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1175462372, 1932, 10, 0, 2, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1156025973, 1933, 5, 1, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1156025972, 1933, 5, 1, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1143411573, 1933, 10, 0, 8, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1143411572, 1933, 10, 0, 8, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1143407973, 1933, 10, 0, 8, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1143407972, 1933, 10, 0, 8, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1124489973, 1934, 5, 2, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1124489972, 1934, 5, 2, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1111961973, 1934, 10, 0, 7, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1111961972, 1934, 10, 0, 7, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1111958373, 1934, 10, 0, 7, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1111958372, 1934, 10, 0, 7, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1092953973, 1935, 5, 3, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1092953972, 1935, 5, 3, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1080512373, 1935, 10, 0, 6, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1080512372, 1935, 10, 0, 6, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1080508773, 1935, 10, 0, 6, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1080508772, 1935, 10, 0, 6, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1061331573, 1936, 5, 5, 15, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1061331572, 1936, 5, 5, 15, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1049062773, 1936, 10, 0, 4, 2, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1049062772, 1936, 10, 0, 4, 2, 0, 0, 0, 1172, "AMT", HINT)
TESTVECTOR(-1049059173, 1936, 10, 0, 4, 2, 59, 59, 0, 1172, "AMT", HINT)
TESTVECTOR(-1049059172, 1936, 10, 0, 4, 3, 0, 0, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1029190773, 1937, 5, 6, 22, 1, 59, 59, 0, 1172, "AMT", BIDI)
TESTVECTOR(-1029190772, 1937, 5, 6, 22, 3, 0, 0, 1, 4772, "NST", BIDI)
TESTVECTOR(-1025745573, 1937, 6, 3, 30, 23, 59, 59, 1, 4772, "NST", BIDI)
TESTVECTOR(-1025745572, 1937, 7, 4, 1, 0, 0, 28, 1, 4800, "+0120", BIDI)
TESTVECTOR(-1017613201, 1937, 10, 0, 3, 2, 59, 59, 1, 4800, "+0120", BIDI)
TESTVECTOR(-1017613200, 1937, 10, 0, 3, 2, 0, 0, 0, 1200, "+0020", HINT)
TESTVECTOR(-1017609601, 1937, 10, 0, 3, 2, 59, 59, 0, 1200, "+0020", HINT)
TESTVECTOR(-1017609600, 1937, 10, 0, 3, 3, 0, 0, 0, 1200, "+0020", BIDI)
TESTVECTOR(-998259601, 1938, 5, 0, 15, 1, 59, 59, 0, 1200, "+0020", BIDI)
TESTVECTOR(-998259600, 1938, 5, 0, 15, 3, 0, 0, 1, 4800, "+0120", BIDI)
TESTVECTOR(-986163601, 1938, 10, 0, 2, 2, 59, 59, 1, 4800, "+0120", BIDI)
TESTVECTOR(-986163600, 1938, 10, 0, 2, 2, 0, 0, 0, 1200, "+0020", HINT)
TESTVECTOR(-986160001, 1938, 10, 0, 2, 2, 59, 59, 0, 1200, "+0020", HINT)
TESTVECTOR(-986160000, 1938, 10, 0, 2, 3, 0, 0, 0, 1200, "+0020", BIDI)
TESTVECTOR(-966723601, 1939, 5, 1, 15, 1, 59, 59, 0, 1200, "+0020", BIDI)
TESTVECTOR(-966723600, 1939, 5, 1, 15, 3, 0, 0, 1, 4800, "+0120", BIDI)
TESTVECTOR(-954109201, 1939, 10, 0, 8, 2, 59, 59, 1, 4800, "+0120", BIDI)
TESTVECTOR(-954109200, 1939, 10, 0, 8, 2, 0, 0, 0, 1200, "+0020", HINT)
TESTVECTOR(-954105601, 1939, 10, 0, 8, 2, 59, 59, 0, 1200, "+0020", HINT)
TESTVECTOR(-954105600, 1939, 10, 0, 8, 3, 0, 0, 0, 1200, "+0020", BIDI)
TESTVECTOR(-935022001, 1940, 5, 3, 15, 23, 59, 59, 0, 1200, "+0020", BIDI)
TESTVECTOR(-935022000, 1940, 5, 4, 16, 1, 40, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-857257201, 1942, 11, 1, 2, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-857257200, 1942, 11, 1, 2, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-857253601, 1942, 11, 1, 2, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-857253600, 1942, 11, 1, 2, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-844556401, 1943, 3, 1, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-844556400, 1943, 3, 1, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-828226801, 1943, 10, 1, 4, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-828226800, 1943, 10, 1, 4, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-828223201, 1943, 10, 1, 4, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-828223200, 1943, 10, 1, 4, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-812502001, 1944, 4, 1, 3, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-812502000, 1944, 4, 1, 3, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-796777201, 1944, 10, 1, 2, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-796777200, 1944, 10, 1, 2, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-796773601, 1944, 10, 1, 2, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-796773600, 1944, 10, 1, 2, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-781052401, 1945, 4, 1, 2, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-781052400, 1945, 4, 1, 2, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-766623601, 1945, 9, 0, 16, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-766623600, 1945, 9, 0, 16, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-766620001, 1945, 9, 0, 16, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-766620000, 1945, 9, 0, 16, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(228877199, 1977, 4, 0, 3, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(228877200, 1977, 4, 0, 3, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(243997199, 1977, 9, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(243997200, 1977, 9, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(244000799, 1977, 9, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(244000800, 1977, 9, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(260326799, 1978, 4, 0, 2, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(260326800, 1978, 4, 0, 2, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(276051599, 1978, 10, 0, 1, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(276051600, 1978, 10, 0, 1, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(276055199, 1978, 10, 0, 1, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(276055200, 1978, 10, 0, 1, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(291776399, 1979, 4, 0, 1, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(291776400, 1979, 4, 0, 1, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(307501199, 1979, 9, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(307501200, 1979, 9, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(307504799, 1979, 9, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(307504800, 1979, 9, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(323830799, 1980, 4, 0, 6, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(323830800, 1980, 4, 0, 6, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(338950799, 1980, 9, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(338950800, 1980, 9, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(338954399, 1980, 9, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(338954400, 1980, 9, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(354675599, 1981, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(354675600, 1981, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(370400399, 1981, 9, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(370400400, 1981, 9, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(370403999, 1981, 9, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(370404000, 1981, 9, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(386125199, 1982, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(386125200, 1982, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(401849999, 1982, 9, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(401850000, 1982, 9, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(401853599, 1982, 9, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(401853600, 1982, 9, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(417574799, 1983, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(417574800, 1983, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(433299599, 1983, 9, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(433299600, 1983, 9, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(433303199, 1983, 9, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(433303200, 1983, 9, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(449024399, 1984, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(449024400, 1984, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(465353999, 1984, 9, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(465354000, 1984, 9, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(465357599, 1984, 9, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(465357600, 1984, 9, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(481078799, 1985, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(481078800, 1985, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(496803599, 1985, 9, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(496803600, 1985, 9, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(496807199, 1985, 9, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(496807200, 1985, 9, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(512528399, 1986, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(512528400, 1986, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(528253199, 1986, 9, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(528253200, 1986, 9, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(528256799, 1986, 9, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(528256800, 1986, 9, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(543977999, 1987, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(543978000, 1987, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(559702799, 1987, 9, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(559702800, 1987, 9, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(559706399, 1987, 9, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(559706400, 1987, 9, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(575427599, 1988, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(575427600, 1988, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(591152399, 1988, 9, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(591152400, 1988, 9, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(591155999, 1988, 9, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(591156000, 1988, 9, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(606877199, 1989, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(606877200, 1989, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(622601999, 1989, 9, 0, 24, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(622602000, 1989, 9, 0, 24, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(622605599, 1989, 9, 0, 24, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(622605600, 1989, 9, 0, 24, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(638326799, 1990, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(638326800, 1990, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(654656399, 1990, 9, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(654656400, 1990, 9, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(654659999, 1990, 9, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(654660000, 1990, 9, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(670381199, 1991, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(670381200, 1991, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(686105999, 1991, 9, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(686106000, 1991, 9, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(686109599, 1991, 9, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(686109600, 1991, 9, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(701830799, 1992, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(701830800, 1992, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(717555599, 1992, 9, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(717555600, 1992, 9, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(717559199, 1992, 9, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(717559200, 1992, 9, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(733280399, 1993, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(733280400, 1993, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(749005199, 1993, 9, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(749005200, 1993, 9, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(749008799, 1993, 9, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(749008800, 1993, 9, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(764729999, 1994, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(764730000, 1994, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(780454799, 1994, 9, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(780454800, 1994, 9, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(780458399, 1994, 9, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(780458400, 1994, 9, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(796179599, 1995, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(796179600, 1995, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(811904399, 1995, 9, 0, 24, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(811904400, 1995, 9, 0, 24, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(811907999, 1995, 9, 0, 24, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(811908000, 1995, 9, 0, 24, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(828233999, 1996, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(828234000, 1996, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(846377999, 1996, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(846378000, 1996, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(846381599, 1996, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(846381600, 1996, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(859683599, 1997, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(859683600, 1997, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(877827599, 1997, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(877827600, 1997, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(877831199, 1997, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(877831200, 1997, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(891133199, 1998, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(891133200, 1998, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(909277199, 1998, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(909277200, 1998, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(909280799, 1998, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(909280800, 1998, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(922582799, 1999, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(922582800, 1999, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(941331599, 1999, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(941331600, 1999, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(941335199, 1999, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(941335200, 1999, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(954032399, 2000, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(954032400, 2000, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(972781199, 2000, 10, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(972781200, 2000, 10, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(972784799, 2000, 10, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(972784800, 2000, 10, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(985481999, 2001, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(985482000, 2001, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1004230799, 2001, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1004230800, 2001, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1004234399, 2001, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1004234400, 2001, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1017536399, 2002, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1017536400, 2002, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1035680399, 2002, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1035680400, 2002, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1035683999, 2002, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1035684000, 2002, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1048985999, 2003, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1048986000, 2003, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1067129999, 2003, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1067130000, 2003, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1067133599, 2003, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1067133600, 2003, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1080435599, 2004, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1080435600, 2004, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1099184399, 2004, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1099184400, 2004, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1099187999, 2004, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1099188000, 2004, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1111885199, 2005, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1111885200, 2005, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1130633999, 2005, 10, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1130634000, 2005, 10, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1130637599, 2005, 10, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1130637600, 2005, 10, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1143334799, 2006, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1143334800, 2006, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1162083599, 2006, 10, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1162083600, 2006, 10, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1162087199, 2006, 10, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1162087200, 2006, 10, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1174784399, 2007, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1174784400, 2007, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1193533199, 2007, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1193533200, 2007, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1193536799, 2007, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1193536800, 2007, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1206838799, 2008, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1206838800, 2008, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1224982799, 2008, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1224982800, 2008, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1224986399, 2008, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1224986400, 2008, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1238288399, 2009, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1238288400, 2009, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1256432399, 2009, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1256432400, 2009, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1256435999, 2009, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1256436000, 2009, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1269737999, 2010, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1269738000, 2010, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1288486799, 2010, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1288486800, 2010, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1288490399, 2010, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1288490400, 2010, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1301187599, 2011, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1301187600, 2011, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1319936399, 2011, 10, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1319936400, 2011, 10, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1319939999, 2011, 10, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1319940000, 2011, 10, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1332637199, 2012, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1332637200, 2012, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1351385999, 2012, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1351386000, 2012, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1351389599, 2012, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1351389600, 2012, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1364691599, 2013, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1364691600, 2013, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1382835599, 2013, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1382835600, 2013, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1382839199, 2013, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1382839200, 2013, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1396141199, 2014, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1396141200, 2014, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1414285199, 2014, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1414285200, 2014, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1414288799, 2014, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1414288800, 2014, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1427590799, 2015, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1427590800, 2015, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1445734799, 2015, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1445734800, 2015, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1445738399, 2015, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1445738400, 2015, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1459040399, 2016, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1459040400, 2016, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1477789199, 2016, 10, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1477789200, 2016, 10, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1477792799, 2016, 10, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1477792800, 2016, 10, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1490489999, 2017, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1490490000, 2017, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1509238799, 2017, 10, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1509238800, 2017, 10, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1509242399, 2017, 10, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1509242400, 2017, 10, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1521939599, 2018, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1521939600, 2018, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1540688399, 2018, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1540688400, 2018, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1540691999, 2018, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1540692000, 2018, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1553993999, 2019, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1553994000, 2019, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1572137999, 2019, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1572138000, 2019, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1572141599, 2019, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1572141600, 2019, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1585443599, 2020, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1585443600, 2020, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1603587599, 2020, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1603587600, 2020, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1603591199, 2020, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1603591200, 2020, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1616893199, 2021, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1616893200, 2021, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1635641999, 2021, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1635642000, 2021, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1635645599, 2021, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1635645600, 2021, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1648342799, 2022, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1648342800, 2022, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1667091599, 2022, 10, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1667091600, 2022, 10, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1667095199, 2022, 10, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1667095200, 2022, 10, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1679792399, 2023, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1679792400, 2023, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1698541199, 2023, 10, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1698541200, 2023, 10, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1698544799, 2023, 10, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1698544800, 2023, 10, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1711846799, 2024, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1711846800, 2024, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1729990799, 2024, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1729990800, 2024, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1729994399, 2024, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1729994400, 2024, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1743296399, 2025, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1743296400, 2025, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1761440399, 2025, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1761440400, 2025, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1761443999, 2025, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1761444000, 2025, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1774745999, 2026, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1774746000, 2026, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1792889999, 2026, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1792890000, 2026, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1792893599, 2026, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1792893600, 2026, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1806195599, 2027, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1806195600, 2027, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1824944399, 2027, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1824944400, 2027, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1824947999, 2027, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1824948000, 2027, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1837645199, 2028, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1837645200, 2028, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1856393999, 2028, 10, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1856394000, 2028, 10, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1856397599, 2028, 10, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1856397600, 2028, 10, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1869094799, 2029, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1869094800, 2029, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1887843599, 2029, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1887843600, 2029, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1887847199, 2029, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1887847200, 2029, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1901149199, 2030, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1901149200, 2030, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1919293199, 2030, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1919293200, 2030, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1919296799, 2030, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1919296800, 2030, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1932598799, 2031, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1932598800, 2031, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1950742799, 2031, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1950742800, 2031, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1950746399, 2031, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1950746400, 2031, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1964048399, 2032, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1964048400, 2032, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1982797199, 2032, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1982797200, 2032, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(1982800799, 2032, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(1982800800, 2032, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1995497999, 2033, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1995498000, 2033, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2014246799, 2033, 10, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2014246800, 2033, 10, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2014250399, 2033, 10, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2014250400, 2033, 10, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2026947599, 2034, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2026947600, 2034, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2045696399, 2034, 10, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2045696400, 2034, 10, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2045699999, 2034, 10, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2045700000, 2034, 10, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2058397199, 2035, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2058397200, 2035, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2077145999, 2035, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2077146000, 2035, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2077149599, 2035, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2077149600, 2035, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2090451599, 2036, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2090451600, 2036, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2108595599, 2036, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2108595600, 2036, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2108599199, 2036, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2108599200, 2036, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2121901199, 2037, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2121901200, 2037, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2140045199, 2037, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2140045200, 2037, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2140048799, 2037, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2140048800, 2037, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2153350799, 2038, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2153350800, 2038, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2172099599, 2038, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2172099600, 2038, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2172103199, 2038, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2172103200, 2038, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2184800399, 2039, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2184800400, 2039, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2203549199, 2039, 10, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2203549200, 2039, 10, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2203552799, 2039, 10, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2203552800, 2039, 10, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2216249999, 2040, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2216250000, 2040, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2234998799, 2040, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2234998800, 2040, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2235002399, 2040, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2235002400, 2040, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2248304399, 2041, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2248304400, 2041, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2266448399, 2041, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2266448400, 2041, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2266451999, 2041, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2266452000, 2041, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2279753999, 2042, 3, 0, 30, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2279754000, 2042, 3, 0, 30, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2297897999, 2042, 10, 0, 26, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2297898000, 2042, 10, 0, 26, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2297901599, 2042, 10, 0, 26, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2297901600, 2042, 10, 0, 26, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2311203599, 2043, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2311203600, 2043, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2329347599, 2043, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2329347600, 2043, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2329351199, 2043, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2329351200, 2043, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2342653199, 2044, 3, 0, 27, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2342653200, 2044, 3, 0, 27, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2361401999, 2044, 10, 0, 30, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2361402000, 2044, 10, 0, 30, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2361405599, 2044, 10, 0, 30, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2361405600, 2044, 10, 0, 30, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2374102799, 2045, 3, 0, 26, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2374102800, 2045, 3, 0, 26, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2392851599, 2045, 10, 0, 29, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2392851600, 2045, 10, 0, 29, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2392855199, 2045, 10, 0, 29, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2392855200, 2045, 10, 0, 29, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2405552399, 2046, 3, 0, 25, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2405552400, 2046, 3, 0, 25, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2424301199, 2046, 10, 0, 28, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2424301200, 2046, 10, 0, 28, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2424304799, 2046, 10, 0, 28, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2424304800, 2046, 10, 0, 28, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2437606799, 2047, 3, 0, 31, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2437606800, 2047, 3, 0, 31, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2455750799, 2047, 10, 0, 27, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2455750800, 2047, 10, 0, 27, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2455754399, 2047, 10, 0, 27, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2455754400, 2047, 10, 0, 27, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2469056399, 2048, 3, 0, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2469056400, 2048, 3, 0, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2487200399, 2048, 10, 0, 25, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2487200400, 2048, 10, 0, 25, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2487203999, 2048, 10, 0, 25, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2487204000, 2048, 10, 0, 25, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(2500505999, 2049, 3, 0, 28, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(2500506000, 2049, 3, 0, 28, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(2519254799, 2049, 10, 0, 31, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(2519254800, 2049, 10, 0, 31, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(2519258399, 2049, 10, 0, 31, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(2519258400, 2049, 10, 0, 31, 3, 0, 0, 0, 3600, "CET", BIDI)
TEST_END

TEST_BEGIN(apia, "Pacific/Apia")
TESTVECTOR(-2445424385, 1892, 7, 1, 4, 23, 59, 59, 0, 45184, "LMT", BIDI)
TESTVECTOR(-2445424384, 1892, 7, 1, 4, 0, 0, 0, 0, -41216, "LMT", UNI)
TESTVECTOR(-2445337985, 1892, 7, 1, 4, 23, 59, 59, 0, -41216, "LMT", UNI)
TESTVECTOR(-2445337984, 1892, 7, 2, 5, 0, 0, 0, 0, -41216, "LMT", BIDI)
TESTVECTOR(-1861878785, 1910, 12, 6, 31, 23, 59, 59, 0, -41216, "LMT", BIDI)
TESTVECTOR(-1861878784, 1910, 12, 6, 31, 23, 56, 56, 0, -41400, "-1130", UNI)
TESTVECTOR(-1861878601, 1910, 12, 6, 31, 23, 59, 59, 0, -41400, "-1130", UNI)
TESTVECTOR(-1861878600, 1911, 1, 0, 1, 0, 0, 0, 0, -41400, "-1130", BIDI)
TESTVECTOR(-631110601, 1949, 12, 6, 31, 23, 59, 59, 0, -41400, "-1130", BIDI)
TESTVECTOR(-631110600, 1950, 1, 0, 1, 0, 30, 0, 0, -39600, "-11", BIDI)
TESTVECTOR(1285498799, 2010, 9, 6, 25, 23, 59, 59, 0, -39600, "-11", BIDI)
TESTVECTOR(1285498800, 2010, 9, 0, 26, 1, 0, 0, 1, -36000, "-10", BIDI)
TESTVECTOR(1301752799, 2011, 4, 6, 2, 3, 59, 59, 1, -36000, "-10", BIDI)
TESTVECTOR(1301752800, 2011, 4, 6, 2, 3, 0, 0, 0, -39600, "-11", HINT)
TESTVECTOR(1301756399, 2011, 4, 6, 2, 3, 59, 59, 0, -39600, "-11", HINT)
TESTVECTOR(1301756400, 2011, 4, 6, 2, 4, 0, 0, 0, -39600, "-11", BIDI)
TESTVECTOR(1316872799, 2011, 9, 6, 24, 2, 59, 59, 0, -39600, "-11", BIDI)
TESTVECTOR(1316872800, 2011, 9, 6, 24, 4, 0, 0, 1, -36000, "-10", BIDI)
TESTVECTOR(1325239199, 2011, 12, 4, 29, 23, 59, 59, 1, -36000, "-10", BIDI)
TESTVECTOR(1325239200, 2011, 12, 6, 31, 0, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1333202399, 2012, 4, 0, 1, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1333202400, 2012, 4, 0, 1, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1333205999, 2012, 4, 0, 1, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1333206000, 2012, 4, 0, 1, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1348927199, 2012, 9, 0, 30, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1348927200, 2012, 9, 0, 30, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1365256799, 2013, 4, 0, 7, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1365256800, 2013, 4, 0, 7, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1365260399, 2013, 4, 0, 7, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1365260400, 2013, 4, 0, 7, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1380376799, 2013, 9, 0, 29, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1380376800, 2013, 9, 0, 29, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1396706399, 2014, 4, 0, 6, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1396706400, 2014, 4, 0, 6, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1396709999, 2014, 4, 0, 6, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1396710000, 2014, 4, 0, 6, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1411826399, 2014, 9, 0, 28, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1411826400, 2014, 9, 0, 28, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1428155999, 2015, 4, 0, 5, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1428156000, 2015, 4, 0, 5, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1428159599, 2015, 4, 0, 5, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1428159600, 2015, 4, 0, 5, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1443275999, 2015, 9, 0, 27, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1443276000, 2015, 9, 0, 27, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1459605599, 2016, 4, 0, 3, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1459605600, 2016, 4, 0, 3, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1459609199, 2016, 4, 0, 3, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1459609200, 2016, 4, 0, 3, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1474725599, 2016, 9, 0, 25, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1474725600, 2016, 9, 0, 25, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1491055199, 2017, 4, 0, 2, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1491055200, 2017, 4, 0, 2, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1491058799, 2017, 4, 0, 2, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1491058800, 2017, 4, 0, 2, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1506175199, 2017, 9, 0, 24, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1506175200, 2017, 9, 0, 24, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1522504799, 2018, 4, 0, 1, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1522504800, 2018, 4, 0, 1, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1522508399, 2018, 4, 0, 1, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1522508400, 2018, 4, 0, 1, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1538229599, 2018, 9, 0, 30, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1538229600, 2018, 9, 0, 30, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1554559199, 2019, 4, 0, 7, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1554559200, 2019, 4, 0, 7, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1554562799, 2019, 4, 0, 7, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1554562800, 2019, 4, 0, 7, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1569679199, 2019, 9, 0, 29, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1569679200, 2019, 9, 0, 29, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1586008799, 2020, 4, 0, 5, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1586008800, 2020, 4, 0, 5, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1586012399, 2020, 4, 0, 5, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1586012400, 2020, 4, 0, 5, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1601128799, 2020, 9, 0, 27, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1601128800, 2020, 9, 0, 27, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1617458399, 2021, 4, 0, 4, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1617458400, 2021, 4, 0, 4, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1617461999, 2021, 4, 0, 4, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1617462000, 2021, 4, 0, 4, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1632578399, 2021, 9, 0, 26, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1632578400, 2021, 9, 0, 26, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1648907999, 2022, 4, 0, 3, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1648908000, 2022, 4, 0, 3, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1648911599, 2022, 4, 0, 3, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1648911600, 2022, 4, 0, 3, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1664027999, 2022, 9, 0, 25, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1664028000, 2022, 9, 0, 25, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1680357599, 2023, 4, 0, 2, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1680357600, 2023, 4, 0, 2, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1680361199, 2023, 4, 0, 2, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1680361200, 2023, 4, 0, 2, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1695477599, 2023, 9, 0, 24, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1695477600, 2023, 9, 0, 24, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1712411999, 2024, 4, 0, 7, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1712412000, 2024, 4, 0, 7, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1712415599, 2024, 4, 0, 7, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1712415600, 2024, 4, 0, 7, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1727531999, 2024, 9, 0, 29, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1727532000, 2024, 9, 0, 29, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1743861599, 2025, 4, 0, 6, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1743861600, 2025, 4, 0, 6, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1743865199, 2025, 4, 0, 6, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1743865200, 2025, 4, 0, 6, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1758981599, 2025, 9, 0, 28, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1758981600, 2025, 9, 0, 28, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1775311199, 2026, 4, 0, 5, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1775311200, 2026, 4, 0, 5, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1775314799, 2026, 4, 0, 5, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1775314800, 2026, 4, 0, 5, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1790431199, 2026, 9, 0, 27, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1790431200, 2026, 9, 0, 27, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1806760799, 2027, 4, 0, 4, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1806760800, 2027, 4, 0, 4, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1806764399, 2027, 4, 0, 4, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1806764400, 2027, 4, 0, 4, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1821880799, 2027, 9, 0, 26, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1821880800, 2027, 9, 0, 26, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1838210399, 2028, 4, 0, 2, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1838210400, 2028, 4, 0, 2, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1838213999, 2028, 4, 0, 2, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1838214000, 2028, 4, 0, 2, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1853330399, 2028, 9, 0, 24, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1853330400, 2028, 9, 0, 24, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1869659999, 2029, 4, 0, 1, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1869660000, 2029, 4, 0, 1, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1869663599, 2029, 4, 0, 1, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1869663600, 2029, 4, 0, 1, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1885384799, 2029, 9, 0, 30, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1885384800, 2029, 9, 0, 30, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1901714399, 2030, 4, 0, 7, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1901714400, 2030, 4, 0, 7, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1901717999, 2030, 4, 0, 7, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1901718000, 2030, 4, 0, 7, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1916834399, 2030, 9, 0, 29, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1916834400, 2030, 9, 0, 29, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1933163999, 2031, 4, 0, 6, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1933164000, 2031, 4, 0, 6, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1933167599, 2031, 4, 0, 6, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1933167600, 2031, 4, 0, 6, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1948283999, 2031, 9, 0, 28, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1948284000, 2031, 9, 0, 28, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1964613599, 2032, 4, 0, 4, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1964613600, 2032, 4, 0, 4, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1964617199, 2032, 4, 0, 4, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1964617200, 2032, 4, 0, 4, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(1979733599, 2032, 9, 0, 26, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(1979733600, 2032, 9, 0, 26, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(1996063199, 2033, 4, 0, 3, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(1996063200, 2033, 4, 0, 3, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(1996066799, 2033, 4, 0, 3, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(1996066800, 2033, 4, 0, 3, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2011183199, 2033, 9, 0, 25, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2011183200, 2033, 9, 0, 25, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2027512799, 2034, 4, 0, 2, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2027512800, 2034, 4, 0, 2, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2027516399, 2034, 4, 0, 2, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2027516400, 2034, 4, 0, 2, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2042632799, 2034, 9, 0, 24, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2042632800, 2034, 9, 0, 24, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2058962399, 2035, 4, 0, 1, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2058962400, 2035, 4, 0, 1, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2058965999, 2035, 4, 0, 1, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2058966000, 2035, 4, 0, 1, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2074687199, 2035, 9, 0, 30, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2074687200, 2035, 9, 0, 30, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2091016799, 2036, 4, 0, 6, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2091016800, 2036, 4, 0, 6, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2091020399, 2036, 4, 0, 6, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2091020400, 2036, 4, 0, 6, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2106136799, 2036, 9, 0, 28, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2106136800, 2036, 9, 0, 28, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2122466399, 2037, 4, 0, 5, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2122466400, 2037, 4, 0, 5, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2122469999, 2037, 4, 0, 5, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2122470000, 2037, 4, 0, 5, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2137586399, 2037, 9, 0, 27, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2137586400, 2037, 9, 0, 27, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2153915999, 2038, 4, 0, 4, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2153916000, 2038, 4, 0, 4, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2153919599, 2038, 4, 0, 4, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2153919600, 2038, 4, 0, 4, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2169035999, 2038, 9, 0, 26, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2169036000, 2038, 9, 0, 26, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2185365599, 2039, 4, 0, 3, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2185365600, 2039, 4, 0, 3, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2185369199, 2039, 4, 0, 3, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2185369200, 2039, 4, 0, 3, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2200485599, 2039, 9, 0, 25, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2200485600, 2039, 9, 0, 25, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2216815199, 2040, 4, 0, 1, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2216815200, 2040, 4, 0, 1, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2216818799, 2040, 4, 0, 1, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2216818800, 2040, 4, 0, 1, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2232539999, 2040, 9, 0, 30, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2232540000, 2040, 9, 0, 30, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2248869599, 2041, 4, 0, 7, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2248869600, 2041, 4, 0, 7, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2248873199, 2041, 4, 0, 7, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2248873200, 2041, 4, 0, 7, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2263989599, 2041, 9, 0, 29, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2263989600, 2041, 9, 0, 29, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2280319199, 2042, 4, 0, 6, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2280319200, 2042, 4, 0, 6, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2280322799, 2042, 4, 0, 6, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2280322800, 2042, 4, 0, 6, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2295439199, 2042, 9, 0, 28, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2295439200, 2042, 9, 0, 28, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2311768799, 2043, 4, 0, 5, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2311768800, 2043, 4, 0, 5, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2311772399, 2043, 4, 0, 5, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2311772400, 2043, 4, 0, 5, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2326888799, 2043, 9, 0, 27, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2326888800, 2043, 9, 0, 27, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2343218399, 2044, 4, 0, 3, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2343218400, 2044, 4, 0, 3, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2343221999, 2044, 4, 0, 3, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2343222000, 2044, 4, 0, 3, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2358338399, 2044, 9, 0, 25, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2358338400, 2044, 9, 0, 25, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2374667999, 2045, 4, 0, 2, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2374668000, 2045, 4, 0, 2, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2374671599, 2045, 4, 0, 2, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2374671600, 2045, 4, 0, 2, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2389787999, 2045, 9, 0, 24, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2389788000, 2045, 9, 0, 24, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2406117599, 2046, 4, 0, 1, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2406117600, 2046, 4, 0, 1, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2406121199, 2046, 4, 0, 1, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2406121200, 2046, 4, 0, 1, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2421842399, 2046, 9, 0, 30, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2421842400, 2046, 9, 0, 30, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2438171999, 2047, 4, 0, 7, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2438172000, 2047, 4, 0, 7, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2438175599, 2047, 4, 0, 7, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2438175600, 2047, 4, 0, 7, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2453291999, 2047, 9, 0, 29, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2453292000, 2047, 9, 0, 29, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2469621599, 2048, 4, 0, 5, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2469621600, 2048, 4, 0, 5, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2469625199, 2048, 4, 0, 5, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2469625200, 2048, 4, 0, 5, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2484741599, 2048, 9, 0, 27, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2484741600, 2048, 9, 0, 27, 4, 0, 0, 1, 50400, "+14", BIDI)
TESTVECTOR(2501071199, 2049, 4, 0, 4, 3, 59, 59, 1, 50400, "+14", BIDI)
TESTVECTOR(2501071200, 2049, 4, 0, 4, 3, 0, 0, 0, 46800, "+13", HINT)
TESTVECTOR(2501074799, 2049, 4, 0, 4, 3, 59, 59, 0, 46800, "+13", HINT)
TESTVECTOR(2501074800, 2049, 4, 0, 4, 4, 0, 0, 0, 46800, "+13", BIDI)
TESTVECTOR(2516191199, 2049, 9, 0, 26, 2, 59, 59, 0, 46800, "+13", BIDI)
TESTVECTOR(2516191200, 2049, 9, 0, 26, 4, 0, 0, 1, 50400, "+14", BIDI)
TEST_END

TEST_BEGIN(ashgabat, "Asia/Ashgabat")
TESTVECTOR(-1441166013, 1924, 5, 4, 1, 23, 59, 59, 0, 14012, "LMT", BIDI)
TESTVECTOR(-1441166012, 1924, 5, 5, 2, 0, 6, 28, 0, 14400, "+04", BIDI)
TESTVECTOR(-1247544001, 1930, 6, 5, 20, 23, 59, 59, 0, 14400, "+04", BIDI)
TESTVECTOR(-1247544000, 1930, 6, 6, 21, 1, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(354913199, 1981, 3, 2, 31, 23, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(354913200, 1981, 4, 3, 1, 1, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(370720799, 1981, 9, 3, 30, 23, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(370720800, 1981, 9, 3, 30, 23, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(370724399, 1981, 9, 3, 30, 23, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(370724400, 1981, 10, 4, 1, 0, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(386449199, 1982, 3, 3, 31, 23, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(386449200, 1982, 4, 4, 1, 1, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(402256799, 1982, 9, 4, 30, 23, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(402256800, 1982, 9, 4, 30, 23, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(402260399, 1982, 9, 4, 30, 23, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(402260400, 1982, 10, 5, 1, 0, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(417985199, 1983, 3, 4, 31, 23, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(417985200, 1983, 4, 5, 1, 1, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(433792799, 1983, 9, 5, 30, 23, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(433792800, 1983, 9, 5, 30, 23, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(433796399, 1983, 9, 5, 30, 23, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(433796400, 1983, 10, 6, 1, 0, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(449607599, 1984, 3, 6, 31, 23, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(449607600, 1984, 4, 0, 1, 1, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(465339599, 1984, 9, 0, 30, 2, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(465339600, 1984, 9, 0, 30, 2, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(465343199, 1984, 9, 0, 30, 2, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(465343200, 1984, 9, 0, 30, 3, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(481064399, 1985, 3, 0, 31, 1, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(481064400, 1985, 3, 0, 31, 3, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(496789199, 1985, 9, 0, 29, 2, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(496789200, 1985, 9, 0, 29, 2, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(496792799, 1985, 9, 0, 29, 2, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(496792800, 1985, 9, 0, 29, 3, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(512513999, 1986, 3, 0, 30, 1, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(512514000, 1986, 3, 0, 30, 3, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(528238799, 1986, 9, 0, 28, 2, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(528238800, 1986, 9, 0, 28, 2, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(528242399, 1986, 9, 0, 28, 2, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(528242400, 1986, 9, 0, 28, 3, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(543963599, 1987, 3, 0, 29, 1, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(543963600, 1987, 3, 0, 29, 3, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(559688399, 1987, 9, 0, 27, 2, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(559688400, 1987, 9, 0, 27, 2, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(559691999, 1987, 9, 0, 27, 2, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(559692000, 1987, 9, 0, 27, 3, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(575413199, 1988, 3, 0, 27, 1, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(575413200, 1988, 3, 0, 27, 3, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(591137999, 1988, 9, 0, 25, 2, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(591138000, 1988, 9, 0, 25, 2, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(591141599, 1988, 9, 0, 25, 2, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(591141600, 1988, 9, 0, 25, 3, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(606862799, 1989, 3, 0, 26, 1, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(606862800, 1989, 3, 0, 26, 3, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(622587599, 1989, 9, 0, 24, 2, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(622587600, 1989, 9, 0, 24, 2, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(622591199, 1989, 9, 0, 24, 2, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(622591200, 1989, 9, 0, 24, 3, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(638312399, 1990, 3, 0, 25, 1, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(638312400, 1990, 3, 0, 25, 3, 0, 0, 1, 21600, "+06", BIDI)
TESTVECTOR(654641999, 1990, 9, 0, 30, 2, 59, 59, 1, 21600, "+06", BIDI)
TESTVECTOR(654642000, 1990, 9, 0, 30, 2, 0, 0, 0, 18000, "+05", HINT)
TESTVECTOR(654645599, 1990, 9, 0, 30, 2, 59, 59, 0, 18000, "+05", HINT)
TESTVECTOR(654645600, 1990, 9, 0, 30, 3, 0, 0, 0, 18000, "+05", BIDI)
TESTVECTOR(670366799, 1991, 3, 0, 31, 1, 59, 59, 0, 18000, "+05", BIDI)
TESTVECTOR(670366800, 1991, 3, 0, 31, 2, 0, 0, 1, 18000, "+05", BIDI)
TESTVECTOR(686095199, 1991, 9, 0, 29, 2, 59, 59, 1, 18000, "+05", BIDI)
TESTVECTOR(686095200, 1991, 9, 0, 29, 2, 0, 0, 0, 14400, "+04", HINT)
TESTVECTOR(686098799, 1991, 9, 0, 29, 2, 59, 59, 0, 14400, "+04", HINT)
TESTVECTOR(686098800, 1991, 9, 0, 29, 3, 0, 0, 0, 14400, "+04", BIDI)
TESTVECTOR(695771999, 1992, 1, 0, 19, 1, 59, 59, 0, 14400, "+04", BIDI)
TESTVECTOR(695772000, 1992, 1, 0, 19, 3, 0, 0, 0, 18000, "+05", BIDI)
TEST_END

TEST_BEGIN(belize, "America/Belize")
TESTVECTOR(-1822500433, 1912, 3, 0, 31, 23, 59, 59, 0, -21168, "LMT", BIDI)
TESTVECTOR(-1822500432, 1912, 3, 0, 31, 23, 52, 48, 0, -21600, "CST", UNI)
TESTVECTOR(-1822500001, 1912, 3, 0, 31, 23, 59, 59, 0, -21600, "CST", UNI)
TESTVECTOR(-1822500000, 1912, 4, 1, 1, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1616954401, 1918, 10, 6, 5, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1616954400, 1918, 10, 0, 6, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1606069801, 1919, 2, 6, 8, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1606069800, 1919, 2, 6, 8, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1606068001, 1919, 2, 6, 8, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1606068000, 1919, 2, 0, 9, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1585504801, 1919, 10, 6, 4, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1585504800, 1919, 10, 0, 5, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1574015401, 1920, 2, 6, 14, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1574015400, 1920, 2, 6, 14, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1574013601, 1920, 2, 6, 14, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1574013600, 1920, 2, 0, 15, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1554055201, 1920, 10, 6, 2, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1554055200, 1920, 10, 0, 3, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1542565801, 1921, 2, 6, 12, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1542565800, 1921, 2, 6, 12, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1542564001, 1921, 2, 6, 12, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1542564000, 1921, 2, 0, 13, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1522605601, 1921, 10, 6, 1, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1522605600, 1921, 10, 0, 2, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1511116201, 1922, 2, 6, 11, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1511116200, 1922, 2, 6, 11, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1511114401, 1922, 2, 6, 11, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1511114400, 1922, 2, 0, 12, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1490551201, 1922, 10, 6, 7, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1490551200, 1922, 10, 0, 8, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1479666601, 1923, 2, 6, 10, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1479666600, 1923, 2, 6, 10, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1479664801, 1923, 2, 6, 10, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1479664800, 1923, 2, 0, 11, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1459101601, 1923, 10, 6, 6, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1459101600, 1923, 10, 0, 7, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1448217001, 1924, 2, 6, 9, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1448217000, 1924, 2, 6, 9, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1448215201, 1924, 2, 6, 9, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1448215200, 1924, 2, 0, 10, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1427652001, 1924, 10, 6, 4, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1427652000, 1924, 10, 0, 5, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1416162601, 1925, 2, 6, 14, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1416162600, 1925, 2, 6, 14, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1416160801, 1925, 2, 6, 14, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1416160800, 1925, 2, 0, 15, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1396202401, 1925, 10, 6, 3, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1396202400, 1925, 10, 0, 4, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1384713001, 1926, 2, 6, 13, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1384713000, 1926, 2, 6, 13, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1384711201, 1926, 2, 6, 13, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1384711200, 1926, 2, 0, 14, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1364752801, 1926, 10, 6, 2, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1364752800, 1926, 10, 0, 3, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1353263401, 1927, 2, 6, 12, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1353263400, 1927, 2, 6, 12, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1353261601, 1927, 2, 6, 12, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1353261600, 1927, 2, 0, 13, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1333303201, 1927, 10, 6, 1, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1333303200, 1927, 10, 0, 2, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1321813801, 1928, 2, 6, 11, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1321813800, 1928, 2, 6, 11, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1321812001, 1928, 2, 6, 11, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1321812000, 1928, 2, 0, 12, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1301248801, 1928, 10, 6, 6, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1301248800, 1928, 10, 0, 7, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1290364201, 1929, 2, 6, 9, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1290364200, 1929, 2, 6, 9, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1290362401, 1929, 2, 6, 9, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1290362400, 1929, 2, 0, 10, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1269799201, 1929, 10, 6, 5, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1269799200, 1929, 10, 0, 6, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1258914601, 1930, 2, 6, 8, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1258914600, 1930, 2, 6, 8, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1258912801, 1930, 2, 6, 8, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1258912800, 1930, 2, 0, 9, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1238349601, 1930, 10, 6, 4, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1238349600, 1930, 10, 0, 5, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1226860201, 1931, 2, 6, 14, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1226860200, 1931, 2, 6, 14, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1226858401, 1931, 2, 6, 14, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1226858400, 1931, 2, 0, 15, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1206900001, 1931, 10, 6, 3, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1206900000, 1931, 10, 0, 4, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1195410601, 1932, 2, 6, 13, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1195410600, 1932, 2, 6, 13, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1195408801, 1932, 2, 6, 13, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1195408800, 1932, 2, 0, 14, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1175450401, 1932, 10, 6, 1, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1175450400, 1932, 10, 0, 2, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1163961001, 1933, 2, 6, 11, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1163961000, 1933, 2, 6, 11, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1163959201, 1933, 2, 6, 11, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1163959200, 1933, 2, 0, 12, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1143396001, 1933, 10, 6, 7, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1143396000, 1933, 10, 0, 8, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1132511401, 1934, 2, 6, 10, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1132511400, 1934, 2, 6, 10, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1132509601, 1934, 2, 6, 10, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1132509600, 1934, 2, 0, 11, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1111946401, 1934, 10, 6, 6, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1111946400, 1934, 10, 0, 7, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1101061801, 1935, 2, 6, 9, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1101061800, 1935, 2, 6, 9, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1101060001, 1935, 2, 6, 9, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1101060000, 1935, 2, 0, 10, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1080496801, 1935, 10, 6, 5, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1080496800, 1935, 10, 0, 6, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1069612201, 1936, 2, 6, 8, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1069612200, 1936, 2, 6, 8, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1069610401, 1936, 2, 6, 8, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1069610400, 1936, 2, 0, 9, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1049047201, 1936, 10, 6, 3, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1049047200, 1936, 10, 0, 4, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1037557801, 1937, 2, 6, 13, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1037557800, 1937, 2, 6, 13, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1037556001, 1937, 2, 6, 13, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1037556000, 1937, 2, 0, 14, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1017597601, 1937, 10, 6, 2, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1017597600, 1937, 10, 0, 3, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1006108201, 1938, 2, 6, 12, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-1006108200, 1938, 2, 6, 12, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1006106401, 1938, 2, 6, 12, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1006106400, 1938, 2, 0, 13, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-986148001, 1938, 10, 6, 1, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-986148000, 1938, 10, 0, 2, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-974658601, 1939, 2, 6, 11, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-974658600, 1939, 2, 6, 11, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-974656801, 1939, 2, 6, 11, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-974656800, 1939, 2, 0, 12, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-954093601, 1939, 10, 6, 7, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-954093600, 1939, 10, 0, 8, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-943209001, 1940, 2, 6, 10, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-943209000, 1940, 2, 6, 10, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-943207201, 1940, 2, 6, 10, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-943207200, 1940, 2, 0, 11, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-922644001, 1940, 10, 6, 5, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-922644000, 1940, 10, 0, 6, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-911759401, 1941, 2, 6, 8, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-911759400, 1941, 2, 6, 8, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-911757601, 1941, 2, 6, 8, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-911757600, 1941, 2, 0, 9, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-891194401, 1941, 10, 6, 4, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-891194400, 1941, 10, 0, 5, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-879705001, 1942, 2, 6, 14, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-879705000, 1942, 2, 6, 14, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-879703201, 1942, 2, 6, 14, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-879703200, 1942, 2, 0, 15, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-859744801, 1942, 10, 6, 3, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-859744800, 1942, 10, 0, 4, 0, 30, 0, 1, -19800, "-0530", BIDI)
TESTVECTOR(-848255401, 1943, 2, 6, 13, 23, 59, 59, 1, -19800, "-0530", BIDI)
TESTVECTOR(-848255400, 1943, 2, 6, 13, 23, 30, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-848253601, 1943, 2, 6, 13, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-848253600, 1943, 2, 0, 14, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(123919199, 1973, 12, 2, 4, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(123919200, 1973, 12, 3, 5, 1, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(129617999, 1974, 2, 5, 8, 23, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(129618000, 1974, 2, 5, 8, 23, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(129621599, 1974, 2, 5, 8, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(129621600, 1974, 2, 6, 9, 0, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(409039199, 1982, 12, 5, 17, 23, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(409039200, 1982, 12, 6, 18, 1, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(413873999, 1983, 2, 5, 11, 23, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(413874000, 1983, 2, 5, 11, 23, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(413877599, 1983, 2, 5, 11, 23, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(413877600, 1983, 2, 6, 12, 0, 0, 0, 0, -21600, "CST", BIDI)
TEST_END

TEST_BEGIN(dhaka, "Asia/Dhaka")
TESTVECTOR(-2524543301, 1889, 12, 2, 31, 23, 59, 59, 0, 21700, "LMT", BIDI)
TESTVECTOR(-2524543300, 1889, 12, 2, 31, 23, 51, 40, 0, 21200, "HMT", UNI)
TESTVECTOR(-2524542801, 1889, 12, 2, 31, 23, 59, 59, 0, 21200, "HMT", UNI)
TESTVECTOR(-2524542800, 1890, 1, 3, 1, 0, 0, 0, 0, 21200, "HMT", BIDI)
TESTVECTOR(-891582801, 1941, 9, 2, 30, 23, 59, 59, 0, 21200, "HMT", BIDI)
TESTVECTOR(-891582800, 1941, 10, 3, 1, 0, 36, 40, 0, 23400, "+0630", BIDI)
TESTVECTOR(-872058601, 1942, 5, 4, 14, 23, 59, 59, 0, 23400, "+0630", BIDI)
TESTVECTOR(-872058600, 1942, 5, 4, 14, 23, 0, 0, 0, 19800, "+0530", UNI)
TESTVECTOR(-872055001, 1942, 5, 4, 14, 23, 59, 59, 0, 19800, "+0530", UNI)
TESTVECTOR(-872055000, 1942, 5, 5, 15, 0, 0, 0, 0, 19800, "+0530", BIDI)
TESTVECTOR(-862637401, 1942, 8, 1, 31, 23, 59, 59, 0, 19800, "+0530", BIDI)
TESTVECTOR(-862637400, 1942, 9, 2, 1, 1, 0, 0, 0, 23400, "+0630", BIDI)
TESTVECTOR(-576138601, 1951, 9, 6, 29, 23, 59, 59, 0, 23400, "+0630", BIDI)
TESTVECTOR(-576138600, 1951, 9, 6, 29, 23, 30, 0, 0, 21600, "+06", UNI)
TESTVECTOR(-576136801, 1951, 9, 6, 29, 23, 59, 59, 0, 21600, "+06", UNI)
TESTVECTOR(-576136800, 1951, 9, 0, 30, 0, 0, 0, 0, 21600, "+06", BIDI)
TESTVECTOR(1245430799, 2009, 6, 5, 19, 22, 59, 59, 0, 21600, "+06", BIDI)
TESTVECTOR(1245430800, 2009, 6, 6, 20, 0, 0, 0, 1, 25200, "+07", BIDI)
TESTVECTOR(1262278799, 2009, 12, 4, 31, 23, 59, 59, 1, 25200, "+07", BIDI)
TESTVECTOR(1262278800, 2009, 12, 4, 31, 23, 0, 0, 0, 21600, "+06", HINT)
TESTVECTOR(1262282399, 2009, 12, 4, 31, 23, 59, 59, 0, 21600, "+06", HINT)
TESTVECTOR(1262282400, 2010, 1, 5, 1, 0, 0, 0, 0, 21600, "+06", BIDI)
TEST_END

TEST_BEGIN(easter_island, "Chile/EasterIsland")
TESTVECTOR(-2524495353, 1889, 12, 2, 31, 23, 59, 59, 0, -26248, "LMT", BIDI)
TESTVECTOR(-2524495352, 1890, 1, 3, 1, 0, 0, 0, 0, -26248, "EMT", BIDI)
TESTVECTOR(-1178124153, 1932, 8, 3, 31, 23, 59, 59, 0, -26248, "EMT", BIDI)
TESTVECTOR(-1178124152, 1932, 9, 4, 1, 0, 17, 28, 0, -25200, "-07", BIDI)
TESTVECTOR(-36619201, 1968, 11, 6, 2, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(-36619200, 1968, 11, 6, 2, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(-23922001, 1969, 3, 6, 29, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(-23922000, 1969, 3, 6, 29, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(-23918401, 1969, 3, 6, 29, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(-23918400, 1969, 3, 6, 29, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(-3355201, 1969, 11, 6, 22, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(-3355200, 1969, 11, 6, 22, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(7527599, 1970, 3, 6, 28, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(7527600, 1970, 3, 6, 28, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(7531199, 1970, 3, 6, 28, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(7531200, 1970, 3, 6, 28, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(24465599, 1970, 10, 6, 10, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(24465600, 1970, 10, 6, 10, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(37767599, 1971, 3, 6, 13, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(37767600, 1971, 3, 6, 13, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(37771199, 1971, 3, 6, 13, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(37771200, 1971, 3, 6, 13, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(55915199, 1971, 10, 6, 9, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(55915200, 1971, 10, 6, 9, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(69217199, 1972, 3, 6, 11, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(69217200, 1972, 3, 6, 11, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(69220799, 1972, 3, 6, 11, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(69220800, 1972, 3, 6, 11, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(87969599, 1972, 10, 6, 14, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(87969600, 1972, 10, 6, 14, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(100666799, 1973, 3, 6, 10, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(100666800, 1973, 3, 6, 10, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(100670399, 1973, 3, 6, 10, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(100670400, 1973, 3, 6, 10, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(118209599, 1973, 9, 6, 29, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(118209600, 1973, 9, 6, 29, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(132116399, 1974, 3, 6, 9, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(132116400, 1974, 3, 6, 9, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(132119999, 1974, 3, 6, 9, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(132120000, 1974, 3, 6, 9, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(150868799, 1974, 10, 6, 12, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(150868800, 1974, 10, 6, 12, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(163565999, 1975, 3, 6, 8, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(163566000, 1975, 3, 6, 8, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(163569599, 1975, 3, 6, 8, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(163569600, 1975, 3, 6, 8, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(182318399, 1975, 10, 6, 11, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(182318400, 1975, 10, 6, 11, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(195620399, 1976, 3, 6, 13, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(195620400, 1976, 3, 6, 13, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(195623999, 1976, 3, 6, 13, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(195624000, 1976, 3, 6, 13, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(213767999, 1976, 10, 6, 9, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(213768000, 1976, 10, 6, 9, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(227069999, 1977, 3, 6, 12, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(227070000, 1977, 3, 6, 12, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(227073599, 1977, 3, 6, 12, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(227073600, 1977, 3, 6, 12, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(245217599, 1977, 10, 6, 8, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(245217600, 1977, 10, 6, 8, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(258519599, 1978, 3, 6, 11, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(258519600, 1978, 3, 6, 11, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(258523199, 1978, 3, 6, 11, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(258523200, 1978, 3, 6, 11, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(277271999, 1978, 10, 6, 14, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(277272000, 1978, 10, 6, 14, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(289969199, 1979, 3, 6, 10, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(289969200, 1979, 3, 6, 10, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(289972799, 1979, 3, 6, 10, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(289972800, 1979, 3, 6, 10, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(308721599, 1979, 10, 6, 13, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(308721600, 1979, 10, 6, 13, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(321418799, 1980, 3, 6, 8, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(321418800, 1980, 3, 6, 8, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(321422399, 1980, 3, 6, 8, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(321422400, 1980, 3, 6, 8, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(340171199, 1980, 10, 6, 11, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(340171200, 1980, 10, 6, 11, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(353473199, 1981, 3, 6, 14, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(353473200, 1981, 3, 6, 14, 20, 0, 0, 0, -25200, "-07", HINT)
TESTVECTOR(353476799, 1981, 3, 6, 14, 20, 59, 59, 0, -25200, "-07", HINT)
TESTVECTOR(353476800, 1981, 3, 6, 14, 21, 0, 0, 0, -25200, "-07", BIDI)
TESTVECTOR(371620799, 1981, 10, 6, 10, 20, 59, 59, 0, -25200, "-07", BIDI)
TESTVECTOR(371620800, 1981, 10, 6, 10, 22, 0, 0, 1, -21600, "-06", BIDI)
TESTVECTOR(384922799, 1982, 3, 6, 13, 20, 59, 59, 1, -21600, "-06", BIDI)
TESTVECTOR(384922800, 1982, 3, 6, 13, 21, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(403070399, 1982, 10, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(403070400, 1982, 10, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(416372399, 1983, 3, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(416372400, 1983, 3, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(416375999, 1983, 3, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(416376000, 1983, 3, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(434519999, 1983, 10, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(434520000, 1983, 10, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(447821999, 1984, 3, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(447822000, 1984, 3, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(447825599, 1984, 3, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(447825600, 1984, 3, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(466574399, 1984, 10, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(466574400, 1984, 10, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(479271599, 1985, 3, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(479271600, 1985, 3, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(479275199, 1985, 3, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(479275200, 1985, 3, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(498023999, 1985, 10, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(498024000, 1985, 10, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(510721199, 1986, 3, 6, 8, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(510721200, 1986, 3, 6, 8, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(510724799, 1986, 3, 6, 8, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(510724800, 1986, 3, 6, 8, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(529473599, 1986, 10, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(529473600, 1986, 10, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(545194799, 1987, 4, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(545194800, 1987, 4, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(545198399, 1987, 4, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(545198400, 1987, 4, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(560923199, 1987, 10, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(560923200, 1987, 10, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(574225199, 1988, 3, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(574225200, 1988, 3, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(574228799, 1988, 3, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(574228800, 1988, 3, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(592372799, 1988, 10, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(592372800, 1988, 10, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(605674799, 1989, 3, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(605674800, 1989, 3, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(605678399, 1989, 3, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(605678400, 1989, 3, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(624427199, 1989, 10, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(624427200, 1989, 10, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(637124399, 1990, 3, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(637124400, 1990, 3, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(637127999, 1990, 3, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(637128000, 1990, 3, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(653457599, 1990, 9, 6, 15, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(653457600, 1990, 9, 6, 15, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(668573999, 1991, 3, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(668574000, 1991, 3, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(668577599, 1991, 3, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(668577600, 1991, 3, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(687326399, 1991, 10, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(687326400, 1991, 10, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(700628399, 1992, 3, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(700628400, 1992, 3, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(700631999, 1992, 3, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(700632000, 1992, 3, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(718775999, 1992, 10, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(718776000, 1992, 10, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(732077999, 1993, 3, 6, 13, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(732078000, 1993, 3, 6, 13, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(732081599, 1993, 3, 6, 13, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(732081600, 1993, 3, 6, 13, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(750225599, 1993, 10, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(750225600, 1993, 10, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(763527599, 1994, 3, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(763527600, 1994, 3, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(763531199, 1994, 3, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(763531200, 1994, 3, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(781675199, 1994, 10, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(781675200, 1994, 10, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(794977199, 1995, 3, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(794977200, 1995, 3, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(794980799, 1995, 3, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(794980800, 1995, 3, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(813729599, 1995, 10, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(813729600, 1995, 10, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(826426799, 1996, 3, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(826426800, 1996, 3, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(826430399, 1996, 3, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(826430400, 1996, 3, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(845179199, 1996, 10, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(845179200, 1996, 10, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(859690799, 1997, 3, 6, 29, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(859690800, 1997, 3, 6, 29, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(859694399, 1997, 3, 6, 29, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(859694400, 1997, 3, 6, 29, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(876628799, 1997, 10, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(876628800, 1997, 10, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(889930799, 1998, 3, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(889930800, 1998, 3, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(889934399, 1998, 3, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(889934400, 1998, 3, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(906868799, 1998, 9, 6, 26, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(906868800, 1998, 9, 6, 26, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(923194799, 1999, 4, 6, 3, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(923194800, 1999, 4, 6, 3, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(923198399, 1999, 4, 6, 3, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(923198400, 1999, 4, 6, 3, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(939527999, 1999, 10, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(939528000, 1999, 10, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(952829999, 2000, 3, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(952830000, 2000, 3, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(952833599, 2000, 3, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(952833600, 2000, 3, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(971582399, 2000, 10, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(971582400, 2000, 10, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(984279599, 2001, 3, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(984279600, 2001, 3, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(984283199, 2001, 3, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(984283200, 2001, 3, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1003031999, 2001, 10, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1003032000, 2001, 10, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1015729199, 2002, 3, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1015729200, 2002, 3, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1015732799, 2002, 3, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1015732800, 2002, 3, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1034481599, 2002, 10, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1034481600, 2002, 10, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1047178799, 2003, 3, 6, 8, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1047178800, 2003, 3, 6, 8, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1047182399, 2003, 3, 6, 8, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1047182400, 2003, 3, 6, 8, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1065931199, 2003, 10, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1065931200, 2003, 10, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1079233199, 2004, 3, 6, 13, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1079233200, 2004, 3, 6, 13, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1079236799, 2004, 3, 6, 13, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1079236800, 2004, 3, 6, 13, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1097380799, 2004, 10, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1097380800, 2004, 10, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1110682799, 2005, 3, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1110682800, 2005, 3, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1110686399, 2005, 3, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1110686400, 2005, 3, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1128830399, 2005, 10, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1128830400, 2005, 10, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1142132399, 2006, 3, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1142132400, 2006, 3, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1142135999, 2006, 3, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1142136000, 2006, 3, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1160884799, 2006, 10, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1160884800, 2006, 10, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1173581999, 2007, 3, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1173582000, 2007, 3, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1173585599, 2007, 3, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1173585600, 2007, 3, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1192334399, 2007, 10, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1192334400, 2007, 10, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1206845999, 2008, 3, 6, 29, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1206846000, 2008, 3, 6, 29, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1206849599, 2008, 3, 6, 29, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1206849600, 2008, 3, 6, 29, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1223783999, 2008, 10, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1223784000, 2008, 10, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1237085999, 2009, 3, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1237086000, 2009, 3, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1237089599, 2009, 3, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1237089600, 2009, 3, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1255233599, 2009, 10, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1255233600, 2009, 10, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1270349999, 2010, 4, 6, 3, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1270350000, 2010, 4, 6, 3, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1270353599, 2010, 4, 6, 3, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1270353600, 2010, 4, 6, 3, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1286683199, 2010, 10, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1286683200, 2010, 10, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1304823599, 2011, 5, 6, 7, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1304823600, 2011, 5, 6, 7, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1304827199, 2011, 5, 6, 7, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1304827200, 2011, 5, 6, 7, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1313899199, 2011, 8, 6, 20, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1313899200, 2011, 8, 6, 20, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1335668399, 2012, 4, 6, 28, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1335668400, 2012, 4, 6, 28, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1335671999, 2012, 4, 6, 28, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1335672000, 2012, 4, 6, 28, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1346558399, 2012, 9, 6, 1, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1346558400, 2012, 9, 6, 1, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1367117999, 2013, 4, 6, 27, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1367118000, 2013, 4, 6, 27, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1367121599, 2013, 4, 6, 27, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1367121600, 2013, 4, 6, 27, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1378612799, 2013, 9, 6, 7, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1378612800, 2013, 9, 6, 7, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1398567599, 2014, 4, 6, 26, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1398567600, 2014, 4, 6, 26, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1398571199, 2014, 4, 6, 26, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1398571200, 2014, 4, 6, 26, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1410062399, 2014, 9, 6, 6, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1410062400, 2014, 9, 6, 6, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1463281199, 2016, 5, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1463281200, 2016, 5, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1463284799, 2016, 5, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1463284800, 2016, 5, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1471147199, 2016, 8, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1471147200, 2016, 8, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1494730799, 2017, 5, 6, 13, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1494730800, 2017, 5, 6, 13, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1494734399, 2017, 5, 6, 13, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1494734400, 2017, 5, 6, 13, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1502596799, 2017, 8, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1502596800, 2017, 8, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1526180399, 2018, 5, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1526180400, 2018, 5, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1526183999, 2018, 5, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1526184000, 2018, 5, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1534046399, 2018, 8, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1534046400, 2018, 8, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1557629999, 2019, 5, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1557630000, 2019, 5, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1557633599, 2019, 5, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1557633600, 2019, 5, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1565495999, 2019, 8, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1565496000, 2019, 8, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1589079599, 2020, 5, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1589079600, 2020, 5, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1589083199, 2020, 5, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1589083200, 2020, 5, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1596945599, 2020, 8, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1596945600, 2020, 8, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1620529199, 2021, 5, 6, 8, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1620529200, 2021, 5, 6, 8, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1620532799, 2021, 5, 6, 8, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1620532800, 2021, 5, 6, 8, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1628999999, 2021, 8, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1629000000, 2021, 8, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1652583599, 2022, 5, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1652583600, 2022, 5, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1652587199, 2022, 5, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1652587200, 2022, 5, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1660449599, 2022, 8, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1660449600, 2022, 8, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1684033199, 2023, 5, 6, 13, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1684033200, 2023, 5, 6, 13, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1684036799, 2023, 5, 6, 13, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1684036800, 2023, 5, 6, 13, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1691899199, 2023, 8, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1691899200, 2023, 8, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1715482799, 2024, 5, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1715482800, 2024, 5, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1715486399, 2024, 5, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1715486400, 2024, 5, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1723348799, 2024, 8, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1723348800, 2024, 8, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1746932399, 2025, 5, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1746932400, 2025, 5, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1746935999, 2025, 5, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1746936000, 2025, 5, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1754798399, 2025, 8, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1754798400, 2025, 8, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1778381999, 2026, 5, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1778382000, 2026, 5, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1778385599, 2026, 5, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1778385600, 2026, 5, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1786247999, 2026, 8, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1786248000, 2026, 8, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1809831599, 2027, 5, 6, 8, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1809831600, 2027, 5, 6, 8, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1809835199, 2027, 5, 6, 8, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1809835200, 2027, 5, 6, 8, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1818302399, 2027, 8, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1818302400, 2027, 8, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1841885999, 2028, 5, 6, 13, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1841886000, 2028, 5, 6, 13, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1841889599, 2028, 5, 6, 13, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1841889600, 2028, 5, 6, 13, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1849751999, 2028, 8, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1849752000, 2028, 8, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1873335599, 2029, 5, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1873335600, 2029, 5, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1873339199, 2029, 5, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1873339200, 2029, 5, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1881201599, 2029, 8, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1881201600, 2029, 8, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1904785199, 2030, 5, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1904785200, 2030, 5, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1904788799, 2030, 5, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1904788800, 2030, 5, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1912651199, 2030, 8, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1912651200, 2030, 8, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1936234799, 2031, 5, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1936234800, 2031, 5, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1936238399, 2031, 5, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1936238400, 2031, 5, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1944100799, 2031, 8, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1944100800, 2031, 8, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1967684399, 2032, 5, 6, 8, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1967684400, 2032, 5, 6, 8, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1967687999, 2032, 5, 6, 8, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1967688000, 2032, 5, 6, 8, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(1976155199, 2032, 8, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(1976155200, 2032, 8, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(1999738799, 2033, 5, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(1999738800, 2033, 5, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(1999742399, 2033, 5, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(1999742400, 2033, 5, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2007604799, 2033, 8, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2007604800, 2033, 8, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2031188399, 2034, 5, 6, 13, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2031188400, 2034, 5, 6, 13, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2031191999, 2034, 5, 6, 13, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2031192000, 2034, 5, 6, 13, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2039054399, 2034, 8, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2039054400, 2034, 8, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2062637999, 2035, 5, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2062638000, 2035, 5, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2062641599, 2035, 5, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2062641600, 2035, 5, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2070503999, 2035, 8, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2070504000, 2035, 8, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2094087599, 2036, 5, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2094087600, 2036, 5, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2094091199, 2036, 5, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2094091200, 2036, 5, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2101953599, 2036, 8, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2101953600, 2036, 8, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2125537199, 2037, 5, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2125537200, 2037, 5, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2125540799, 2037, 5, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2125540800, 2037, 5, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2133403199, 2037, 8, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2133403200, 2037, 8, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2156986799, 2038, 5, 6, 8, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2156986800, 2038, 5, 6, 8, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2156990399, 2038, 5, 6, 8, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2156990400, 2038, 5, 6, 8, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2165457599, 2038, 8, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2165457600, 2038, 8, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2189041199, 2039, 5, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2189041200, 2039, 5, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2189044799, 2039, 5, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2189044800, 2039, 5, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2196907199, 2039, 8, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2196907200, 2039, 8, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2220490799, 2040, 5, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2220490800, 2040, 5, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2220494399, 2040, 5, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2220494400, 2040, 5, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2228356799, 2040, 8, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2228356800, 2040, 8, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2251940399, 2041, 5, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2251940400, 2041, 5, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2251943999, 2041, 5, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2251944000, 2041, 5, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2259806399, 2041, 8, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2259806400, 2041, 8, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2283389999, 2042, 5, 6, 10, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2283390000, 2042, 5, 6, 10, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2283393599, 2042, 5, 6, 10, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2283393600, 2042, 5, 6, 10, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2291255999, 2042, 8, 6, 9, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2291256000, 2042, 8, 6, 9, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2314839599, 2043, 5, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2314839600, 2043, 5, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2314843199, 2043, 5, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2314843200, 2043, 5, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2322705599, 2043, 8, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2322705600, 2043, 8, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2346893999, 2044, 5, 6, 14, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2346894000, 2044, 5, 6, 14, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2346897599, 2044, 5, 6, 14, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2346897600, 2044, 5, 6, 14, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2354759999, 2044, 8, 6, 13, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2354760000, 2044, 8, 6, 13, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2378343599, 2045, 5, 6, 13, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2378343600, 2045, 5, 6, 13, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2378347199, 2045, 5, 6, 13, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2378347200, 2045, 5, 6, 13, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2386209599, 2045, 8, 6, 12, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2386209600, 2045, 8, 6, 12, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2409793199, 2046, 5, 6, 12, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2409793200, 2046, 5, 6, 12, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2409796799, 2046, 5, 6, 12, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2409796800, 2046, 5, 6, 12, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2417659199, 2046, 8, 6, 11, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2417659200, 2046, 8, 6, 11, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2441242799, 2047, 5, 6, 11, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2441242800, 2047, 5, 6, 11, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2441246399, 2047, 5, 6, 11, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2441246400, 2047, 5, 6, 11, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2449108799, 2047, 8, 6, 10, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2449108800, 2047, 8, 6, 10, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2472692399, 2048, 5, 6, 9, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2472692400, 2048, 5, 6, 9, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2472695999, 2048, 5, 6, 9, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2472696000, 2048, 5, 6, 9, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2480558399, 2048, 8, 6, 8, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2480558400, 2048, 8, 6, 8, 23, 0, 0, 1, -18000, "-05", BIDI)
TESTVECTOR(2504141999, 2049, 5, 6, 8, 21, 59, 59, 1, -18000, "-05", BIDI)
TESTVECTOR(2504142000, 2049, 5, 6, 8, 21, 0, 0, 0, -21600, "-06", HINT)
TESTVECTOR(2504145599, 2049, 5, 6, 8, 21, 59, 59, 0, -21600, "-06", HINT)
TESTVECTOR(2504145600, 2049, 5, 6, 8, 22, 0, 0, 0, -21600, "-06", BIDI)
TESTVECTOR(2512612799, 2049, 8, 6, 14, 21, 59, 59, 0, -21600, "-06", BIDI)
TESTVECTOR(2512612800, 2049, 8, 6, 14, 23, 0, 0, 1, -18000, "-05", BIDI)
TEST_END

TEST_BEGIN(gaza, "Asia/Gaza")
TESTVECTOR(-2185409873, 1900, 9, 0, 30, 23, 59, 59, 0, 8272, "LMT", BIDI)
TESTVECTOR(-2185409872, 1900, 9, 0, 30, 23, 42, 8, 0, 7200, "EET", UNI)
TESTVECTOR(-2185408801, 1900, 9, 0, 30, 23, 59, 59, 0, 7200, "EET", UNI)
TESTVECTOR(-2185408800, 1900, 10, 1, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-933645601, 1940, 5, 5, 31, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-933645600, 1940, 6, 6, 1, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-857358001, 1942, 10, 6, 31, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-857358000, 1942, 10, 6, 31, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-857354401, 1942, 10, 6, 31, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-857354400, 1942, 11, 0, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-844300801, 1943, 4, 4, 1, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-844300800, 1943, 4, 4, 1, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-825822001, 1943, 10, 0, 31, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-825822000, 1943, 10, 0, 31, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-825818401, 1943, 10, 0, 31, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-825818400, 1943, 11, 1, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-812685601, 1944, 3, 5, 31, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-812685600, 1944, 4, 6, 1, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-794199601, 1944, 10, 2, 31, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-794199600, 1944, 10, 2, 31, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-794196001, 1944, 10, 2, 31, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-794196000, 1944, 11, 3, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-779853601, 1945, 4, 0, 15, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-779853600, 1945, 4, 1, 16, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-762656401, 1945, 11, 4, 1, 1, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-762656400, 1945, 11, 4, 1, 1, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-762652801, 1945, 11, 4, 1, 1, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-762652800, 1945, 11, 4, 1, 2, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-748310401, 1946, 4, 2, 16, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-748310400, 1946, 4, 2, 16, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-731127601, 1946, 10, 4, 31, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-731127600, 1946, 10, 4, 31, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-731124001, 1946, 10, 4, 31, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-731124000, 1946, 11, 5, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-399088801, 1957, 5, 4, 9, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-399088800, 1957, 5, 5, 10, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-386650801, 1957, 9, 1, 30, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-386650800, 1957, 9, 1, 30, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-386647201, 1957, 9, 1, 30, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-386647200, 1957, 10, 2, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-368330401, 1958, 4, 3, 30, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-368330400, 1958, 5, 4, 1, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-355114801, 1958, 9, 2, 30, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-355114800, 1958, 9, 2, 30, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-355111201, 1958, 9, 2, 30, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-355111200, 1958, 10, 3, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-336790801, 1959, 5, 5, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-336790800, 1959, 5, 5, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-323654401, 1959, 9, 3, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-323654400, 1959, 9, 3, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-323650801, 1959, 9, 3, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-323650800, 1959, 9, 3, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-305168401, 1960, 5, 0, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-305168400, 1960, 5, 0, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-292032001, 1960, 9, 5, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-292032000, 1960, 9, 5, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-292028401, 1960, 9, 5, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-292028400, 1960, 9, 5, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-273632401, 1961, 5, 1, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-273632400, 1961, 5, 1, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-260496001, 1961, 9, 6, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-260496000, 1961, 9, 6, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-260492401, 1961, 9, 6, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-260492400, 1961, 9, 6, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-242096401, 1962, 5, 2, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-242096400, 1962, 5, 2, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-228960001, 1962, 9, 0, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-228960000, 1962, 9, 0, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-228956401, 1962, 9, 0, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-228956400, 1962, 9, 0, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-210560401, 1963, 5, 3, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-210560400, 1963, 5, 3, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-197424001, 1963, 9, 1, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-197424000, 1963, 9, 1, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-197420401, 1963, 9, 1, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-197420400, 1963, 9, 1, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-178938001, 1964, 5, 5, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-178938000, 1964, 5, 5, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-165801601, 1964, 9, 3, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-165801600, 1964, 9, 3, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-165798001, 1964, 9, 3, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-165798000, 1964, 9, 3, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-147402001, 1965, 5, 6, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-147402000, 1965, 5, 6, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-134265601, 1965, 9, 4, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-134265600, 1965, 9, 4, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-134262001, 1965, 9, 4, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-134262000, 1965, 9, 4, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-115866001, 1966, 5, 0, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-115866000, 1966, 5, 0, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-102643201, 1966, 10, 6, 1, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-102643200, 1966, 10, 6, 1, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(-102639601, 1966, 10, 6, 1, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(-102639600, 1966, 10, 6, 1, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-84330001, 1967, 5, 1, 1, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-84330000, 1967, 5, 1, 1, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(-81313201, 1967, 6, 0, 4, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(-81313200, 1967, 6, 0, 4, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(-81309601, 1967, 6, 0, 4, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(-81309600, 1967, 6, 1, 5, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(142379999, 1974, 7, 6, 6, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(142380000, 1974, 7, 0, 7, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(150843599, 1974, 10, 6, 12, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(150843600, 1974, 10, 6, 12, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(150847199, 1974, 10, 6, 12, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(150847200, 1974, 10, 0, 13, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(167176799, 1975, 4, 6, 19, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(167176800, 1975, 4, 0, 20, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(178664399, 1975, 8, 6, 30, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(178664400, 1975, 8, 6, 30, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(178667999, 1975, 8, 6, 30, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(178668000, 1975, 8, 0, 31, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(482277599, 1985, 4, 6, 13, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(482277600, 1985, 4, 0, 14, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(495579599, 1985, 9, 6, 14, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(495579600, 1985, 9, 6, 14, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(495583199, 1985, 9, 6, 14, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(495583200, 1985, 9, 0, 15, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(516751199, 1986, 5, 6, 17, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(516751200, 1986, 5, 0, 18, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(526424399, 1986, 9, 6, 6, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(526424400, 1986, 9, 6, 6, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(526427999, 1986, 9, 6, 6, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(526428000, 1986, 9, 0, 7, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(545435999, 1987, 4, 2, 14, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(545436000, 1987, 4, 3, 15, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(558478799, 1987, 9, 6, 12, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(558478800, 1987, 9, 6, 12, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(558482399, 1987, 9, 6, 12, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(558482400, 1987, 9, 0, 13, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(576626399, 1988, 4, 6, 9, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(576626400, 1988, 4, 0, 10, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(589323599, 1988, 9, 6, 3, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(589323600, 1988, 9, 6, 3, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(589327199, 1988, 9, 6, 3, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(589327200, 1988, 9, 0, 4, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(609890399, 1989, 4, 6, 29, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(609890400, 1989, 4, 0, 30, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(620773199, 1989, 9, 6, 2, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(620773200, 1989, 9, 6, 2, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(620776799, 1989, 9, 6, 2, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(620776800, 1989, 9, 0, 3, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(638315999, 1990, 3, 6, 24, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(638316000, 1990, 3, 0, 25, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(651617999, 1990, 8, 6, 25, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(651618000, 1990, 8, 6, 25, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(651621599, 1990, 8, 6, 25, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(651621600, 1990, 8, 0, 26, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(669765599, 1991, 3, 6, 23, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(669765600, 1991, 3, 0, 24, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(683672399, 1991, 8, 6, 31, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(683672400, 1991, 8, 6, 31, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(683675999, 1991, 8, 6, 31, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(683676000, 1991, 9, 0, 1, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(701819999, 1992, 3, 6, 28, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(701820000, 1992, 3, 0, 29, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(715726799, 1992, 9, 6, 5, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(715726800, 1992, 9, 6, 5, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(715730399, 1992, 9, 6, 5, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(715730400, 1992, 9, 0, 6, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(733701599, 1993, 4, 4, 1, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(733701600, 1993, 4, 5, 2, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(747176399, 1993, 9, 6, 4, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(747176400, 1993, 9, 6, 4, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(747179999, 1993, 9, 6, 4, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(747180000, 1993, 9, 0, 5, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(765151199, 1994, 3, 4, 31, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(765151200, 1994, 4, 5, 1, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(778021199, 1994, 8, 6, 27, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(778021200, 1994, 8, 6, 27, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(778024799, 1994, 8, 6, 27, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(778024800, 1994, 8, 0, 28, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(796600799, 1995, 3, 4, 30, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(796600800, 1995, 3, 5, 31, 1, 0, 0, 1, 10800, "IDT", BIDI)
TESTVECTOR(810075599, 1995, 9, 6, 2, 23, 59, 59, 1, 10800, "IDT", BIDI)
TESTVECTOR(810075600, 1995, 9, 6, 2, 23, 0, 0, 0, 7200, "IST", HINT)
TESTVECTOR(810079199, 1995, 9, 6, 2, 23, 59, 59, 0, 7200, "IST", HINT)
TESTVECTOR(810079200, 1995, 9, 0, 3, 0, 0, 0, 0, 7200, "IST", BIDI)
TESTVECTOR(820447199, 1995, 12, 0, 31, 23, 59, 59, 0, 7200, "IST", BIDI)
TESTVECTOR(820447200, 1996, 1, 1, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(828655199, 1996, 4, 4, 4, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(828655200, 1996, 4, 5, 5, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(843170399, 1996, 9, 5, 20, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(843170400, 1996, 9, 5, 20, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(843173999, 1996, 9, 5, 20, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(843174000, 1996, 9, 5, 20, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(860104799, 1997, 4, 4, 3, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(860104800, 1997, 4, 5, 4, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(874619999, 1997, 9, 5, 19, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(874620000, 1997, 9, 5, 19, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(874623599, 1997, 9, 5, 19, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(874623600, 1997, 9, 5, 19, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(891554399, 1998, 4, 4, 2, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(891554400, 1998, 4, 5, 3, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(906069599, 1998, 9, 5, 18, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(906069600, 1998, 9, 5, 18, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(906073199, 1998, 9, 5, 18, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(906073200, 1998, 9, 5, 18, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(924213599, 1999, 4, 4, 15, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(924213600, 1999, 4, 5, 16, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(939934799, 1999, 10, 4, 14, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(939934800, 1999, 10, 4, 14, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(939938399, 1999, 10, 4, 14, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(939938400, 1999, 10, 5, 15, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(956267999, 2000, 4, 4, 20, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(956268000, 2000, 4, 5, 21, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(971989199, 2000, 10, 4, 19, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(971989200, 2000, 10, 4, 19, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(971992799, 2000, 10, 4, 19, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(971992800, 2000, 10, 5, 20, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(987717599, 2001, 4, 4, 19, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(987717600, 2001, 4, 5, 20, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1003438799, 2001, 10, 4, 18, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1003438800, 2001, 10, 4, 18, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1003442399, 2001, 10, 4, 18, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1003442400, 2001, 10, 5, 19, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1019167199, 2002, 4, 4, 18, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1019167200, 2002, 4, 5, 19, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1034888399, 2002, 10, 4, 17, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1034888400, 2002, 10, 4, 17, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1034891999, 2002, 10, 4, 17, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1034892000, 2002, 10, 5, 18, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1050616799, 2003, 4, 4, 17, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1050616800, 2003, 4, 5, 18, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1066337999, 2003, 10, 4, 16, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1066338000, 2003, 10, 4, 16, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1066341599, 2003, 10, 4, 16, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1066341600, 2003, 10, 5, 17, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1082066399, 2004, 4, 4, 15, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1082066400, 2004, 4, 5, 16, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1096581599, 2004, 10, 5, 1, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1096581600, 2004, 10, 5, 1, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1096585199, 2004, 10, 5, 1, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1096585200, 2004, 10, 5, 1, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1113515999, 2005, 4, 4, 14, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1113516000, 2005, 4, 5, 15, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1128380399, 2005, 10, 2, 4, 1, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1128380400, 2005, 10, 2, 4, 1, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1128383999, 2005, 10, 2, 4, 1, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1128384000, 2005, 10, 2, 4, 2, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1143842399, 2006, 3, 5, 31, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1143842400, 2006, 4, 6, 1, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1158872399, 2006, 9, 4, 21, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1158872400, 2006, 9, 4, 21, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1158875999, 2006, 9, 4, 21, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1158876000, 2006, 9, 5, 22, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1175378399, 2007, 3, 6, 31, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1175378400, 2007, 4, 0, 1, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1189637999, 2007, 9, 4, 13, 1, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1189638000, 2007, 9, 4, 13, 1, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1189641599, 2007, 9, 4, 13, 1, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1189641600, 2007, 9, 4, 13, 2, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1206655199, 2008, 3, 4, 27, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1206655200, 2008, 3, 5, 28, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1219957199, 2008, 8, 4, 28, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1219957200, 2008, 8, 4, 28, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1219960799, 2008, 8, 4, 28, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1219960800, 2008, 8, 5, 29, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1238104799, 2009, 3, 4, 26, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1238104800, 2009, 3, 5, 27, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1252015199, 2009, 9, 5, 4, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1252015200, 2009, 9, 5, 4, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1252018799, 2009, 9, 5, 4, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1252018800, 2009, 9, 5, 4, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1269640859, 2010, 3, 6, 27, 0, 0, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1269640860, 2010, 3, 6, 27, 1, 1, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1281473999, 2010, 8, 2, 10, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1281474000, 2010, 8, 2, 10, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1281477599, 2010, 8, 2, 10, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1281477600, 2010, 8, 3, 11, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1301608859, 2011, 4, 5, 1, 0, 0, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1301608860, 2011, 4, 5, 1, 1, 1, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1312145999, 2011, 7, 0, 31, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1312146000, 2011, 7, 0, 31, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1312149599, 2011, 7, 0, 31, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1312149600, 2011, 8, 1, 1, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1333058399, 2012, 3, 4, 29, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1333058400, 2012, 3, 5, 30, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1348178399, 2012, 9, 5, 21, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1348178400, 2012, 9, 5, 21, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1348181999, 2012, 9, 5, 21, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1348182000, 2012, 9, 5, 21, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1364507999, 2013, 3, 4, 28, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1364508000, 2013, 3, 5, 29, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1380229199, 2013, 9, 4, 26, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1380229200, 2013, 9, 4, 26, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1380232799, 2013, 9, 4, 26, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1380232800, 2013, 9, 5, 27, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1395957599, 2014, 3, 4, 27, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1395957600, 2014, 3, 5, 28, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1414097999, 2014, 10, 4, 23, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1414098000, 2014, 10, 4, 23, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1414101599, 2014, 10, 4, 23, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1414101600, 2014, 10, 5, 24, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1427493599, 2015, 3, 5, 27, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1427493600, 2015, 3, 6, 28, 1, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1445547599, 2015, 10, 4, 22, 23, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1445547600, 2015, 10, 4, 22, 23, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1445551199, 2015, 10, 4, 22, 23, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1445551200, 2015, 10, 5, 23, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1458946799, 2016, 3, 6, 26, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1458946800, 2016, 3, 6, 26, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1477691999, 2016, 10, 6, 29, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1477692000, 2016, 10, 6, 29, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1477695599, 2016, 10, 6, 29, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1477695600, 2016, 10, 6, 29, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1490396399, 2017, 3, 6, 25, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1490396400, 2017, 3, 6, 25, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1509141599, 2017, 10, 6, 28, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1509141600, 2017, 10, 6, 28, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1509145199, 2017, 10, 6, 28, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1509145200, 2017, 10, 6, 28, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1522450799, 2018, 3, 6, 31, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1522450800, 2018, 3, 6, 31, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1540591199, 2018, 10, 6, 27, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1540591200, 2018, 10, 6, 27, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1540594799, 2018, 10, 6, 27, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1540594800, 2018, 10, 6, 27, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1553900399, 2019, 3, 6, 30, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1553900400, 2019, 3, 6, 30, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1572040799, 2019, 10, 6, 26, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1572040800, 2019, 10, 6, 26, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1572044399, 2019, 10, 6, 26, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1572044400, 2019, 10, 6, 26, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1585349999, 2020, 3, 6, 28, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1585350000, 2020, 3, 6, 28, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1604095199, 2020, 10, 6, 31, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1604095200, 2020, 10, 6, 31, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1604098799, 2020, 10, 6, 31, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1604098800, 2020, 10, 6, 31, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1616799599, 2021, 3, 6, 27, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1616799600, 2021, 3, 6, 27, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1635544799, 2021, 10, 6, 30, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1635544800, 2021, 10, 6, 30, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1635548399, 2021, 10, 6, 30, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1635548400, 2021, 10, 6, 30, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1648249199, 2022, 3, 6, 26, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1648249200, 2022, 3, 6, 26, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1666994399, 2022, 10, 6, 29, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1666994400, 2022, 10, 6, 29, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1666997999, 2022, 10, 6, 29, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1666998000, 2022, 10, 6, 29, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1679698799, 2023, 3, 6, 25, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1679698800, 2023, 3, 6, 25, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1698443999, 2023, 10, 6, 28, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1698444000, 2023, 10, 6, 28, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1698447599, 2023, 10, 6, 28, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1698447600, 2023, 10, 6, 28, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1711753199, 2024, 3, 6, 30, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1711753200, 2024, 3, 6, 30, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1729893599, 2024, 10, 6, 26, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1729893600, 2024, 10, 6, 26, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1729897199, 2024, 10, 6, 26, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1729897200, 2024, 10, 6, 26, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1743202799, 2025, 3, 6, 29, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1743202800, 2025, 3, 6, 29, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1761343199, 2025, 10, 6, 25, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1761343200, 2025, 10, 6, 25, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1761346799, 2025, 10, 6, 25, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1761346800, 2025, 10, 6, 25, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1774652399, 2026, 3, 6, 28, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1774652400, 2026, 3, 6, 28, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1793397599, 2026, 10, 6, 31, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1793397600, 2026, 10, 6, 31, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1793401199, 2026, 10, 6, 31, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1793401200, 2026, 10, 6, 31, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1806101999, 2027, 3, 6, 27, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1806102000, 2027, 3, 6, 27, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1824847199, 2027, 10, 6, 30, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1824847200, 2027, 10, 6, 30, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1824850799, 2027, 10, 6, 30, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1824850800, 2027, 10, 6, 30, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1837551599, 2028, 3, 6, 25, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1837551600, 2028, 3, 6, 25, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1856296799, 2028, 10, 6, 28, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1856296800, 2028, 10, 6, 28, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1856300399, 2028, 10, 6, 28, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1856300400, 2028, 10, 6, 28, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1869605999, 2029, 3, 6, 31, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1869606000, 2029, 3, 6, 31, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1887746399, 2029, 10, 6, 27, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1887746400, 2029, 10, 6, 27, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1887749999, 2029, 10, 6, 27, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1887750000, 2029, 10, 6, 27, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1901055599, 2030, 3, 6, 30, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1901055600, 2030, 3, 6, 30, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1919195999, 2030, 10, 6, 26, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1919196000, 2030, 10, 6, 26, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1919199599, 2030, 10, 6, 26, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1919199600, 2030, 10, 6, 26, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1932505199, 2031, 3, 6, 29, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1932505200, 2031, 3, 6, 29, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1950645599, 2031, 10, 6, 25, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1950645600, 2031, 10, 6, 25, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1950649199, 2031, 10, 6, 25, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1950649200, 2031, 10, 6, 25, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1963954799, 2032, 3, 6, 27, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1963954800, 2032, 3, 6, 27, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1982699999, 2032, 10, 6, 30, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1982700000, 2032, 10, 6, 30, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1982703599, 2032, 10, 6, 30, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1982703600, 2032, 10, 6, 30, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1995404399, 2033, 3, 6, 26, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1995404400, 2033, 3, 6, 26, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2014149599, 2033, 10, 6, 29, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2014149600, 2033, 10, 6, 29, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2014153199, 2033, 10, 6, 29, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2014153200, 2033, 10, 6, 29, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2026853999, 2034, 3, 6, 25, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2026854000, 2034, 3, 6, 25, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2045599199, 2034, 10, 6, 28, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2045599200, 2034, 10, 6, 28, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2045602799, 2034, 10, 6, 28, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2045602800, 2034, 10, 6, 28, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2058908399, 2035, 3, 6, 31, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2058908400, 2035, 3, 6, 31, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2077048799, 2035, 10, 6, 27, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2077048800, 2035, 10, 6, 27, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2077052399, 2035, 10, 6, 27, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2077052400, 2035, 10, 6, 27, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2090357999, 2036, 3, 6, 29, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2090358000, 2036, 3, 6, 29, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2108498399, 2036, 10, 6, 25, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2108498400, 2036, 10, 6, 25, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2108501999, 2036, 10, 6, 25, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2108502000, 2036, 10, 6, 25, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2121807599, 2037, 3, 6, 28, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2121807600, 2037, 3, 6, 28, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2140552799, 2037, 10, 6, 31, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2140552800, 2037, 10, 6, 31, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2140556399, 2037, 10, 6, 31, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2140556400, 2037, 10, 6, 31, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2153257199, 2038, 3, 6, 27, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2153257200, 2038, 3, 6, 27, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2172002399, 2038, 10, 6, 30, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2172002400, 2038, 10, 6, 30, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2172005999, 2038, 10, 6, 30, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2172006000, 2038, 10, 6, 30, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2184706799, 2039, 3, 6, 26, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2184706800, 2039, 3, 6, 26, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2203451999, 2039, 10, 6, 29, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2203452000, 2039, 10, 6, 29, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2203455599, 2039, 10, 6, 29, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2203455600, 2039, 10, 6, 29, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2216761199, 2040, 3, 6, 31, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2216761200, 2040, 3, 6, 31, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2234901599, 2040, 10, 6, 27, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2234901600, 2040, 10, 6, 27, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2234905199, 2040, 10, 6, 27, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2234905200, 2040, 10, 6, 27, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2248210799, 2041, 3, 6, 30, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2248210800, 2041, 3, 6, 30, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2266351199, 2041, 10, 6, 26, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2266351200, 2041, 10, 6, 26, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2266354799, 2041, 10, 6, 26, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2266354800, 2041, 10, 6, 26, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2279660399, 2042, 3, 6, 29, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2279660400, 2042, 3, 6, 29, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2297800799, 2042, 10, 6, 25, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2297800800, 2042, 10, 6, 25, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2297804399, 2042, 10, 6, 25, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2297804400, 2042, 10, 6, 25, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2311109999, 2043, 3, 6, 28, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2311110000, 2043, 3, 6, 28, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2329855199, 2043, 10, 6, 31, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2329855200, 2043, 10, 6, 31, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2329858799, 2043, 10, 6, 31, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2329858800, 2043, 10, 6, 31, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2342559599, 2044, 3, 6, 26, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2342559600, 2044, 3, 6, 26, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2361304799, 2044, 10, 6, 29, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2361304800, 2044, 10, 6, 29, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2361308399, 2044, 10, 6, 29, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2361308400, 2044, 10, 6, 29, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2374009199, 2045, 3, 6, 25, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2374009200, 2045, 3, 6, 25, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2392754399, 2045, 10, 6, 28, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2392754400, 2045, 10, 6, 28, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2392757999, 2045, 10, 6, 28, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2392758000, 2045, 10, 6, 28, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2406063599, 2046, 3, 6, 31, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2406063600, 2046, 3, 6, 31, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2424203999, 2046, 10, 6, 27, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2424204000, 2046, 10, 6, 27, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2424207599, 2046, 10, 6, 27, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2424207600, 2046, 10, 6, 27, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2437513199, 2047, 3, 6, 30, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2437513200, 2047, 3, 6, 30, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2455653599, 2047, 10, 6, 26, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2455653600, 2047, 10, 6, 26, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2455657199, 2047, 10, 6, 26, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2455657200, 2047, 10, 6, 26, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2468962799, 2048, 3, 6, 28, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2468962800, 2048, 3, 6, 28, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2487707999, 2048, 10, 6, 31, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2487708000, 2048, 10, 6, 31, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2487711599, 2048, 10, 6, 31, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2487711600, 2048, 10, 6, 31, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(2500412399, 2049, 3, 6, 27, 0, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(2500412400, 2049, 3, 6, 27, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(2519157599, 2049, 10, 6, 30, 0, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(2519157600, 2049, 10, 6, 30, 0, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(2519161199, 2049, 10, 6, 30, 0, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(2519161200, 2049, 10, 6, 30, 1, 0, 0, 0, 7200, "EET", BIDI)
TEST_END

TEST_BEGIN(jujuy, "America/Argentina/Jujuy")
TESTVECTOR(-2372096329, 1894, 10, 2, 30, 23, 59, 59, 0, -15672, "LMT", BIDI)
TESTVECTOR(-2372096328, 1894, 10, 3, 31, 0, 4, 24, 0, -15408, "CMT", BIDI)
TESTVECTOR(-1567453393, 1920, 4, 5, 30, 23, 59, 59, 0, -15408, "CMT", BIDI)
TESTVECTOR(-1567453392, 1920, 5, 6, 1, 0, 16, 48, 0, -14400, "-04", BIDI)
TESTVECTOR(-1233432001, 1930, 11, 0, 30, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1233432000, 1930, 12, 1, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1222981201, 1931, 3, 2, 31, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1222981200, 1931, 3, 2, 31, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1222977601, 1931, 3, 2, 31, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1222977600, 1931, 4, 3, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-1205956801, 1931, 10, 3, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1205956800, 1931, 10, 4, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1194037201, 1932, 2, 1, 29, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1194037200, 1932, 2, 1, 29, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1194033601, 1932, 2, 1, 29, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1194033600, 1932, 3, 2, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-1172865601, 1932, 10, 1, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1172865600, 1932, 11, 2, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1162501201, 1933, 2, 2, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1162501200, 1933, 2, 2, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1162497601, 1933, 2, 2, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1162497600, 1933, 3, 3, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-1141329601, 1933, 10, 2, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1141329600, 1933, 11, 3, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1130965201, 1934, 2, 3, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1130965200, 1934, 2, 3, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1130961601, 1934, 2, 3, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1130961600, 1934, 3, 4, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-1109793601, 1934, 10, 3, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1109793600, 1934, 11, 4, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1099429201, 1935, 2, 4, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1099429200, 1935, 2, 4, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1099425601, 1935, 2, 4, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1099425600, 1935, 3, 5, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-1078257601, 1935, 10, 4, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1078257600, 1935, 11, 5, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1067806801, 1936, 2, 6, 29, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1067806800, 1936, 2, 6, 29, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1067803201, 1936, 2, 6, 29, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1067803200, 1936, 3, 0, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-1046635201, 1936, 10, 6, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1046635200, 1936, 11, 0, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1036270801, 1937, 2, 0, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1036270800, 1937, 2, 0, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1036267201, 1937, 2, 0, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1036267200, 1937, 3, 1, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-1015099201, 1937, 10, 0, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1015099200, 1937, 11, 1, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-1004734801, 1938, 2, 1, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-1004734800, 1938, 2, 1, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-1004731201, 1938, 2, 1, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-1004731200, 1938, 3, 2, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-983563201, 1938, 10, 1, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-983563200, 1938, 11, 2, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-973198801, 1939, 2, 2, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-973198800, 1939, 2, 2, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-973195201, 1939, 2, 2, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-973195200, 1939, 3, 3, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-952027201, 1939, 10, 2, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-952027200, 1939, 11, 3, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-941576401, 1940, 2, 4, 29, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-941576400, 1940, 2, 4, 29, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-941572801, 1940, 2, 4, 29, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-941572800, 1940, 3, 5, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-931032001, 1940, 6, 0, 30, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-931032000, 1940, 7, 1, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-900882001, 1941, 6, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-900882000, 1941, 6, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-900878401, 1941, 6, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-900878400, 1941, 6, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-890337601, 1941, 10, 2, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-890337600, 1941, 10, 3, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-833749201, 1943, 7, 6, 31, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-833749200, 1943, 7, 6, 31, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-833745601, 1943, 7, 6, 31, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-833745600, 1943, 8, 0, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-827265601, 1943, 10, 4, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-827265600, 1943, 10, 5, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-752274001, 1946, 2, 4, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-752274000, 1946, 2, 4, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-752270401, 1946, 2, 4, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-752270400, 1946, 3, 5, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-733780801, 1946, 9, 1, 30, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-733780800, 1946, 10, 2, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-197326801, 1963, 9, 1, 30, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-197326800, 1963, 9, 1, 30, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-197323201, 1963, 9, 1, 30, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-197323200, 1963, 10, 2, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-190843201, 1963, 12, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-190843200, 1963, 12, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-184194001, 1964, 2, 6, 29, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-184194000, 1964, 2, 6, 29, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-184190401, 1964, 2, 6, 29, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-184190400, 1964, 3, 0, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-164491201, 1964, 10, 3, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-164491200, 1964, 10, 4, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-152658001, 1965, 2, 0, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-152658000, 1965, 2, 0, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-152654401, 1965, 2, 0, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-152654400, 1965, 3, 1, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-132955201, 1965, 10, 4, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-132955200, 1965, 10, 5, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-121122001, 1966, 2, 1, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-121122000, 1966, 2, 1, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-121118401, 1966, 2, 1, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-121118400, 1966, 3, 2, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-101419201, 1966, 10, 5, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-101419200, 1966, 10, 6, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-86821201, 1967, 4, 6, 1, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-86821200, 1967, 4, 6, 1, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-86817601, 1967, 4, 6, 1, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-86817600, 1967, 4, 0, 2, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-71092801, 1967, 9, 6, 30, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-71092800, 1967, 10, 0, 1, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-54766801, 1968, 4, 6, 6, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-54766800, 1968, 4, 6, 6, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-54763201, 1968, 4, 6, 6, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-54763200, 1968, 4, 0, 7, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-39038401, 1968, 10, 6, 5, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-39038400, 1968, 10, 0, 6, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-23317201, 1969, 4, 6, 5, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-23317200, 1969, 4, 6, 5, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-23313601, 1969, 4, 6, 5, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-23313600, 1969, 4, 0, 6, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-7588801, 1969, 10, 6, 4, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-7588800, 1969, 10, 0, 5, 1, 0, 0, 0, -10800, "-03", BIDI)
TESTVECTOR(128141999, 1974, 1, 2, 22, 23, 59, 59, 0, -10800, "-03", BIDI)
TESTVECTOR(128142000, 1974, 1, 3, 23, 1, 0, 0, 1, -7200, "-02", BIDI)
TESTVECTOR(136605599, 1974, 4, 2, 30, 23, 59, 59, 1, -7200, "-02", BIDI)
TESTVECTOR(136605600, 1974, 4, 2, 30, 23, 0, 0, 0, -10800, "-03", HINT)
TESTVECTOR(136609199, 1974, 4, 2, 30, 23, 59, 59, 0, -10800, "-03", HINT)
TESTVECTOR(136609200, 1974, 5, 3, 1, 0, 0, 0, 0, -10800, "-03", BIDI)
TESTVECTOR(596948399, 1988, 11, 3, 30, 23, 59, 59, 0, -10800, "-03", BIDI)
TESTVECTOR(596948400, 1988, 12, 4, 1, 1, 0, 0, 1, -7200, "-02", BIDI)
TESTVECTOR(605066399, 1989, 3, 6, 4, 23, 59, 59, 1, -7200, "-02", BIDI)
TESTVECTOR(605066400, 1989, 3, 6, 4, 23, 0, 0, 0, -10800, "-03", HINT)
TESTVECTOR(605069999, 1989, 3, 6, 4, 23, 59, 59, 0, -10800, "-03", HINT)
TESTVECTOR(605070000, 1989, 3, 0, 5, 0, 0, 0, 0, -10800, "-03", BIDI)
TESTVECTOR(624423599, 1989, 10, 6, 14, 23, 59, 59, 0, -10800, "-03", BIDI)
TESTVECTOR(624423600, 1989, 10, 0, 15, 1, 0, 0, 1, -7200, "-02", BIDI)
TESTVECTOR(636515999, 1990, 3, 6, 3, 23, 59, 59, 1, -7200, "-02", BIDI)
TESTVECTOR(636516000, 1990, 3, 6, 3, 22, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(636523199, 1990, 3, 6, 3, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(636523200, 1990, 3, 0, 4, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(657086399, 1990, 10, 6, 27, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(657086400, 1990, 10, 0, 28, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(669178799, 1991, 3, 6, 16, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(669178800, 1991, 3, 6, 16, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(669182399, 1991, 3, 6, 16, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(669182400, 1991, 3, 0, 17, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(686721599, 1991, 10, 6, 5, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(686721600, 1991, 10, 0, 6, 2, 0, 0, 1, -7200, "-02", BIDI)
TESTVECTOR(699415199, 1992, 2, 6, 29, 23, 59, 59, 1, -7200, "-02", BIDI)
TESTVECTOR(699415200, 1992, 2, 6, 29, 23, 0, 0, 0, -10800, "-03", HINT)
TESTVECTOR(699418799, 1992, 2, 6, 29, 23, 59, 59, 0, -10800, "-03", HINT)
TESTVECTOR(699418800, 1992, 3, 0, 1, 0, 0, 0, 0, -10800, "-03", BIDI)
TESTVECTOR(719377199, 1992, 10, 6, 17, 23, 59, 59, 0, -10800, "-03", BIDI)
TESTVECTOR(719377200, 1992, 10, 0, 18, 1, 0, 0, 1, -7200, "-02", BIDI)
TESTVECTOR(731469599, 1993, 3, 6, 6, 23, 59, 59, 1, -7200, "-02", BIDI)
TESTVECTOR(731469600, 1993, 3, 6, 6, 23, 0, 0, 0, -10800, "-03", HINT)
TESTVECTOR(731473199, 1993, 3, 6, 6, 23, 59, 59, 0, -10800, "-03", HINT)
TESTVECTOR(731473200, 1993, 3, 0, 7, 0, 0, 0, 0, -10800, "-03", BIDI)
TESTVECTOR(938919599, 1999, 10, 6, 2, 23, 59, 59, 0, -10800, "-03", BIDI)
TESTVECTOR(938919600, 1999, 10, 0, 3, 0, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(952052399, 2000, 3, 4, 2, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(952052400, 2000, 3, 5, 3, 0, 0, 0, 0, -10800, "-03", BIDI)
TESTVECTOR(1198983599, 2007, 12, 6, 29, 23, 59, 59, 0, -10800, "-03", BIDI)
TESTVECTOR(1198983600, 2007, 12, 0, 30, 1, 0, 0, 1, -7200, "-02", BIDI)
TESTVECTOR(1205632799, 2008, 3, 6, 15, 23, 59, 59, 1, -7200, "-02", BIDI)
TESTVECTOR(1205632800, 2008, 3, 6, 15, 23, 0, 0, 0, -10800, "-03", HINT)
TESTVECTOR(1205636399, 2008, 3, 6, 15, 23, 59, 59, 0, -10800, "-03", HINT)
TESTVECTOR(1205636400, 2008, 3, 0, 16, 0, 0, 0, 0, -10800, "-03", BIDI)
TEST_END

TEST_BEGIN(kuala_lumpur, "Asia/Kuala_Lumpur")
TESTVECTOR(-2177477207, 1900, 12, 1, 31, 23, 59, 59, 0, 24406, "LMT", BIDI)
TESTVECTOR(-2177477206, 1901, 1, 2, 1, 0, 8, 39, 0, 24925, "SMT", BIDI)
TESTVECTOR(-2038200926, 1905, 5, 3, 31, 23, 59, 59, 0, 24925, "SMT", BIDI)
TESTVECTOR(-2038200925, 1905, 6, 4, 1, 0, 4, 35, 0, 25200, "+07", BIDI)
TESTVECTOR(-1167634801, 1932, 12, 6, 31, 23, 59, 59, 0, 25200, "+07", BIDI)
TESTVECTOR(-1167634800, 1933, 1, 0, 1, 0, 20, 0, 1, 26400, "+0720", BIDI)
TESTVECTOR(-1073028001, 1935, 12, 2, 31, 23, 59, 59, 1, 26400, "+0720", BIDI)
TESTVECTOR(-1073028000, 1936, 1, 3, 1, 0, 0, 0, 0, 26400, "+0720", BIDI)
TESTVECTOR(-894180001, 1941, 8, 0, 31, 23, 59, 59, 0, 26400, "+0720", BIDI)
TESTVECTOR(-894180000, 1941, 9, 1, 1, 0, 10, 0, 0, 27000, "+0730", BIDI)
TESTVECTOR(-879665401, 1942, 2, 0, 15, 23, 59, 59, 0, 27000, "+0730", BIDI)
TESTVECTOR(-879665400, 1942, 2, 1, 16, 1, 30, 0, 0, 32400, "+09", BIDI)
TESTVECTOR(-767005201, 1945, 9, 2, 11, 23, 59, 59, 0, 32400, "+09", BIDI)
TESTVECTOR(-767005200, 1945, 9, 2, 11, 22, 30, 0, 0, 27000, "+0730", UNI)
TESTVECTOR(-766999801, 1945, 9, 2, 11, 23, 59, 59, 0, 27000, "+0730", UNI)
TESTVECTOR(-766999800, 1945, 9, 3, 12, 0, 0, 0, 0, 27000, "+0730", BIDI)
TESTVECTOR(378664199, 1981, 12, 4, 31, 23, 59, 59, 0, 27000, "+0730", BIDI)
TESTVECTOR(378664200, 1982, 1, 5, 1, 0, 30, 0, 0, 28800, "+08", BIDI)
TEST_END

TEST_BEGIN(london, "Europe/London")
TESTVECTOR(-3852662326, 1847, 11, 2, 30, 23, 59, 59, 0, -75, "LMT", BIDI)
TESTVECTOR(-3852662325, 1847, 12, 3, 1, 0, 1, 15, 0, 0, "GMT", BIDI)
TESTVECTOR(-1691964001, 1916, 5, 0, 21, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1691964000, 1916, 5, 0, 21, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1680472801, 1916, 10, 0, 1, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1680472800, 1916, 10, 0, 1, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1680469201, 1916, 10, 0, 1, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1680469200, 1916, 10, 0, 1, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1664143201, 1917, 4, 0, 8, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1664143200, 1917, 4, 0, 8, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1650146401, 1917, 9, 1, 17, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1650146400, 1917, 9, 1, 17, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1650142801, 1917, 9, 1, 17, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1650142800, 1917, 9, 1, 17, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1633903201, 1918, 3, 0, 24, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1633903200, 1918, 3, 0, 24, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1617487201, 1918, 9, 1, 30, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1617487200, 1918, 9, 1, 30, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1617483601, 1918, 9, 1, 30, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1617483600, 1918, 9, 1, 30, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1601848801, 1919, 3, 0, 30, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1601848800, 1919, 3, 0, 30, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1586037601, 1919, 9, 1, 29, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1586037600, 1919, 9, 1, 29, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1586034001, 1919, 9, 1, 29, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1586034000, 1919, 9, 1, 29, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1570399201, 1920, 3, 0, 28, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1570399200, 1920, 3, 0, 28, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1552168801, 1920, 10, 1, 25, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1552168800, 1920, 10, 1, 25, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1552165201, 1920, 10, 1, 25, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1552165200, 1920, 10, 1, 25, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1538344801, 1921, 4, 0, 3, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1538344800, 1921, 4, 0, 3, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1522533601, 1921, 10, 1, 3, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1522533600, 1921, 10, 1, 3, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1522530001, 1921, 10, 1, 3, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1522530000, 1921, 10, 1, 3, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1507500001, 1922, 3, 0, 26, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1507500000, 1922, 3, 0, 26, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1490565601, 1922, 10, 0, 8, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1490565600, 1922, 10, 0, 8, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1490562001, 1922, 10, 0, 8, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1490562000, 1922, 10, 0, 8, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1473631201, 1923, 4, 0, 22, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1473631200, 1923, 4, 0, 22, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1460930401, 1923, 9, 0, 16, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1460930400, 1923, 9, 0, 16, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1460926801, 1923, 9, 0, 16, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1460926800, 1923, 9, 0, 16, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1442786401, 1924, 4, 0, 13, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1442786400, 1924, 4, 0, 13, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1428876001, 1924, 9, 0, 21, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1428876000, 1924, 9, 0, 21, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1428872401, 1924, 9, 0, 21, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1428872400, 1924, 9, 0, 21, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1410732001, 1925, 4, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1410732000, 1925, 4, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1396216801, 1925, 10, 0, 4, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1396216800, 1925, 10, 0, 4, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1396213201, 1925, 10, 0, 4, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1396213200, 1925, 10, 0, 4, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1379282401, 1926, 4, 0, 18, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1379282400, 1926, 4, 0, 18, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1364767201, 1926, 10, 0, 3, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1364767200, 1926, 10, 0, 3, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1364763601, 1926, 10, 0, 3, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1364763600, 1926, 10, 0, 3, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1348437601, 1927, 4, 0, 10, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1348437600, 1927, 4, 0, 10, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1333317601, 1927, 10, 0, 2, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1333317600, 1927, 10, 0, 2, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1333314001, 1927, 10, 0, 2, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1333314000, 1927, 10, 0, 2, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1315778401, 1928, 4, 0, 22, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1315778400, 1928, 4, 0, 22, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1301263201, 1928, 10, 0, 7, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1301263200, 1928, 10, 0, 7, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1301259601, 1928, 10, 0, 7, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1301259600, 1928, 10, 0, 7, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1284328801, 1929, 4, 0, 21, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1284328800, 1929, 4, 0, 21, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1269813601, 1929, 10, 0, 6, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1269813600, 1929, 10, 0, 6, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1269810001, 1929, 10, 0, 6, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1269810000, 1929, 10, 0, 6, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1253484001, 1930, 4, 0, 13, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1253484000, 1930, 4, 0, 13, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1238364001, 1930, 10, 0, 5, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1238364000, 1930, 10, 0, 5, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1238360401, 1930, 10, 0, 5, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1238360400, 1930, 10, 0, 5, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1221429601, 1931, 4, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1221429600, 1931, 4, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1206914401, 1931, 10, 0, 4, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1206914400, 1931, 10, 0, 4, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1206910801, 1931, 10, 0, 4, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1206910800, 1931, 10, 0, 4, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1189980001, 1932, 4, 0, 17, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1189980000, 1932, 4, 0, 17, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1175464801, 1932, 10, 0, 2, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1175464800, 1932, 10, 0, 2, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1175461201, 1932, 10, 0, 2, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1175461200, 1932, 10, 0, 2, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1159135201, 1933, 4, 0, 9, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1159135200, 1933, 4, 0, 9, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1143410401, 1933, 10, 0, 8, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1143410400, 1933, 10, 0, 8, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1143406801, 1933, 10, 0, 8, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1143406800, 1933, 10, 0, 8, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1126476001, 1934, 4, 0, 22, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1126476000, 1934, 4, 0, 22, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1111960801, 1934, 10, 0, 7, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1111960800, 1934, 10, 0, 7, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1111957201, 1934, 10, 0, 7, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1111957200, 1934, 10, 0, 7, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1095631201, 1935, 4, 0, 14, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1095631200, 1935, 4, 0, 14, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1080511201, 1935, 10, 0, 6, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1080511200, 1935, 10, 0, 6, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1080507601, 1935, 10, 0, 6, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1080507600, 1935, 10, 0, 6, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1063576801, 1936, 4, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1063576800, 1936, 4, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1049061601, 1936, 10, 0, 4, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1049061600, 1936, 10, 0, 4, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1049058001, 1936, 10, 0, 4, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1049058000, 1936, 10, 0, 4, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1032127201, 1937, 4, 0, 18, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1032127200, 1937, 4, 0, 18, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-1017612001, 1937, 10, 0, 3, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-1017612000, 1937, 10, 0, 3, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-1017608401, 1937, 10, 0, 3, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-1017608400, 1937, 10, 0, 3, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-1001282401, 1938, 4, 0, 10, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-1001282400, 1938, 4, 0, 10, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-986162401, 1938, 10, 0, 2, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-986162400, 1938, 10, 0, 2, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-986158801, 1938, 10, 0, 2, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-986158800, 1938, 10, 0, 2, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-969228001, 1939, 4, 0, 16, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-969228000, 1939, 4, 0, 16, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-950479201, 1939, 11, 0, 19, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-950479200, 1939, 11, 0, 19, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-950475601, 1939, 11, 0, 19, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-950475600, 1939, 11, 0, 19, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-942012001, 1940, 2, 0, 25, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-942012000, 1940, 2, 0, 25, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-904518001, 1941, 5, 0, 4, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-904518000, 1941, 5, 0, 4, 3, 0, 0, 1, 7200, "BDST", BIDI)
TESTVECTOR(-896050801, 1941, 8, 0, 10, 2, 59, 59, 1, 7200, "BDST", BIDI)
TESTVECTOR(-896050800, 1941, 8, 0, 10, 2, 0, 0, 1, 3600, "BST", UNI)
TESTVECTOR(-896047201, 1941, 8, 0, 10, 2, 59, 59, 1, 3600, "BST", UNI)
TESTVECTOR(-896047200, 1941, 8, 0, 10, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-875487601, 1942, 4, 0, 5, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-875487600, 1942, 4, 0, 5, 3, 0, 0, 1, 7200, "BDST", BIDI)
TESTVECTOR(-864601201, 1942, 8, 0, 9, 2, 59, 59, 1, 7200, "BDST", BIDI)
TESTVECTOR(-864601200, 1942, 8, 0, 9, 2, 0, 0, 1, 3600, "BST", UNI)
TESTVECTOR(-864597601, 1942, 8, 0, 9, 2, 59, 59, 1, 3600, "BST", UNI)
TESTVECTOR(-864597600, 1942, 8, 0, 9, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-844038001, 1943, 4, 0, 4, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-844038000, 1943, 4, 0, 4, 3, 0, 0, 1, 7200, "BDST", BIDI)
TESTVECTOR(-832546801, 1943, 8, 0, 15, 2, 59, 59, 1, 7200, "BDST", BIDI)
TESTVECTOR(-832546800, 1943, 8, 0, 15, 2, 0, 0, 1, 3600, "BST", UNI)
TESTVECTOR(-832543201, 1943, 8, 0, 15, 2, 59, 59, 1, 3600, "BST", UNI)
TESTVECTOR(-832543200, 1943, 8, 0, 15, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-812588401, 1944, 4, 0, 2, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-812588400, 1944, 4, 0, 2, 3, 0, 0, 1, 7200, "BDST", BIDI)
TESTVECTOR(-798073201, 1944, 9, 0, 17, 2, 59, 59, 1, 7200, "BDST", BIDI)
TESTVECTOR(-798073200, 1944, 9, 0, 17, 2, 0, 0, 1, 3600, "BST", UNI)
TESTVECTOR(-798069601, 1944, 9, 0, 17, 2, 59, 59, 1, 3600, "BST", UNI)
TESTVECTOR(-798069600, 1944, 9, 0, 17, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-781052401, 1945, 4, 1, 2, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-781052400, 1945, 4, 1, 2, 3, 0, 0, 1, 7200, "BDST", BIDI)
TESTVECTOR(-772066801, 1945, 7, 0, 15, 2, 59, 59, 1, 7200, "BDST", BIDI)
TESTVECTOR(-772066800, 1945, 7, 0, 15, 2, 0, 0, 1, 3600, "BST", UNI)
TESTVECTOR(-772063201, 1945, 7, 0, 15, 2, 59, 59, 1, 3600, "BST", UNI)
TESTVECTOR(-772063200, 1945, 7, 0, 15, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-764805601, 1945, 10, 0, 7, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-764805600, 1945, 10, 0, 7, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-764802001, 1945, 10, 0, 7, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-764802000, 1945, 10, 0, 7, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-748476001, 1946, 4, 0, 14, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-748476000, 1946, 4, 0, 14, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-733356001, 1946, 10, 0, 6, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-733356000, 1946, 10, 0, 6, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-733352401, 1946, 10, 0, 6, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-733352400, 1946, 10, 0, 6, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-719445601, 1947, 3, 0, 16, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-719445600, 1947, 3, 0, 16, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-717030001, 1947, 4, 0, 13, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-717030000, 1947, 4, 0, 13, 3, 0, 0, 1, 7200, "BDST", BIDI)
TESTVECTOR(-706748401, 1947, 8, 0, 10, 2, 59, 59, 1, 7200, "BDST", BIDI)
TESTVECTOR(-706748400, 1947, 8, 0, 10, 2, 0, 0, 1, 3600, "BST", UNI)
TESTVECTOR(-706744801, 1947, 8, 0, 10, 2, 59, 59, 1, 3600, "BST", UNI)
TESTVECTOR(-706744800, 1947, 8, 0, 10, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-699487201, 1947, 11, 0, 2, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-699487200, 1947, 11, 0, 2, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-699483601, 1947, 11, 0, 2, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-699483600, 1947, 11, 0, 2, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-687996001, 1948, 3, 0, 14, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-687996000, 1948, 3, 0, 14, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-668037601, 1948, 10, 0, 31, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-668037600, 1948, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-668034001, 1948, 10, 0, 31, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-668034000, 1948, 10, 0, 31, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-654732001, 1949, 4, 0, 3, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-654732000, 1949, 4, 0, 3, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-636588001, 1949, 10, 0, 30, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-636588000, 1949, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-636584401, 1949, 10, 0, 30, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-636584400, 1949, 10, 0, 30, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-622072801, 1950, 4, 0, 16, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-622072800, 1950, 4, 0, 16, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-605743201, 1950, 10, 0, 22, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-605743200, 1950, 10, 0, 22, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-605739601, 1950, 10, 0, 22, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-605739600, 1950, 10, 0, 22, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-590623201, 1951, 4, 0, 15, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-590623200, 1951, 4, 0, 15, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-574293601, 1951, 10, 0, 21, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-574293600, 1951, 10, 0, 21, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-574290001, 1951, 10, 0, 21, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-574290000, 1951, 10, 0, 21, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-558568801, 1952, 4, 0, 20, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-558568800, 1952, 4, 0, 20, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-542239201, 1952, 10, 0, 26, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-542239200, 1952, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-542235601, 1952, 10, 0, 26, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-542235600, 1952, 10, 0, 26, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-527119201, 1953, 4, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-527119200, 1953, 4, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-512604001, 1953, 10, 0, 4, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-512604000, 1953, 10, 0, 4, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-512600401, 1953, 10, 0, 4, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-512600400, 1953, 10, 0, 4, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-496274401, 1954, 4, 0, 11, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-496274400, 1954, 4, 0, 11, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-481154401, 1954, 10, 0, 3, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-481154400, 1954, 10, 0, 3, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-481150801, 1954, 10, 0, 3, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-481150800, 1954, 10, 0, 3, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-464220001, 1955, 4, 0, 17, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-464220000, 1955, 4, 0, 17, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-449704801, 1955, 10, 0, 2, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-449704800, 1955, 10, 0, 2, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-449701201, 1955, 10, 0, 2, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-449701200, 1955, 10, 0, 2, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-432165601, 1956, 4, 0, 22, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-432165600, 1956, 4, 0, 22, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-417650401, 1956, 10, 0, 7, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-417650400, 1956, 10, 0, 7, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-417646801, 1956, 10, 0, 7, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-417646800, 1956, 10, 0, 7, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-401320801, 1957, 4, 0, 14, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-401320800, 1957, 4, 0, 14, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-386200801, 1957, 10, 0, 6, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-386200800, 1957, 10, 0, 6, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-386197201, 1957, 10, 0, 6, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-386197200, 1957, 10, 0, 6, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-369266401, 1958, 4, 0, 20, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-369266400, 1958, 4, 0, 20, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-354751201, 1958, 10, 0, 5, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-354751200, 1958, 10, 0, 5, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-354747601, 1958, 10, 0, 5, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-354747600, 1958, 10, 0, 5, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-337816801, 1959, 4, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-337816800, 1959, 4, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-323301601, 1959, 10, 0, 4, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-323301600, 1959, 10, 0, 4, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-323298001, 1959, 10, 0, 4, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-323298000, 1959, 10, 0, 4, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-306972001, 1960, 4, 0, 10, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-306972000, 1960, 4, 0, 10, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-291852001, 1960, 10, 0, 2, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-291852000, 1960, 10, 0, 2, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-291848401, 1960, 10, 0, 2, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-291848400, 1960, 10, 0, 2, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-276732001, 1961, 3, 0, 26, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-276732000, 1961, 3, 0, 26, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-257983201, 1961, 10, 0, 29, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-257983200, 1961, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-257979601, 1961, 10, 0, 29, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-257979600, 1961, 10, 0, 29, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-245282401, 1962, 3, 0, 25, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-245282400, 1962, 3, 0, 25, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-226533601, 1962, 10, 0, 28, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-226533600, 1962, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-226530001, 1962, 10, 0, 28, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-226530000, 1962, 10, 0, 28, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-213228001, 1963, 3, 0, 31, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-213228000, 1963, 3, 0, 31, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-195084001, 1963, 10, 0, 27, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-195084000, 1963, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-195080401, 1963, 10, 0, 27, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-195080400, 1963, 10, 0, 27, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-182383201, 1964, 3, 0, 22, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-182383200, 1964, 3, 0, 22, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-163634401, 1964, 10, 0, 25, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-163634400, 1964, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-163630801, 1964, 10, 0, 25, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-163630800, 1964, 10, 0, 25, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-150933601, 1965, 3, 0, 21, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-150933600, 1965, 3, 0, 21, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-132184801, 1965, 10, 0, 24, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-132184800, 1965, 10, 0, 24, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-132181201, 1965, 10, 0, 24, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-132181200, 1965, 10, 0, 24, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-119484001, 1966, 3, 0, 20, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-119484000, 1966, 3, 0, 20, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-100735201, 1966, 10, 0, 23, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-100735200, 1966, 10, 0, 23, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-100731601, 1966, 10, 0, 23, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-100731600, 1966, 10, 0, 23, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-88034401, 1967, 3, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-88034400, 1967, 3, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-68680801, 1967, 10, 0, 29, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-68680800, 1967, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(-68677201, 1967, 10, 0, 29, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(-68677200, 1967, 10, 0, 29, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(-59004001, 1968, 2, 0, 18, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(-59004000, 1968, 2, 0, 18, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(-37242001, 1968, 10, 6, 26, 23, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(-37242000, 1968, 10, 0, 27, 0, 0, 0, 0, 3600, "BST", BIDI)
TESTVECTOR(57722399, 1971, 10, 0, 31, 2, 59, 59, 0, 3600, "BST", BIDI)
TESTVECTOR(57722400, 1971, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", UNI)
TESTVECTOR(57725999, 1971, 10, 0, 31, 2, 59, 59, 0, 0, "GMT", UNI)
TESTVECTOR(57726000, 1971, 10, 0, 31, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(69818399, 1972, 3, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(69818400, 1972, 3, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(89171999, 1972, 10, 0, 29, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(89172000, 1972, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(89175599, 1972, 10, 0, 29, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(89175600, 1972, 10, 0, 29, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(101267999, 1973, 3, 0, 18, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(101268000, 1973, 3, 0, 18, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(120621599, 1973, 10, 0, 28, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(120621600, 1973, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(120625199, 1973, 10, 0, 28, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(120625200, 1973, 10, 0, 28, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(132717599, 1974, 3, 0, 17, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(132717600, 1974, 3, 0, 17, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(152071199, 1974, 10, 0, 27, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(152071200, 1974, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(152074799, 1974, 10, 0, 27, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(152074800, 1974, 10, 0, 27, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(164167199, 1975, 3, 0, 16, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(164167200, 1975, 3, 0, 16, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(183520799, 1975, 10, 0, 26, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(183520800, 1975, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(183524399, 1975, 10, 0, 26, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(183524400, 1975, 10, 0, 26, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(196221599, 1976, 3, 0, 21, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(196221600, 1976, 3, 0, 21, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(214970399, 1976, 10, 0, 24, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(214970400, 1976, 10, 0, 24, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(214973999, 1976, 10, 0, 24, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(214974000, 1976, 10, 0, 24, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(227671199, 1977, 3, 0, 20, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(227671200, 1977, 3, 0, 20, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(246419999, 1977, 10, 0, 23, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(246420000, 1977, 10, 0, 23, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(246423599, 1977, 10, 0, 23, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(246423600, 1977, 10, 0, 23, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(259120799, 1978, 3, 0, 19, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(259120800, 1978, 3, 0, 19, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(278474399, 1978, 10, 0, 29, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(278474400, 1978, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(278477999, 1978, 10, 0, 29, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(278478000, 1978, 10, 0, 29, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(290570399, 1979, 3, 0, 18, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(290570400, 1979, 3, 0, 18, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(309923999, 1979, 10, 0, 28, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(309924000, 1979, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(309927599, 1979, 10, 0, 28, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(309927600, 1979, 10, 0, 28, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(322019999, 1980, 3, 0, 16, 1, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(322020000, 1980, 3, 0, 16, 3, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(341373599, 1980, 10, 0, 26, 2, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(341373600, 1980, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(341377199, 1980, 10, 0, 26, 2, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(341377200, 1980, 10, 0, 26, 3, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(354675599, 1981, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(354675600, 1981, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(372819599, 1981, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(372819600, 1981, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(372823199, 1981, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(372823200, 1981, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(386125199, 1982, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(386125200, 1982, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(404269199, 1982, 10, 0, 24, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(404269200, 1982, 10, 0, 24, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(404272799, 1982, 10, 0, 24, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(404272800, 1982, 10, 0, 24, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(417574799, 1983, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(417574800, 1983, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(435718799, 1983, 10, 0, 23, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(435718800, 1983, 10, 0, 23, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(435722399, 1983, 10, 0, 23, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(435722400, 1983, 10, 0, 23, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(449024399, 1984, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(449024400, 1984, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(467773199, 1984, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(467773200, 1984, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(467776799, 1984, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(467776800, 1984, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(481078799, 1985, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(481078800, 1985, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(499222799, 1985, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(499222800, 1985, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(499226399, 1985, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(499226400, 1985, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(512528399, 1986, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(512528400, 1986, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(530672399, 1986, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(530672400, 1986, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(530675999, 1986, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(530676000, 1986, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(543977999, 1987, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(543978000, 1987, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(562121999, 1987, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(562122000, 1987, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(562125599, 1987, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(562125600, 1987, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(575427599, 1988, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(575427600, 1988, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(593571599, 1988, 10, 0, 23, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(593571600, 1988, 10, 0, 23, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(593575199, 1988, 10, 0, 23, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(593575200, 1988, 10, 0, 23, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(606877199, 1989, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(606877200, 1989, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(625625999, 1989, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(625626000, 1989, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(625629599, 1989, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(625629600, 1989, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(638326799, 1990, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(638326800, 1990, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(657075599, 1990, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(657075600, 1990, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(657079199, 1990, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(657079200, 1990, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(670381199, 1991, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(670381200, 1991, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(688525199, 1991, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(688525200, 1991, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(688528799, 1991, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(688528800, 1991, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(701830799, 1992, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(701830800, 1992, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(719974799, 1992, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(719974800, 1992, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(719978399, 1992, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(719978400, 1992, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(733280399, 1993, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(733280400, 1993, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(751424399, 1993, 10, 0, 24, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(751424400, 1993, 10, 0, 24, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(751427999, 1993, 10, 0, 24, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(751428000, 1993, 10, 0, 24, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(764729999, 1994, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(764730000, 1994, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(782873999, 1994, 10, 0, 23, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(782874000, 1994, 10, 0, 23, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(782877599, 1994, 10, 0, 23, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(782877600, 1994, 10, 0, 23, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(796179599, 1995, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(796179600, 1995, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(814323599, 1995, 10, 0, 22, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(814323600, 1995, 10, 0, 22, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(814327199, 1995, 10, 0, 22, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(814327200, 1995, 10, 0, 22, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(828233999, 1996, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(828234000, 1996, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(846377999, 1996, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(846378000, 1996, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(846381599, 1996, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(846381600, 1996, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(859683599, 1997, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(859683600, 1997, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(877827599, 1997, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(877827600, 1997, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(877831199, 1997, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(877831200, 1997, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(891133199, 1998, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(891133200, 1998, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(909277199, 1998, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(909277200, 1998, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(909280799, 1998, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(909280800, 1998, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(922582799, 1999, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(922582800, 1999, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(941331599, 1999, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(941331600, 1999, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(941335199, 1999, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(941335200, 1999, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(954032399, 2000, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(954032400, 2000, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(972781199, 2000, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(972781200, 2000, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(972784799, 2000, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(972784800, 2000, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(985481999, 2001, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(985482000, 2001, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1004230799, 2001, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1004230800, 2001, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1004234399, 2001, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1004234400, 2001, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1017536399, 2002, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1017536400, 2002, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1035680399, 2002, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1035680400, 2002, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1035683999, 2002, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1035684000, 2002, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1048985999, 2003, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1048986000, 2003, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1067129999, 2003, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1067130000, 2003, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1067133599, 2003, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1067133600, 2003, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1080435599, 2004, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1080435600, 2004, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1099184399, 2004, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1099184400, 2004, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1099187999, 2004, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1099188000, 2004, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1111885199, 2005, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1111885200, 2005, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1130633999, 2005, 10, 0, 30, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1130634000, 2005, 10, 0, 30, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1130637599, 2005, 10, 0, 30, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1130637600, 2005, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1143334799, 2006, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1143334800, 2006, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1162083599, 2006, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1162083600, 2006, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1162087199, 2006, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1162087200, 2006, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1174784399, 2007, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1174784400, 2007, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1193533199, 2007, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1193533200, 2007, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1193536799, 2007, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1193536800, 2007, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1206838799, 2008, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1206838800, 2008, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1224982799, 2008, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1224982800, 2008, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1224986399, 2008, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1224986400, 2008, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1238288399, 2009, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1238288400, 2009, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1256432399, 2009, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1256432400, 2009, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1256435999, 2009, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1256436000, 2009, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1269737999, 2010, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1269738000, 2010, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1288486799, 2010, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1288486800, 2010, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1288490399, 2010, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1288490400, 2010, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1301187599, 2011, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1301187600, 2011, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1319936399, 2011, 10, 0, 30, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1319936400, 2011, 10, 0, 30, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1319939999, 2011, 10, 0, 30, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1319940000, 2011, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1332637199, 2012, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1332637200, 2012, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1351385999, 2012, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1351386000, 2012, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1351389599, 2012, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1351389600, 2012, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1364691599, 2013, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1364691600, 2013, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1382835599, 2013, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1382835600, 2013, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1382839199, 2013, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1382839200, 2013, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1396141199, 2014, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1396141200, 2014, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1414285199, 2014, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1414285200, 2014, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1414288799, 2014, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1414288800, 2014, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1427590799, 2015, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1427590800, 2015, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1445734799, 2015, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1445734800, 2015, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1445738399, 2015, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1445738400, 2015, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1459040399, 2016, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1459040400, 2016, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1477789199, 2016, 10, 0, 30, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1477789200, 2016, 10, 0, 30, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1477792799, 2016, 10, 0, 30, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1477792800, 2016, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1490489999, 2017, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1490490000, 2017, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1509238799, 2017, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1509238800, 2017, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1509242399, 2017, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1509242400, 2017, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1521939599, 2018, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1521939600, 2018, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1540688399, 2018, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1540688400, 2018, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1540691999, 2018, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1540692000, 2018, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1553993999, 2019, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1553994000, 2019, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1572137999, 2019, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1572138000, 2019, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1572141599, 2019, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1572141600, 2019, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1585443599, 2020, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1585443600, 2020, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1603587599, 2020, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1603587600, 2020, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1603591199, 2020, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1603591200, 2020, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1616893199, 2021, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1616893200, 2021, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1635641999, 2021, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1635642000, 2021, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1635645599, 2021, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1635645600, 2021, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1648342799, 2022, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1648342800, 2022, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1667091599, 2022, 10, 0, 30, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1667091600, 2022, 10, 0, 30, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1667095199, 2022, 10, 0, 30, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1667095200, 2022, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1679792399, 2023, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1679792400, 2023, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1698541199, 2023, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1698541200, 2023, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1698544799, 2023, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1698544800, 2023, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1711846799, 2024, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1711846800, 2024, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1729990799, 2024, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1729990800, 2024, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1729994399, 2024, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1729994400, 2024, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1743296399, 2025, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1743296400, 2025, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1761440399, 2025, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1761440400, 2025, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1761443999, 2025, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1761444000, 2025, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1774745999, 2026, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1774746000, 2026, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1792889999, 2026, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1792890000, 2026, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1792893599, 2026, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1792893600, 2026, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1806195599, 2027, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1806195600, 2027, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1824944399, 2027, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1824944400, 2027, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1824947999, 2027, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1824948000, 2027, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1837645199, 2028, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1837645200, 2028, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1856393999, 2028, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1856394000, 2028, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1856397599, 2028, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1856397600, 2028, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1869094799, 2029, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1869094800, 2029, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1887843599, 2029, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1887843600, 2029, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1887847199, 2029, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1887847200, 2029, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1901149199, 2030, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1901149200, 2030, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1919293199, 2030, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1919293200, 2030, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1919296799, 2030, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1919296800, 2030, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1932598799, 2031, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1932598800, 2031, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1950742799, 2031, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1950742800, 2031, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1950746399, 2031, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1950746400, 2031, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1964048399, 2032, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1964048400, 2032, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(1982797199, 2032, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(1982797200, 2032, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(1982800799, 2032, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(1982800800, 2032, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(1995497999, 2033, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(1995498000, 2033, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2014246799, 2033, 10, 0, 30, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2014246800, 2033, 10, 0, 30, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2014250399, 2033, 10, 0, 30, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2014250400, 2033, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2026947599, 2034, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2026947600, 2034, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2045696399, 2034, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2045696400, 2034, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2045699999, 2034, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2045700000, 2034, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2058397199, 2035, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2058397200, 2035, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2077145999, 2035, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2077146000, 2035, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2077149599, 2035, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2077149600, 2035, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2090451599, 2036, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2090451600, 2036, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2108595599, 2036, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2108595600, 2036, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2108599199, 2036, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2108599200, 2036, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2121901199, 2037, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2121901200, 2037, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2140045199, 2037, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2140045200, 2037, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2140048799, 2037, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2140048800, 2037, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2153350799, 2038, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2153350800, 2038, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2172099599, 2038, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2172099600, 2038, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2172103199, 2038, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2172103200, 2038, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2184800399, 2039, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2184800400, 2039, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2203549199, 2039, 10, 0, 30, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2203549200, 2039, 10, 0, 30, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2203552799, 2039, 10, 0, 30, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2203552800, 2039, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2216249999, 2040, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2216250000, 2040, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2234998799, 2040, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2234998800, 2040, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2235002399, 2040, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2235002400, 2040, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2248304399, 2041, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2248304400, 2041, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2266448399, 2041, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2266448400, 2041, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2266451999, 2041, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2266452000, 2041, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2279753999, 2042, 3, 0, 30, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2279754000, 2042, 3, 0, 30, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2297897999, 2042, 10, 0, 26, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2297898000, 2042, 10, 0, 26, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2297901599, 2042, 10, 0, 26, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2297901600, 2042, 10, 0, 26, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2311203599, 2043, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2311203600, 2043, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2329347599, 2043, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2329347600, 2043, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2329351199, 2043, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2329351200, 2043, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2342653199, 2044, 3, 0, 27, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2342653200, 2044, 3, 0, 27, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2361401999, 2044, 10, 0, 30, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2361402000, 2044, 10, 0, 30, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2361405599, 2044, 10, 0, 30, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2361405600, 2044, 10, 0, 30, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2374102799, 2045, 3, 0, 26, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2374102800, 2045, 3, 0, 26, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2392851599, 2045, 10, 0, 29, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2392851600, 2045, 10, 0, 29, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2392855199, 2045, 10, 0, 29, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2392855200, 2045, 10, 0, 29, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2405552399, 2046, 3, 0, 25, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2405552400, 2046, 3, 0, 25, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2424301199, 2046, 10, 0, 28, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2424301200, 2046, 10, 0, 28, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2424304799, 2046, 10, 0, 28, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2424304800, 2046, 10, 0, 28, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2437606799, 2047, 3, 0, 31, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2437606800, 2047, 3, 0, 31, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2455750799, 2047, 10, 0, 27, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2455750800, 2047, 10, 0, 27, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2455754399, 2047, 10, 0, 27, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2455754400, 2047, 10, 0, 27, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2469056399, 2048, 3, 0, 29, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2469056400, 2048, 3, 0, 29, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2487200399, 2048, 10, 0, 25, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2487200400, 2048, 10, 0, 25, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2487203999, 2048, 10, 0, 25, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2487204000, 2048, 10, 0, 25, 2, 0, 0, 0, 0, "GMT", BIDI)
TESTVECTOR(2500505999, 2049, 3, 0, 28, 0, 59, 59, 0, 0, "GMT", BIDI)
TESTVECTOR(2500506000, 2049, 3, 0, 28, 2, 0, 0, 1, 3600, "BST", BIDI)
TESTVECTOR(2519254799, 2049, 10, 0, 31, 1, 59, 59, 1, 3600, "BST", BIDI)
TESTVECTOR(2519254800, 2049, 10, 0, 31, 1, 0, 0, 0, 0, "GMT", HINT)
TESTVECTOR(2519258399, 2049, 10, 0, 31, 1, 59, 59, 0, 0, "GMT", HINT)
TESTVECTOR(2519258400, 2049, 10, 0, 31, 2, 0, 0, 0, 0, "GMT", BIDI)
TEST_END

TEST_BEGIN(los_angeles, "America/Los_Angeles")
TESTVECTOR(-2717640001, 1883, 11, 0, 18, 12, 7, 1, 0, -28378, "LMT", BIDI)
TESTVECTOR(-2717640000, 1883, 11, 0, 18, 12, 0, 0, 0, -28800, "PST", UNI)
TESTVECTOR(-2717639579, 1883, 11, 0, 18, 12, 7, 1, 0, -28800, "PST", UNI)
TESTVECTOR(-2717639578, 1883, 11, 0, 18, 12, 7, 2, 0, -28800, "PST", BIDI)
TESTVECTOR(-1633269601, 1918, 3, 0, 31, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-1633269600, 1918, 3, 0, 31, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-1615129201, 1918, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-1615129200, 1918, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-1615125601, 1918, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-1615125600, 1918, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-1601820001, 1919, 3, 0, 30, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-1601820000, 1919, 3, 0, 30, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-1583679601, 1919, 10, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-1583679600, 1919, 10, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-1583676001, 1919, 10, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-1583676000, 1919, 10, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-880207201, 1942, 2, 1, 9, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-880207200, 1942, 2, 1, 9, 3, 0, 0, 1, -25200, "PWT", BIDI)
TESTVECTOR(-769395601, 1945, 8, 2, 14, 15, 59, 59, 1, -25200, "PWT", BIDI)
TESTVECTOR(-769395600, 1945, 8, 2, 14, 16, 0, 0, 1, -25200, "PPT", BIDI)
TESTVECTOR(-765385201, 1945, 9, 0, 30, 1, 59, 59, 1, -25200, "PPT", BIDI)
TESTVECTOR(-765385200, 1945, 9, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-765381601, 1945, 9, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-765381600, 1945, 9, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-687967141, 1948, 3, 0, 14, 2, 0, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-687967140, 1948, 3, 0, 14, 3, 1, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-662655601, 1949, 1, 6, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-662655600, 1949, 1, 6, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-662652001, 1949, 1, 6, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-662652000, 1949, 1, 6, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-620838001, 1950, 4, 0, 30, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-620838000, 1950, 4, 0, 30, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-608137201, 1950, 9, 0, 24, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-608137200, 1950, 9, 0, 24, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-608133601, 1950, 9, 0, 24, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-608133600, 1950, 9, 0, 24, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-589388401, 1951, 4, 0, 29, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-589388400, 1951, 4, 0, 29, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-576082801, 1951, 9, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-576082800, 1951, 9, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-576079201, 1951, 9, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-576079200, 1951, 9, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-557938801, 1952, 4, 0, 27, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-557938800, 1952, 4, 0, 27, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-544633201, 1952, 9, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-544633200, 1952, 9, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-544629601, 1952, 9, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-544629600, 1952, 9, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-526489201, 1953, 4, 0, 26, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-526489200, 1953, 4, 0, 26, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-513183601, 1953, 9, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-513183600, 1953, 9, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-513180001, 1953, 9, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-513180000, 1953, 9, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-495039601, 1954, 4, 0, 25, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-495039600, 1954, 4, 0, 25, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-481734001, 1954, 9, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-481734000, 1954, 9, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-481730401, 1954, 9, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-481730400, 1954, 9, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-463590001, 1955, 4, 0, 24, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-463590000, 1955, 4, 0, 24, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-450284401, 1955, 9, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-450284400, 1955, 9, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-450280801, 1955, 9, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-450280800, 1955, 9, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-431535601, 1956, 4, 0, 29, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-431535600, 1956, 4, 0, 29, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-418230001, 1956, 9, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-418230000, 1956, 9, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-418226401, 1956, 9, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-418226400, 1956, 9, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-400086001, 1957, 4, 0, 28, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-400086000, 1957, 4, 0, 28, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-386780401, 1957, 9, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-386780400, 1957, 9, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-386776801, 1957, 9, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-386776800, 1957, 9, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-368636401, 1958, 4, 0, 27, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-368636400, 1958, 4, 0, 27, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-355330801, 1958, 9, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-355330800, 1958, 9, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-355327201, 1958, 9, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-355327200, 1958, 9, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-337186801, 1959, 4, 0, 26, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-337186800, 1959, 4, 0, 26, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-323881201, 1959, 9, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-323881200, 1959, 9, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-323877601, 1959, 9, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-323877600, 1959, 9, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-305737201, 1960, 4, 0, 24, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-305737200, 1960, 4, 0, 24, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-292431601, 1960, 9, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-292431600, 1960, 9, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-292428001, 1960, 9, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-292428000, 1960, 9, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-273682801, 1961, 4, 0, 30, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-273682800, 1961, 4, 0, 30, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-260982001, 1961, 9, 0, 24, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-260982000, 1961, 9, 0, 24, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-260978401, 1961, 9, 0, 24, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-260978400, 1961, 9, 0, 24, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-242233201, 1962, 4, 0, 29, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-242233200, 1962, 4, 0, 29, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-226508401, 1962, 10, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-226508400, 1962, 10, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-226504801, 1962, 10, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-226504800, 1962, 10, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-210783601, 1963, 4, 0, 28, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-210783600, 1963, 4, 0, 28, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-195058801, 1963, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-195058800, 1963, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-195055201, 1963, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-195055200, 1963, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-179334001, 1964, 4, 0, 26, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-179334000, 1964, 4, 0, 26, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-163609201, 1964, 10, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-163609200, 1964, 10, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-163605601, 1964, 10, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-163605600, 1964, 10, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-147884401, 1965, 4, 0, 25, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-147884400, 1965, 4, 0, 25, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-131554801, 1965, 10, 0, 31, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-131554800, 1965, 10, 0, 31, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-131551201, 1965, 10, 0, 31, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-131551200, 1965, 10, 0, 31, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-116434801, 1966, 4, 0, 24, 0, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-116434800, 1966, 4, 0, 24, 2, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-100105201, 1966, 10, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-100105200, 1966, 10, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-100101601, 1966, 10, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-100101600, 1966, 10, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-84376801, 1967, 4, 0, 30, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-84376800, 1967, 4, 0, 30, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-68655601, 1967, 10, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-68655600, 1967, 10, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-68652001, 1967, 10, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-68652000, 1967, 10, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-52927201, 1968, 4, 0, 28, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-52927200, 1968, 4, 0, 28, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-37206001, 1968, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-37206000, 1968, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-37202401, 1968, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-37202400, 1968, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(-21477601, 1969, 4, 0, 27, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(-21477600, 1969, 4, 0, 27, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(-5756401, 1969, 10, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(-5756400, 1969, 10, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(-5752801, 1969, 10, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(-5752800, 1969, 10, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(9971999, 1970, 4, 0, 26, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(9972000, 1970, 4, 0, 26, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(25693199, 1970, 10, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(25693200, 1970, 10, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(25696799, 1970, 10, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(25696800, 1970, 10, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(41421599, 1971, 4, 0, 25, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(41421600, 1971, 4, 0, 25, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(57747599, 1971, 10, 0, 31, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(57747600, 1971, 10, 0, 31, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(57751199, 1971, 10, 0, 31, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(57751200, 1971, 10, 0, 31, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(73475999, 1972, 4, 0, 30, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(73476000, 1972, 4, 0, 30, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(89197199, 1972, 10, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(89197200, 1972, 10, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(89200799, 1972, 10, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(89200800, 1972, 10, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(104925599, 1973, 4, 0, 29, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(104925600, 1973, 4, 0, 29, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(120646799, 1973, 10, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(120646800, 1973, 10, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(120650399, 1973, 10, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(120650400, 1973, 10, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(126698399, 1974, 1, 0, 6, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(126698400, 1974, 1, 0, 6, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(152096399, 1974, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(152096400, 1974, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(152099999, 1974, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(152100000, 1974, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(162381599, 1975, 2, 0, 23, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(162381600, 1975, 2, 0, 23, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(183545999, 1975, 10, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(183546000, 1975, 10, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(183549599, 1975, 10, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(183549600, 1975, 10, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(199274399, 1976, 4, 0, 25, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(199274400, 1976, 4, 0, 25, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(215600399, 1976, 10, 0, 31, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(215600400, 1976, 10, 0, 31, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(215603999, 1976, 10, 0, 31, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(215604000, 1976, 10, 0, 31, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(230723999, 1977, 4, 0, 24, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(230724000, 1977, 4, 0, 24, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(247049999, 1977, 10, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(247050000, 1977, 10, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(247053599, 1977, 10, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(247053600, 1977, 10, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(262778399, 1978, 4, 0, 30, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(262778400, 1978, 4, 0, 30, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(278499599, 1978, 10, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(278499600, 1978, 10, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(278503199, 1978, 10, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(278503200, 1978, 10, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(294227999, 1979, 4, 0, 29, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(294228000, 1979, 4, 0, 29, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(309949199, 1979, 10, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(309949200, 1979, 10, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(309952799, 1979, 10, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(309952800, 1979, 10, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(325677599, 1980, 4, 0, 27, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(325677600, 1980, 4, 0, 27, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(341398799, 1980, 10, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(341398800, 1980, 10, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(341402399, 1980, 10, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(341402400, 1980, 10, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(357127199, 1981, 4, 0, 26, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(357127200, 1981, 4, 0, 26, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(372848399, 1981, 10, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(372848400, 1981, 10, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(372851999, 1981, 10, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(372852000, 1981, 10, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(388576799, 1982, 4, 0, 25, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(388576800, 1982, 4, 0, 25, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(404902799, 1982, 10, 0, 31, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(404902800, 1982, 10, 0, 31, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(404906399, 1982, 10, 0, 31, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(404906400, 1982, 10, 0, 31, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(420026399, 1983, 4, 0, 24, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(420026400, 1983, 4, 0, 24, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(436352399, 1983, 10, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(436352400, 1983, 10, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(436355999, 1983, 10, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(436356000, 1983, 10, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(452080799, 1984, 4, 0, 29, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(452080800, 1984, 4, 0, 29, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(467801999, 1984, 10, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(467802000, 1984, 10, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(467805599, 1984, 10, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(467805600, 1984, 10, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(483530399, 1985, 4, 0, 28, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(483530400, 1985, 4, 0, 28, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(499251599, 1985, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(499251600, 1985, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(499255199, 1985, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(499255200, 1985, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(514979999, 1986, 4, 0, 27, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(514980000, 1986, 4, 0, 27, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(530701199, 1986, 10, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(530701200, 1986, 10, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(530704799, 1986, 10, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(530704800, 1986, 10, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(544615199, 1987, 4, 0, 5, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(544615200, 1987, 4, 0, 5, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(562150799, 1987, 10, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(562150800, 1987, 10, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(562154399, 1987, 10, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(562154400, 1987, 10, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(576064799, 1988, 4, 0, 3, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(576064800, 1988, 4, 0, 3, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(594205199, 1988, 10, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(594205200, 1988, 10, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(594208799, 1988, 10, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(594208800, 1988, 10, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(607514399, 1989, 4, 0, 2, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(607514400, 1989, 4, 0, 2, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(625654799, 1989, 10, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(625654800, 1989, 10, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(625658399, 1989, 10, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(625658400, 1989, 10, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(638963999, 1990, 4, 0, 1, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(638964000, 1990, 4, 0, 1, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(657104399, 1990, 10, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(657104400, 1990, 10, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(657107999, 1990, 10, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(657108000, 1990, 10, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(671018399, 1991, 4, 0, 7, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(671018400, 1991, 4, 0, 7, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(688553999, 1991, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(688554000, 1991, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(688557599, 1991, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(688557600, 1991, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(702467999, 1992, 4, 0, 5, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(702468000, 1992, 4, 0, 5, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(720003599, 1992, 10, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(720003600, 1992, 10, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(720007199, 1992, 10, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(720007200, 1992, 10, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(733917599, 1993, 4, 0, 4, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(733917600, 1993, 4, 0, 4, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(752057999, 1993, 10, 0, 31, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(752058000, 1993, 10, 0, 31, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(752061599, 1993, 10, 0, 31, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(752061600, 1993, 10, 0, 31, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(765367199, 1994, 4, 0, 3, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(765367200, 1994, 4, 0, 3, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(783507599, 1994, 10, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(783507600, 1994, 10, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(783511199, 1994, 10, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(783511200, 1994, 10, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(796816799, 1995, 4, 0, 2, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(796816800, 1995, 4, 0, 2, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(814957199, 1995, 10, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(814957200, 1995, 10, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(814960799, 1995, 10, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(814960800, 1995, 10, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(828871199, 1996, 4, 0, 7, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(828871200, 1996, 4, 0, 7, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(846406799, 1996, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(846406800, 1996, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(846410399, 1996, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(846410400, 1996, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(860320799, 1997, 4, 0, 6, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(860320800, 1997, 4, 0, 6, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(877856399, 1997, 10, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(877856400, 1997, 10, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(877859999, 1997, 10, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(877860000, 1997, 10, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(891770399, 1998, 4, 0, 5, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(891770400, 1998, 4, 0, 5, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(909305999, 1998, 10, 0, 25, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(909306000, 1998, 10, 0, 25, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(909309599, 1998, 10, 0, 25, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(909309600, 1998, 10, 0, 25, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(923219999, 1999, 4, 0, 4, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(923220000, 1999, 4, 0, 4, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(941360399, 1999, 10, 0, 31, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(941360400, 1999, 10, 0, 31, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(941363999, 1999, 10, 0, 31, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(941364000, 1999, 10, 0, 31, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(954669599, 2000, 4, 0, 2, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(954669600, 2000, 4, 0, 2, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(972809999, 2000, 10, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(972810000, 2000, 10, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(972813599, 2000, 10, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(972813600, 2000, 10, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(986119199, 2001, 4, 0, 1, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(986119200, 2001, 4, 0, 1, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1004259599, 2001, 10, 0, 28, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1004259600, 2001, 10, 0, 28, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1004263199, 2001, 10, 0, 28, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1004263200, 2001, 10, 0, 28, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1018173599, 2002, 4, 0, 7, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1018173600, 2002, 4, 0, 7, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1035709199, 2002, 10, 0, 27, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1035709200, 2002, 10, 0, 27, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1035712799, 2002, 10, 0, 27, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1035712800, 2002, 10, 0, 27, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1049623199, 2003, 4, 0, 6, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1049623200, 2003, 4, 0, 6, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1067158799, 2003, 10, 0, 26, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1067158800, 2003, 10, 0, 26, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1067162399, 2003, 10, 0, 26, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1067162400, 2003, 10, 0, 26, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1081072799, 2004, 4, 0, 4, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1081072800, 2004, 4, 0, 4, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1099213199, 2004, 10, 0, 31, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1099213200, 2004, 10, 0, 31, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1099216799, 2004, 10, 0, 31, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1099216800, 2004, 10, 0, 31, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1112522399, 2005, 4, 0, 3, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1112522400, 2005, 4, 0, 3, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1130662799, 2005, 10, 0, 30, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1130662800, 2005, 10, 0, 30, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1130666399, 2005, 10, 0, 30, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1130666400, 2005, 10, 0, 30, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1143971999, 2006, 4, 0, 2, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1143972000, 2006, 4, 0, 2, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1162112399, 2006, 10, 0, 29, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1162112400, 2006, 10, 0, 29, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1162115999, 2006, 10, 0, 29, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1162116000, 2006, 10, 0, 29, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1173607199, 2007, 3, 0, 11, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1173607200, 2007, 3, 0, 11, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1194166799, 2007, 11, 0, 4, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1194166800, 2007, 11, 0, 4, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1194170399, 2007, 11, 0, 4, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1194170400, 2007, 11, 0, 4, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1205056799, 2008, 3, 0, 9, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1205056800, 2008, 3, 0, 9, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1225616399, 2008, 11, 0, 2, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1225616400, 2008, 11, 0, 2, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1225619999, 2008, 11, 0, 2, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1225620000, 2008, 11, 0, 2, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1236506399, 2009, 3, 0, 8, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1236506400, 2009, 3, 0, 8, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1257065999, 2009, 11, 0, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1257066000, 2009, 11, 0, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1257069599, 2009, 11, 0, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1257069600, 2009, 11, 0, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1268560799, 2010, 3, 0, 14, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1268560800, 2010, 3, 0, 14, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1289120399, 2010, 11, 0, 7, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1289120400, 2010, 11, 0, 7, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1289123999, 2010, 11, 0, 7, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1289124000, 2010, 11, 0, 7, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1300010399, 2011, 3, 0, 13, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1300010400, 2011, 3, 0, 13, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1320569999, 2011, 11, 0, 6, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1320570000, 2011, 11, 0, 6, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1320573599, 2011, 11, 0, 6, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1320573600, 2011, 11, 0, 6, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1331459999, 2012, 3, 0, 11, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1331460000, 2012, 3, 0, 11, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1352019599, 2012, 11, 0, 4, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1352019600, 2012, 11, 0, 4, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1352023199, 2012, 11, 0, 4, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1352023200, 2012, 11, 0, 4, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1362909599, 2013, 3, 0, 10, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1362909600, 2013, 3, 0, 10, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1383469199, 2013, 11, 0, 3, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1383469200, 2013, 11, 0, 3, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1383472799, 2013, 11, 0, 3, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1383472800, 2013, 11, 0, 3, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1394359199, 2014, 3, 0, 9, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1394359200, 2014, 3, 0, 9, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1414918799, 2014, 11, 0, 2, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1414918800, 2014, 11, 0, 2, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1414922399, 2014, 11, 0, 2, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1414922400, 2014, 11, 0, 2, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1425808799, 2015, 3, 0, 8, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1425808800, 2015, 3, 0, 8, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1446368399, 2015, 11, 0, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1446368400, 2015, 11, 0, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1446371999, 2015, 11, 0, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1446372000, 2015, 11, 0, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1457863199, 2016, 3, 0, 13, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1457863200, 2016, 3, 0, 13, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1478422799, 2016, 11, 0, 6, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1478422800, 2016, 11, 0, 6, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1478426399, 2016, 11, 0, 6, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1478426400, 2016, 11, 0, 6, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1489312799, 2017, 3, 0, 12, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1489312800, 2017, 3, 0, 12, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1509872399, 2017, 11, 0, 5, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1509872400, 2017, 11, 0, 5, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1509875999, 2017, 11, 0, 5, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1509876000, 2017, 11, 0, 5, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1520762399, 2018, 3, 0, 11, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1520762400, 2018, 3, 0, 11, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1541321999, 2018, 11, 0, 4, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1541322000, 2018, 11, 0, 4, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1541325599, 2018, 11, 0, 4, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1541325600, 2018, 11, 0, 4, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1552211999, 2019, 3, 0, 10, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1552212000, 2019, 3, 0, 10, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1572771599, 2019, 11, 0, 3, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1572771600, 2019, 11, 0, 3, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1572775199, 2019, 11, 0, 3, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1572775200, 2019, 11, 0, 3, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1583661599, 2020, 3, 0, 8, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1583661600, 2020, 3, 0, 8, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1604221199, 2020, 11, 0, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1604221200, 2020, 11, 0, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1604224799, 2020, 11, 0, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1604224800, 2020, 11, 0, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1615715999, 2021, 3, 0, 14, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1615716000, 2021, 3, 0, 14, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1636275599, 2021, 11, 0, 7, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1636275600, 2021, 11, 0, 7, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1636279199, 2021, 11, 0, 7, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1636279200, 2021, 11, 0, 7, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1647165599, 2022, 3, 0, 13, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1647165600, 2022, 3, 0, 13, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1667725199, 2022, 11, 0, 6, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1667725200, 2022, 11, 0, 6, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1667728799, 2022, 11, 0, 6, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1667728800, 2022, 11, 0, 6, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1678615199, 2023, 3, 0, 12, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1678615200, 2023, 3, 0, 12, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1699174799, 2023, 11, 0, 5, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1699174800, 2023, 11, 0, 5, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1699178399, 2023, 11, 0, 5, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1699178400, 2023, 11, 0, 5, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1710064799, 2024, 3, 0, 10, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1710064800, 2024, 3, 0, 10, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1730624399, 2024, 11, 0, 3, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1730624400, 2024, 11, 0, 3, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1730627999, 2024, 11, 0, 3, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1730628000, 2024, 11, 0, 3, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1741514399, 2025, 3, 0, 9, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1741514400, 2025, 3, 0, 9, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1762073999, 2025, 11, 0, 2, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1762074000, 2025, 11, 0, 2, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1762077599, 2025, 11, 0, 2, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1762077600, 2025, 11, 0, 2, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1772963999, 2026, 3, 0, 8, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1772964000, 2026, 3, 0, 8, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1793523599, 2026, 11, 0, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1793523600, 2026, 11, 0, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1793527199, 2026, 11, 0, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1793527200, 2026, 11, 0, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1805018399, 2027, 3, 0, 14, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1805018400, 2027, 3, 0, 14, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1825577999, 2027, 11, 0, 7, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1825578000, 2027, 11, 0, 7, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1825581599, 2027, 11, 0, 7, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1825581600, 2027, 11, 0, 7, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1836467999, 2028, 3, 0, 12, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1836468000, 2028, 3, 0, 12, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1857027599, 2028, 11, 0, 5, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1857027600, 2028, 11, 0, 5, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1857031199, 2028, 11, 0, 5, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1857031200, 2028, 11, 0, 5, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1867917599, 2029, 3, 0, 11, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1867917600, 2029, 3, 0, 11, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1888477199, 2029, 11, 0, 4, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1888477200, 2029, 11, 0, 4, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1888480799, 2029, 11, 0, 4, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1888480800, 2029, 11, 0, 4, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1899367199, 2030, 3, 0, 10, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1899367200, 2030, 3, 0, 10, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1919926799, 2030, 11, 0, 3, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1919926800, 2030, 11, 0, 3, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1919930399, 2030, 11, 0, 3, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1919930400, 2030, 11, 0, 3, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1930816799, 2031, 3, 0, 9, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1930816800, 2031, 3, 0, 9, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1951376399, 2031, 11, 0, 2, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1951376400, 2031, 11, 0, 2, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1951379999, 2031, 11, 0, 2, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1951380000, 2031, 11, 0, 2, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1962871199, 2032, 3, 0, 14, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1962871200, 2032, 3, 0, 14, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(1983430799, 2032, 11, 0, 7, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(1983430800, 2032, 11, 0, 7, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(1983434399, 2032, 11, 0, 7, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(1983434400, 2032, 11, 0, 7, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(1994320799, 2033, 3, 0, 13, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(1994320800, 2033, 3, 0, 13, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2014880399, 2033, 11, 0, 6, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2014880400, 2033, 11, 0, 6, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2014883999, 2033, 11, 0, 6, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2014884000, 2033, 11, 0, 6, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2025770399, 2034, 3, 0, 12, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2025770400, 2034, 3, 0, 12, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2046329999, 2034, 11, 0, 5, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2046330000, 2034, 11, 0, 5, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2046333599, 2034, 11, 0, 5, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2046333600, 2034, 11, 0, 5, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2057219999, 2035, 3, 0, 11, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2057220000, 2035, 3, 0, 11, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2077779599, 2035, 11, 0, 4, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2077779600, 2035, 11, 0, 4, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2077783199, 2035, 11, 0, 4, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2077783200, 2035, 11, 0, 4, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2088669599, 2036, 3, 0, 9, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2088669600, 2036, 3, 0, 9, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2109229199, 2036, 11, 0, 2, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2109229200, 2036, 11, 0, 2, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2109232799, 2036, 11, 0, 2, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2109232800, 2036, 11, 0, 2, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2120119199, 2037, 3, 0, 8, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2120119200, 2037, 3, 0, 8, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2140678799, 2037, 11, 0, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2140678800, 2037, 11, 0, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2140682399, 2037, 11, 0, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2140682400, 2037, 11, 0, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2152173599, 2038, 3, 0, 14, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2152173600, 2038, 3, 0, 14, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2172733199, 2038, 11, 0, 7, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2172733200, 2038, 11, 0, 7, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2172736799, 2038, 11, 0, 7, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2172736800, 2038, 11, 0, 7, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2183623199, 2039, 3, 0, 13, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2183623200, 2039, 3, 0, 13, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2204182799, 2039, 11, 0, 6, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2204182800, 2039, 11, 0, 6, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2204186399, 2039, 11, 0, 6, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2204186400, 2039, 11, 0, 6, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2215072799, 2040, 3, 0, 11, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2215072800, 2040, 3, 0, 11, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2235632399, 2040, 11, 0, 4, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2235632400, 2040, 11, 0, 4, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2235635999, 2040, 11, 0, 4, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2235636000, 2040, 11, 0, 4, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2246522399, 2041, 3, 0, 10, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2246522400, 2041, 3, 0, 10, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2267081999, 2041, 11, 0, 3, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2267082000, 2041, 11, 0, 3, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2267085599, 2041, 11, 0, 3, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2267085600, 2041, 11, 0, 3, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2277971999, 2042, 3, 0, 9, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2277972000, 2042, 3, 0, 9, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2298531599, 2042, 11, 0, 2, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2298531600, 2042, 11, 0, 2, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2298535199, 2042, 11, 0, 2, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2298535200, 2042, 11, 0, 2, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2309421599, 2043, 3, 0, 8, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2309421600, 2043, 3, 0, 8, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2329981199, 2043, 11, 0, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2329981200, 2043, 11, 0, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2329984799, 2043, 11, 0, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2329984800, 2043, 11, 0, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2341475999, 2044, 3, 0, 13, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2341476000, 2044, 3, 0, 13, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2362035599, 2044, 11, 0, 6, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2362035600, 2044, 11, 0, 6, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2362039199, 2044, 11, 0, 6, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2362039200, 2044, 11, 0, 6, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2372925599, 2045, 3, 0, 12, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2372925600, 2045, 3, 0, 12, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2393485199, 2045, 11, 0, 5, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2393485200, 2045, 11, 0, 5, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2393488799, 2045, 11, 0, 5, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2393488800, 2045, 11, 0, 5, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2404375199, 2046, 3, 0, 11, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2404375200, 2046, 3, 0, 11, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2424934799, 2046, 11, 0, 4, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2424934800, 2046, 11, 0, 4, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2424938399, 2046, 11, 0, 4, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2424938400, 2046, 11, 0, 4, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2435824799, 2047, 3, 0, 10, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2435824800, 2047, 3, 0, 10, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2456384399, 2047, 11, 0, 3, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2456384400, 2047, 11, 0, 3, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2456387999, 2047, 11, 0, 3, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2456388000, 2047, 11, 0, 3, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2467274399, 2048, 3, 0, 8, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2467274400, 2048, 3, 0, 8, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2487833999, 2048, 11, 0, 1, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2487834000, 2048, 11, 0, 1, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2487837599, 2048, 11, 0, 1, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2487837600, 2048, 11, 0, 1, 2, 0, 0, 0, -28800, "PST", BIDI)
TESTVECTOR(2499328799, 2049, 3, 0, 14, 1, 59, 59, 0, -28800, "PST", BIDI)
TESTVECTOR(2499328800, 2049, 3, 0, 14, 3, 0, 0, 1, -25200, "PDT", BIDI)
TESTVECTOR(2519888399, 2049, 11, 0, 7, 1, 59, 59, 1, -25200, "PDT", BIDI)
TESTVECTOR(2519888400, 2049, 11, 0, 7, 1, 0, 0, 0, -28800, "PST", HINT)
TESTVECTOR(2519891999, 2049, 11, 0, 7, 1, 59, 59, 0, -28800, "PST", HINT)
TESTVECTOR(2519892000, 2049, 11, 0, 7, 2, 0, 0, 0, -28800, "PST", BIDI)
TEST_END

TEST_BEGIN(marengo, "America/Indiana/Marengo")
TESTVECTOR(-2717647201, 1883, 11, 0, 18, 12, 14, 36, 0, -20723, "LMT", BIDI)
TESTVECTOR(-2717647200, 1883, 11, 0, 18, 12, 0, 0, 0, -21600, "CST", UNI)
TESTVECTOR(-2717646324, 1883, 11, 0, 18, 12, 14, 36, 0, -21600, "CST", UNI)
TESTVECTOR(-2717646323, 1883, 11, 0, 18, 12, 14, 37, 0, -21600, "CST", BIDI)
TESTVECTOR(-1633276801, 1918, 3, 0, 31, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1633276800, 1918, 3, 0, 31, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-1615136401, 1918, 10, 0, 27, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-1615136400, 1918, 10, 0, 27, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1615132801, 1918, 10, 0, 27, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1615132800, 1918, 10, 0, 27, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-1601827201, 1919, 3, 0, 30, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-1601827200, 1919, 3, 0, 30, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-1583686801, 1919, 10, 0, 26, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-1583686800, 1919, 10, 0, 26, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-1583683201, 1919, 10, 0, 26, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-1583683200, 1919, 10, 0, 26, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-880214401, 1942, 2, 1, 9, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-880214400, 1942, 2, 1, 9, 3, 0, 0, 1, -18000, "CWT", BIDI)
TESTVECTOR(-769395601, 1945, 8, 2, 14, 17, 59, 59, 1, -18000, "CWT", BIDI)
TESTVECTOR(-769395600, 1945, 8, 2, 14, 18, 0, 0, 1, -18000, "CPT", BIDI)
TESTVECTOR(-765392401, 1945, 9, 0, 30, 1, 59, 59, 1, -18000, "CPT", BIDI)
TESTVECTOR(-765392400, 1945, 9, 0, 30, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-765388801, 1945, 9, 0, 30, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-765388800, 1945, 9, 0, 30, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-589392001, 1951, 4, 0, 29, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-589392000, 1951, 4, 0, 29, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-576090001, 1951, 9, 0, 30, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-576090000, 1951, 9, 0, 30, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-576086401, 1951, 9, 0, 30, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-576086400, 1951, 9, 0, 30, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-495043201, 1954, 4, 0, 25, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-495043200, 1954, 4, 0, 25, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-481741201, 1954, 9, 0, 26, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-481741200, 1954, 9, 0, 26, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-481737601, 1954, 9, 0, 26, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-481737600, 1954, 9, 0, 26, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-463593601, 1955, 4, 0, 24, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-463593600, 1955, 4, 0, 24, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-450291601, 1955, 9, 0, 25, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-450291600, 1955, 9, 0, 25, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-450288001, 1955, 9, 0, 25, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-450288000, 1955, 9, 0, 25, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-431539201, 1956, 4, 0, 29, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-431539200, 1956, 4, 0, 29, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-418237201, 1956, 9, 0, 30, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-418237200, 1956, 9, 0, 30, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-418233601, 1956, 9, 0, 30, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-418233600, 1956, 9, 0, 30, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-400089601, 1957, 4, 0, 28, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-400089600, 1957, 4, 0, 28, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-386787601, 1957, 9, 0, 29, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-386787600, 1957, 9, 0, 29, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-386784001, 1957, 9, 0, 29, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-386784000, 1957, 9, 0, 29, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-368640001, 1958, 4, 0, 27, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-368640000, 1958, 4, 0, 27, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-355338001, 1958, 9, 0, 28, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-355338000, 1958, 9, 0, 28, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-355334401, 1958, 9, 0, 28, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-355334400, 1958, 9, 0, 28, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-337190401, 1959, 4, 0, 26, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-337190400, 1959, 4, 0, 26, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-323888401, 1959, 9, 0, 27, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-323888400, 1959, 9, 0, 27, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-323884801, 1959, 9, 0, 27, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-323884800, 1959, 9, 0, 27, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-305740801, 1960, 4, 0, 24, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-305740800, 1960, 4, 0, 24, 3, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(-292438801, 1960, 9, 0, 25, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(-292438800, 1960, 9, 0, 25, 1, 0, 0, 0, -21600, "CST", HINT)
TESTVECTOR(-292435201, 1960, 9, 0, 25, 1, 59, 59, 0, -21600, "CST", HINT)
TESTVECTOR(-292435200, 1960, 9, 0, 25, 2, 0, 0, 0, -21600, "CST", BIDI)
TESTVECTOR(-273686401, 1961, 4, 0, 30, 1, 59, 59, 0, -21600, "CST", BIDI)
TESTVECTOR(-273686400, 1961, 4, 0, 30, 3, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(-21488401, 1969, 4, 0, 27, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(-21488400, 1969, 4, 0, 27, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(-5767201, 1969, 10, 0, 26, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(-5767200, 1969, 10, 0, 26, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(-5763601, 1969, 10, 0, 26, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(-5763600, 1969, 10, 0, 26, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(9961199, 1970, 4, 0, 26, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(9961200, 1970, 4, 0, 26, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(25682399, 1970, 10, 0, 25, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(25682400, 1970, 10, 0, 25, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(25685999, 1970, 10, 0, 25, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(25686000, 1970, 10, 0, 25, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(41410799, 1971, 4, 0, 25, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(41410800, 1971, 4, 0, 25, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(57736799, 1971, 10, 0, 31, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(57736800, 1971, 10, 0, 31, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(57740399, 1971, 10, 0, 31, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(57740400, 1971, 10, 0, 31, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(73465199, 1972, 4, 0, 30, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(73465200, 1972, 4, 0, 30, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(89186399, 1972, 10, 0, 29, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(89186400, 1972, 10, 0, 29, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(89189999, 1972, 10, 0, 29, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(89190000, 1972, 10, 0, 29, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(104914799, 1973, 4, 0, 29, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(104914800, 1973, 4, 0, 29, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(120635999, 1973, 10, 0, 28, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(120636000, 1973, 10, 0, 28, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(120639599, 1973, 10, 0, 28, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(120639600, 1973, 10, 0, 28, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(126687599, 1974, 1, 0, 6, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(126687600, 1974, 1, 0, 6, 2, 0, 0, 1, -18000, "CDT", BIDI)
TESTVECTOR(152089199, 1974, 10, 0, 27, 1, 59, 59, 1, -18000, "CDT", BIDI)
TESTVECTOR(152089200, 1974, 10, 0, 27, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(162370799, 1975, 2, 0, 23, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(162370800, 1975, 2, 0, 23, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(183535199, 1975, 10, 0, 26, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(183535200, 1975, 10, 0, 26, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(183538799, 1975, 10, 0, 26, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(183538800, 1975, 10, 0, 26, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1143961199, 2006, 4, 0, 2, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1143961200, 2006, 4, 0, 2, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1162101599, 2006, 10, 0, 29, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1162101600, 2006, 10, 0, 29, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1162105199, 2006, 10, 0, 29, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1162105200, 2006, 10, 0, 29, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1173596399, 2007, 3, 0, 11, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1173596400, 2007, 3, 0, 11, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1194155999, 2007, 11, 0, 4, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1194156000, 2007, 11, 0, 4, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1194159599, 2007, 11, 0, 4, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1194159600, 2007, 11, 0, 4, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1205045999, 2008, 3, 0, 9, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1205046000, 2008, 3, 0, 9, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1225605599, 2008, 11, 0, 2, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1225605600, 2008, 11, 0, 2, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1225609199, 2008, 11, 0, 2, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1225609200, 2008, 11, 0, 2, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1236495599, 2009, 3, 0, 8, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1236495600, 2009, 3, 0, 8, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1257055199, 2009, 11, 0, 1, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1257055200, 2009, 11, 0, 1, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1257058799, 2009, 11, 0, 1, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1257058800, 2009, 11, 0, 1, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1268549999, 2010, 3, 0, 14, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1268550000, 2010, 3, 0, 14, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1289109599, 2010, 11, 0, 7, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1289109600, 2010, 11, 0, 7, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1289113199, 2010, 11, 0, 7, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1289113200, 2010, 11, 0, 7, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1299999599, 2011, 3, 0, 13, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1299999600, 2011, 3, 0, 13, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1320559199, 2011, 11, 0, 6, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1320559200, 2011, 11, 0, 6, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1320562799, 2011, 11, 0, 6, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1320562800, 2011, 11, 0, 6, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1331449199, 2012, 3, 0, 11, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1331449200, 2012, 3, 0, 11, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1352008799, 2012, 11, 0, 4, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1352008800, 2012, 11, 0, 4, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1352012399, 2012, 11, 0, 4, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1352012400, 2012, 11, 0, 4, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1362898799, 2013, 3, 0, 10, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1362898800, 2013, 3, 0, 10, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1383458399, 2013, 11, 0, 3, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1383458400, 2013, 11, 0, 3, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1383461999, 2013, 11, 0, 3, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1383462000, 2013, 11, 0, 3, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1394348399, 2014, 3, 0, 9, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1394348400, 2014, 3, 0, 9, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1414907999, 2014, 11, 0, 2, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1414908000, 2014, 11, 0, 2, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1414911599, 2014, 11, 0, 2, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1414911600, 2014, 11, 0, 2, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1425797999, 2015, 3, 0, 8, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1425798000, 2015, 3, 0, 8, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1446357599, 2015, 11, 0, 1, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1446357600, 2015, 11, 0, 1, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1446361199, 2015, 11, 0, 1, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1446361200, 2015, 11, 0, 1, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1457852399, 2016, 3, 0, 13, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1457852400, 2016, 3, 0, 13, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1478411999, 2016, 11, 0, 6, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1478412000, 2016, 11, 0, 6, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1478415599, 2016, 11, 0, 6, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1478415600, 2016, 11, 0, 6, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1489301999, 2017, 3, 0, 12, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1489302000, 2017, 3, 0, 12, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1509861599, 2017, 11, 0, 5, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1509861600, 2017, 11, 0, 5, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1509865199, 2017, 11, 0, 5, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1509865200, 2017, 11, 0, 5, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1520751599, 2018, 3, 0, 11, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1520751600, 2018, 3, 0, 11, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1541311199, 2018, 11, 0, 4, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1541311200, 2018, 11, 0, 4, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1541314799, 2018, 11, 0, 4, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1541314800, 2018, 11, 0, 4, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1552201199, 2019, 3, 0, 10, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1552201200, 2019, 3, 0, 10, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1572760799, 2019, 11, 0, 3, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1572760800, 2019, 11, 0, 3, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1572764399, 2019, 11, 0, 3, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1572764400, 2019, 11, 0, 3, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1583650799, 2020, 3, 0, 8, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1583650800, 2020, 3, 0, 8, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1604210399, 2020, 11, 0, 1, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1604210400, 2020, 11, 0, 1, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1604213999, 2020, 11, 0, 1, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1604214000, 2020, 11, 0, 1, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1615705199, 2021, 3, 0, 14, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1615705200, 2021, 3, 0, 14, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1636264799, 2021, 11, 0, 7, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1636264800, 2021, 11, 0, 7, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1636268399, 2021, 11, 0, 7, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1636268400, 2021, 11, 0, 7, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1647154799, 2022, 3, 0, 13, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1647154800, 2022, 3, 0, 13, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1667714399, 2022, 11, 0, 6, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1667714400, 2022, 11, 0, 6, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1667717999, 2022, 11, 0, 6, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1667718000, 2022, 11, 0, 6, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1678604399, 2023, 3, 0, 12, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1678604400, 2023, 3, 0, 12, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1699163999, 2023, 11, 0, 5, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1699164000, 2023, 11, 0, 5, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1699167599, 2023, 11, 0, 5, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1699167600, 2023, 11, 0, 5, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1710053999, 2024, 3, 0, 10, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1710054000, 2024, 3, 0, 10, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1730613599, 2024, 11, 0, 3, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1730613600, 2024, 11, 0, 3, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1730617199, 2024, 11, 0, 3, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1730617200, 2024, 11, 0, 3, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1741503599, 2025, 3, 0, 9, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1741503600, 2025, 3, 0, 9, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1762063199, 2025, 11, 0, 2, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1762063200, 2025, 11, 0, 2, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1762066799, 2025, 11, 0, 2, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1762066800, 2025, 11, 0, 2, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1772953199, 2026, 3, 0, 8, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1772953200, 2026, 3, 0, 8, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1793512799, 2026, 11, 0, 1, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1793512800, 2026, 11, 0, 1, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1793516399, 2026, 11, 0, 1, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1793516400, 2026, 11, 0, 1, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1805007599, 2027, 3, 0, 14, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1805007600, 2027, 3, 0, 14, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1825567199, 2027, 11, 0, 7, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1825567200, 2027, 11, 0, 7, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1825570799, 2027, 11, 0, 7, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1825570800, 2027, 11, 0, 7, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1836457199, 2028, 3, 0, 12, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1836457200, 2028, 3, 0, 12, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1857016799, 2028, 11, 0, 5, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1857016800, 2028, 11, 0, 5, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1857020399, 2028, 11, 0, 5, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1857020400, 2028, 11, 0, 5, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1867906799, 2029, 3, 0, 11, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1867906800, 2029, 3, 0, 11, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1888466399, 2029, 11, 0, 4, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1888466400, 2029, 11, 0, 4, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1888469999, 2029, 11, 0, 4, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1888470000, 2029, 11, 0, 4, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1899356399, 2030, 3, 0, 10, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1899356400, 2030, 3, 0, 10, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1919915999, 2030, 11, 0, 3, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1919916000, 2030, 11, 0, 3, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1919919599, 2030, 11, 0, 3, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1919919600, 2030, 11, 0, 3, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1930805999, 2031, 3, 0, 9, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1930806000, 2031, 3, 0, 9, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1951365599, 2031, 11, 0, 2, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1951365600, 2031, 11, 0, 2, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1951369199, 2031, 11, 0, 2, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1951369200, 2031, 11, 0, 2, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1962860399, 2032, 3, 0, 14, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1962860400, 2032, 3, 0, 14, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(1983419999, 2032, 11, 0, 7, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(1983420000, 2032, 11, 0, 7, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(1983423599, 2032, 11, 0, 7, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(1983423600, 2032, 11, 0, 7, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(1994309999, 2033, 3, 0, 13, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(1994310000, 2033, 3, 0, 13, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2014869599, 2033, 11, 0, 6, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2014869600, 2033, 11, 0, 6, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2014873199, 2033, 11, 0, 6, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2014873200, 2033, 11, 0, 6, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2025759599, 2034, 3, 0, 12, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2025759600, 2034, 3, 0, 12, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2046319199, 2034, 11, 0, 5, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2046319200, 2034, 11, 0, 5, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2046322799, 2034, 11, 0, 5, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2046322800, 2034, 11, 0, 5, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2057209199, 2035, 3, 0, 11, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2057209200, 2035, 3, 0, 11, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2077768799, 2035, 11, 0, 4, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2077768800, 2035, 11, 0, 4, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2077772399, 2035, 11, 0, 4, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2077772400, 2035, 11, 0, 4, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2088658799, 2036, 3, 0, 9, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2088658800, 2036, 3, 0, 9, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2109218399, 2036, 11, 0, 2, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2109218400, 2036, 11, 0, 2, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2109221999, 2036, 11, 0, 2, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2109222000, 2036, 11, 0, 2, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2120108399, 2037, 3, 0, 8, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2120108400, 2037, 3, 0, 8, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2140667999, 2037, 11, 0, 1, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2140668000, 2037, 11, 0, 1, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2140671599, 2037, 11, 0, 1, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2140671600, 2037, 11, 0, 1, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2152162799, 2038, 3, 0, 14, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2152162800, 2038, 3, 0, 14, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2172722399, 2038, 11, 0, 7, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2172722400, 2038, 11, 0, 7, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2172725999, 2038, 11, 0, 7, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2172726000, 2038, 11, 0, 7, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2183612399, 2039, 3, 0, 13, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2183612400, 2039, 3, 0, 13, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2204171999, 2039, 11, 0, 6, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2204172000, 2039, 11, 0, 6, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2204175599, 2039, 11, 0, 6, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2204175600, 2039, 11, 0, 6, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2215061999, 2040, 3, 0, 11, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2215062000, 2040, 3, 0, 11, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2235621599, 2040, 11, 0, 4, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2235621600, 2040, 11, 0, 4, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2235625199, 2040, 11, 0, 4, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2235625200, 2040, 11, 0, 4, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2246511599, 2041, 3, 0, 10, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2246511600, 2041, 3, 0, 10, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2267071199, 2041, 11, 0, 3, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2267071200, 2041, 11, 0, 3, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2267074799, 2041, 11, 0, 3, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2267074800, 2041, 11, 0, 3, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2277961199, 2042, 3, 0, 9, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2277961200, 2042, 3, 0, 9, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2298520799, 2042, 11, 0, 2, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2298520800, 2042, 11, 0, 2, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2298524399, 2042, 11, 0, 2, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2298524400, 2042, 11, 0, 2, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2309410799, 2043, 3, 0, 8, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2309410800, 2043, 3, 0, 8, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2329970399, 2043, 11, 0, 1, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2329970400, 2043, 11, 0, 1, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2329973999, 2043, 11, 0, 1, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2329974000, 2043, 11, 0, 1, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2341465199, 2044, 3, 0, 13, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2341465200, 2044, 3, 0, 13, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2362024799, 2044, 11, 0, 6, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2362024800, 2044, 11, 0, 6, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2362028399, 2044, 11, 0, 6, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2362028400, 2044, 11, 0, 6, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2372914799, 2045, 3, 0, 12, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2372914800, 2045, 3, 0, 12, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2393474399, 2045, 11, 0, 5, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2393474400, 2045, 11, 0, 5, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2393477999, 2045, 11, 0, 5, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2393478000, 2045, 11, 0, 5, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2404364399, 2046, 3, 0, 11, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2404364400, 2046, 3, 0, 11, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2424923999, 2046, 11, 0, 4, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2424924000, 2046, 11, 0, 4, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2424927599, 2046, 11, 0, 4, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2424927600, 2046, 11, 0, 4, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2435813999, 2047, 3, 0, 10, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2435814000, 2047, 3, 0, 10, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2456373599, 2047, 11, 0, 3, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2456373600, 2047, 11, 0, 3, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2456377199, 2047, 11, 0, 3, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2456377200, 2047, 11, 0, 3, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2467263599, 2048, 3, 0, 8, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2467263600, 2048, 3, 0, 8, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2487823199, 2048, 11, 0, 1, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2487823200, 2048, 11, 0, 1, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2487826799, 2048, 11, 0, 1, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2487826800, 2048, 11, 0, 1, 2, 0, 0, 0, -18000, "EST", BIDI)
TESTVECTOR(2499317999, 2049, 3, 0, 14, 1, 59, 59, 0, -18000, "EST", BIDI)
TESTVECTOR(2499318000, 2049, 3, 0, 14, 3, 0, 0, 1, -14400, "EDT", BIDI)
TESTVECTOR(2519877599, 2049, 11, 0, 7, 1, 59, 59, 1, -14400, "EDT", BIDI)
TESTVECTOR(2519877600, 2049, 11, 0, 7, 1, 0, 0, 0, -18000, "EST", HINT)
TESTVECTOR(2519881199, 2049, 11, 0, 7, 1, 59, 59, 0, -18000, "EST", HINT)
TESTVECTOR(2519881200, 2049, 11, 0, 7, 2, 0, 0, 0, -18000, "EST", BIDI)
TEST_END

TEST_BEGIN(minsk, "Europe/Minsk")
TESTVECTOR(-2840147417, 1879, 12, 3, 31, 23, 59, 59, 0, 6616, "LMT", BIDI)
TESTVECTOR(-2840147416, 1879, 12, 3, 31, 23, 59, 44, 0, 6600, "MMT", UNI)
TESTVECTOR(-2840147401, 1879, 12, 3, 31, 23, 59, 59, 0, 6600, "MMT", UNI)
TESTVECTOR(-2840147400, 1880, 1, 4, 1, 0, 0, 0, 0, 6600, "MMT", BIDI)
TESTVECTOR(-1441158601, 1924, 5, 4, 1, 23, 59, 59, 0, 6600, "MMT", BIDI)
TESTVECTOR(-1441158600, 1924, 5, 5, 2, 0, 10, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(-1247536801, 1930, 6, 5, 20, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(-1247536800, 1930, 6, 6, 21, 1, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(-899780401, 1941, 6, 5, 27, 23, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(-899780400, 1941, 6, 5, 27, 23, 0, 0, 1, 7200, "CEST", UNI)
TESTVECTOR(-899776801, 1941, 6, 5, 27, 23, 59, 59, 1, 7200, "CEST", UNI)
TESTVECTOR(-899776800, 1941, 6, 6, 28, 0, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-857257201, 1942, 11, 1, 2, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-857257200, 1942, 11, 1, 2, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-857253601, 1942, 11, 1, 2, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-857253600, 1942, 11, 1, 2, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-844556401, 1943, 3, 1, 29, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-844556400, 1943, 3, 1, 29, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-828226801, 1943, 10, 1, 4, 2, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-828226800, 1943, 10, 1, 4, 2, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-828223201, 1943, 10, 1, 4, 2, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-828223200, 1943, 10, 1, 4, 3, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-812502001, 1944, 4, 1, 3, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-812502000, 1944, 4, 1, 3, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-804650401, 1944, 7, 0, 2, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-804650400, 1944, 7, 1, 3, 1, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(354920399, 1981, 3, 2, 31, 23, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(354920400, 1981, 4, 3, 1, 1, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(370727999, 1981, 9, 3, 30, 23, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(370728000, 1981, 9, 3, 30, 23, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(370731599, 1981, 9, 3, 30, 23, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(370731600, 1981, 10, 4, 1, 0, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(386456399, 1982, 3, 3, 31, 23, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(386456400, 1982, 4, 4, 1, 1, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(402263999, 1982, 9, 4, 30, 23, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(402264000, 1982, 9, 4, 30, 23, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(402267599, 1982, 9, 4, 30, 23, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(402267600, 1982, 10, 5, 1, 0, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(417992399, 1983, 3, 4, 31, 23, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(417992400, 1983, 4, 5, 1, 1, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(433799999, 1983, 9, 5, 30, 23, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(433800000, 1983, 9, 5, 30, 23, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(433803599, 1983, 9, 5, 30, 23, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(433803600, 1983, 10, 6, 1, 0, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(449614799, 1984, 3, 6, 31, 23, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(449614800, 1984, 4, 0, 1, 1, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(465346799, 1984, 9, 0, 30, 2, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(465346800, 1984, 9, 0, 30, 2, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(465350399, 1984, 9, 0, 30, 2, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(465350400, 1984, 9, 0, 30, 3, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(481071599, 1985, 3, 0, 31, 1, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(481071600, 1985, 3, 0, 31, 3, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(496796399, 1985, 9, 0, 29, 2, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(496796400, 1985, 9, 0, 29, 2, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(496799999, 1985, 9, 0, 29, 2, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(496800000, 1985, 9, 0, 29, 3, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(512521199, 1986, 3, 0, 30, 1, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(512521200, 1986, 3, 0, 30, 3, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(528245999, 1986, 9, 0, 28, 2, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(528246000, 1986, 9, 0, 28, 2, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(528249599, 1986, 9, 0, 28, 2, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(528249600, 1986, 9, 0, 28, 3, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(543970799, 1987, 3, 0, 29, 1, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(543970800, 1987, 3, 0, 29, 3, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(559695599, 1987, 9, 0, 27, 2, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(559695600, 1987, 9, 0, 27, 2, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(559699199, 1987, 9, 0, 27, 2, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(559699200, 1987, 9, 0, 27, 3, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(575420399, 1988, 3, 0, 27, 1, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(575420400, 1988, 3, 0, 27, 3, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(591145199, 1988, 9, 0, 25, 2, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(591145200, 1988, 9, 0, 25, 2, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(591148799, 1988, 9, 0, 25, 2, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(591148800, 1988, 9, 0, 25, 3, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(606869999, 1989, 3, 0, 26, 1, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(606870000, 1989, 3, 0, 26, 3, 0, 0, 1, 14400, "MSD", BIDI)
TESTVECTOR(622594799, 1989, 9, 0, 24, 2, 59, 59, 1, 14400, "MSD", BIDI)
TESTVECTOR(622594800, 1989, 9, 0, 24, 2, 0, 0, 0, 10800, "MSK", HINT)
TESTVECTOR(622598399, 1989, 9, 0, 24, 2, 59, 59, 0, 10800, "MSK", HINT)
TESTVECTOR(622598400, 1989, 9, 0, 24, 3, 0, 0, 0, 10800, "MSK", BIDI)
TESTVECTOR(670373999, 1991, 3, 0, 31, 1, 59, 59, 0, 10800, "MSK", BIDI)
TESTVECTOR(670374000, 1991, 3, 0, 31, 2, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(686102399, 1991, 9, 0, 29, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(686102400, 1991, 9, 0, 29, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(686105999, 1991, 9, 0, 29, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(686106000, 1991, 9, 0, 29, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(701827199, 1992, 3, 0, 29, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(701827200, 1992, 3, 0, 29, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(717551999, 1992, 9, 0, 27, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(717552000, 1992, 9, 0, 27, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(717555599, 1992, 9, 0, 27, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(717555600, 1992, 9, 0, 27, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(733276799, 1993, 3, 0, 28, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(733276800, 1993, 3, 0, 28, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(749001599, 1993, 9, 0, 26, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(749001600, 1993, 9, 0, 26, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(749005199, 1993, 9, 0, 26, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(749005200, 1993, 9, 0, 26, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(764726399, 1994, 3, 0, 27, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(764726400, 1994, 3, 0, 27, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(780451199, 1994, 9, 0, 25, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(780451200, 1994, 9, 0, 25, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(780454799, 1994, 9, 0, 25, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(780454800, 1994, 9, 0, 25, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(796175999, 1995, 3, 0, 26, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(796176000, 1995, 3, 0, 26, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(811900799, 1995, 9, 0, 24, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(811900800, 1995, 9, 0, 24, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(811904399, 1995, 9, 0, 24, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(811904400, 1995, 9, 0, 24, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(828230399, 1996, 3, 0, 31, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(828230400, 1996, 3, 0, 31, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(846374399, 1996, 10, 0, 27, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(846374400, 1996, 10, 0, 27, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(846377999, 1996, 10, 0, 27, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(846378000, 1996, 10, 0, 27, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(859679999, 1997, 3, 0, 30, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(859680000, 1997, 3, 0, 30, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(877823999, 1997, 10, 0, 26, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(877824000, 1997, 10, 0, 26, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(877827599, 1997, 10, 0, 26, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(877827600, 1997, 10, 0, 26, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(891129599, 1998, 3, 0, 29, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(891129600, 1998, 3, 0, 29, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(909273599, 1998, 10, 0, 25, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(909273600, 1998, 10, 0, 25, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(909277199, 1998, 10, 0, 25, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(909277200, 1998, 10, 0, 25, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(922579199, 1999, 3, 0, 28, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(922579200, 1999, 3, 0, 28, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(941327999, 1999, 10, 0, 31, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(941328000, 1999, 10, 0, 31, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(941331599, 1999, 10, 0, 31, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(941331600, 1999, 10, 0, 31, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(954028799, 2000, 3, 0, 26, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(954028800, 2000, 3, 0, 26, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(972777599, 2000, 10, 0, 29, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(972777600, 2000, 10, 0, 29, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(972781199, 2000, 10, 0, 29, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(972781200, 2000, 10, 0, 29, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(985478399, 2001, 3, 0, 25, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(985478400, 2001, 3, 0, 25, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1004227199, 2001, 10, 0, 28, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1004227200, 2001, 10, 0, 28, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1004230799, 2001, 10, 0, 28, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1004230800, 2001, 10, 0, 28, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1017532799, 2002, 3, 0, 31, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1017532800, 2002, 3, 0, 31, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1035676799, 2002, 10, 0, 27, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1035676800, 2002, 10, 0, 27, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1035680399, 2002, 10, 0, 27, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1035680400, 2002, 10, 0, 27, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1048982399, 2003, 3, 0, 30, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1048982400, 2003, 3, 0, 30, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1067126399, 2003, 10, 0, 26, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1067126400, 2003, 10, 0, 26, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1067129999, 2003, 10, 0, 26, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1067130000, 2003, 10, 0, 26, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1080431999, 2004, 3, 0, 28, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1080432000, 2004, 3, 0, 28, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1099180799, 2004, 10, 0, 31, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1099180800, 2004, 10, 0, 31, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1099184399, 2004, 10, 0, 31, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1099184400, 2004, 10, 0, 31, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1111881599, 2005, 3, 0, 27, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1111881600, 2005, 3, 0, 27, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1130630399, 2005, 10, 0, 30, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1130630400, 2005, 10, 0, 30, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1130633999, 2005, 10, 0, 30, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1130634000, 2005, 10, 0, 30, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1143331199, 2006, 3, 0, 26, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1143331200, 2006, 3, 0, 26, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1162079999, 2006, 10, 0, 29, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1162080000, 2006, 10, 0, 29, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1162083599, 2006, 10, 0, 29, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1162083600, 2006, 10, 0, 29, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1174780799, 2007, 3, 0, 25, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1174780800, 2007, 3, 0, 25, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1193529599, 2007, 10, 0, 28, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1193529600, 2007, 10, 0, 28, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1193533199, 2007, 10, 0, 28, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1193533200, 2007, 10, 0, 28, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1206835199, 2008, 3, 0, 30, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1206835200, 2008, 3, 0, 30, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1224979199, 2008, 10, 0, 26, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1224979200, 2008, 10, 0, 26, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1224982799, 2008, 10, 0, 26, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1224982800, 2008, 10, 0, 26, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1238284799, 2009, 3, 0, 29, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1238284800, 2009, 3, 0, 29, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1256428799, 2009, 10, 0, 25, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1256428800, 2009, 10, 0, 25, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1256432399, 2009, 10, 0, 25, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1256432400, 2009, 10, 0, 25, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1269734399, 2010, 3, 0, 28, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1269734400, 2010, 3, 0, 28, 3, 0, 0, 1, 10800, "EEST", BIDI)
TESTVECTOR(1288483199, 2010, 10, 0, 31, 2, 59, 59, 1, 10800, "EEST", BIDI)
TESTVECTOR(1288483200, 2010, 10, 0, 31, 2, 0, 0, 0, 7200, "EET", HINT)
TESTVECTOR(1288486799, 2010, 10, 0, 31, 2, 59, 59, 0, 7200, "EET", HINT)
TESTVECTOR(1288486800, 2010, 10, 0, 31, 3, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1301183999, 2011, 3, 0, 27, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1301184000, 2011, 3, 0, 27, 3, 0, 0, 0, 10800, "+03", BIDI)
TEST_END

TEST_BEGIN(santiago, "America/Santiago")
TESTVECTOR(-2524504635, 1889, 12, 2, 31, 23, 59, 59, 0, -16966, "LMT", BIDI)
TESTVECTOR(-2524504634, 1890, 1, 3, 1, 0, 0, 0, 0, -16966, "SMT", BIDI)
TESTVECTOR(-1892661435, 1910, 1, 0, 9, 23, 59, 59, 0, -16966, "SMT", BIDI)
TESTVECTOR(-1892661434, 1910, 1, 0, 9, 23, 42, 46, 0, -18000, "-05", UNI)
TESTVECTOR(-1892660401, 1910, 1, 0, 9, 23, 59, 59, 0, -18000, "-05", UNI)
TESTVECTOR(-1892660400, 1910, 1, 1, 10, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-1688410801, 1916, 6, 5, 30, 23, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-1688410800, 1916, 7, 6, 1, 0, 17, 14, 0, -16966, "SMT", BIDI)
TESTVECTOR(-1619205435, 1918, 9, 1, 9, 23, 59, 59, 0, -16966, "SMT", BIDI)
TESTVECTOR(-1619205434, 1918, 9, 2, 10, 0, 42, 46, 0, -14400, "-04", BIDI)
TESTVECTOR(-1593806401, 1919, 6, 1, 30, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-1593806400, 1919, 6, 1, 30, 23, 17, 14, 0, -16966, "SMT", UNI)
TESTVECTOR(-1593803835, 1919, 6, 1, 30, 23, 59, 59, 0, -16966, "SMT", UNI)
TESTVECTOR(-1593803834, 1919, 7, 2, 1, 0, 0, 0, 0, -16966, "SMT", BIDI)
TESTVECTOR(-1335986235, 1927, 8, 3, 31, 23, 59, 59, 0, -16966, "SMT", BIDI)
TESTVECTOR(-1335986234, 1927, 9, 4, 1, 0, 42, 46, 1, -14400, "-04", BIDI)
TESTVECTOR(-1317585601, 1928, 3, 6, 31, 23, 59, 59, 1, -14400, "-04", BIDI)
TESTVECTOR(-1317585600, 1928, 3, 6, 31, 23, 0, 0, 0, -18000, "-05", HINT)
TESTVECTOR(-1317582001, 1928, 3, 6, 31, 23, 59, 59, 0, -18000, "-05", HINT)
TESTVECTOR(-1317582000, 1928, 4, 0, 1, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-1304362801, 1928, 8, 5, 31, 23, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-1304362800, 1928, 9, 6, 1, 1, 0, 0, 1, -14400, "-04", BIDI)
TESTVECTOR(-1286049601, 1929, 3, 0, 31, 23, 59, 59, 1, -14400, "-04", BIDI)
TESTVECTOR(-1286049600, 1929, 3, 0, 31, 23, 0, 0, 0, -18000, "-05", HINT)
TESTVECTOR(-1286046001, 1929, 3, 0, 31, 23, 59, 59, 0, -18000, "-05", HINT)
TESTVECTOR(-1286046000, 1929, 4, 1, 1, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-1272826801, 1929, 8, 6, 31, 23, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-1272826800, 1929, 9, 0, 1, 1, 0, 0, 1, -14400, "-04", BIDI)
TESTVECTOR(-1254513601, 1930, 3, 1, 31, 23, 59, 59, 1, -14400, "-04", BIDI)
TESTVECTOR(-1254513600, 1930, 3, 1, 31, 23, 0, 0, 0, -18000, "-05", HINT)
TESTVECTOR(-1254510001, 1930, 3, 1, 31, 23, 59, 59, 0, -18000, "-05", HINT)
TESTVECTOR(-1254510000, 1930, 4, 2, 1, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-1241290801, 1930, 8, 0, 31, 23, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-1241290800, 1930, 9, 1, 1, 1, 0, 0, 1, -14400, "-04", BIDI)
TESTVECTOR(-1222977601, 1931, 3, 2, 31, 23, 59, 59, 1, -14400, "-04", BIDI)
TESTVECTOR(-1222977600, 1931, 3, 2, 31, 23, 0, 0, 0, -18000, "-05", HINT)
TESTVECTOR(-1222974001, 1931, 3, 2, 31, 23, 59, 59, 0, -18000, "-05", HINT)
TESTVECTOR(-1222974000, 1931, 4, 3, 1, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-1209754801, 1931, 8, 1, 31, 23, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-1209754800, 1931, 9, 2, 1, 1, 0, 0, 1, -14400, "-04", BIDI)
TESTVECTOR(-1191355201, 1932, 3, 4, 31, 23, 59, 59, 1, -14400, "-04", BIDI)
TESTVECTOR(-1191355200, 1932, 3, 4, 31, 23, 0, 0, 0, -18000, "-05", HINT)
TESTVECTOR(-1191351601, 1932, 3, 4, 31, 23, 59, 59, 0, -18000, "-05", HINT)
TESTVECTOR(-1191351600, 1932, 4, 5, 1, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-1178132401, 1932, 8, 3, 31, 23, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-1178132400, 1932, 9, 4, 1, 1, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-870552001, 1942, 5, 0, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-870552000, 1942, 5, 0, 31, 23, 0, 0, 0, -18000, "-05", UNI)
TESTVECTOR(-870548401, 1942, 5, 0, 31, 23, 59, 59, 0, -18000, "-05", UNI)
TESTVECTOR(-870548400, 1942, 6, 1, 1, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-865278001, 1942, 7, 5, 31, 23, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-865278000, 1942, 8, 6, 1, 1, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-740520001, 1946, 7, 0, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-740520000, 1946, 7, 1, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-736376401, 1946, 8, 6, 31, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-736376400, 1946, 8, 6, 31, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-736372801, 1946, 8, 6, 31, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-736372800, 1946, 9, 0, 1, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-718056001, 1947, 3, 1, 31, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-718056000, 1947, 3, 1, 31, 23, 0, 0, 0, -18000, "-05", UNI)
TESTVECTOR(-718052401, 1947, 3, 1, 31, 23, 59, 59, 0, -18000, "-05", UNI)
TESTVECTOR(-718052400, 1947, 4, 2, 1, 0, 0, 0, 0, -18000, "-05", BIDI)
TESTVECTOR(-713649601, 1947, 5, 3, 21, 22, 59, 59, 0, -18000, "-05", BIDI)
TESTVECTOR(-713649600, 1947, 5, 4, 22, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-36619201, 1968, 11, 6, 2, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-36619200, 1968, 11, 0, 3, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(-23922001, 1969, 3, 6, 29, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(-23922000, 1969, 3, 6, 29, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(-23918401, 1969, 3, 6, 29, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(-23918400, 1969, 3, 0, 30, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(-3355201, 1969, 11, 6, 22, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(-3355200, 1969, 11, 0, 23, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(7527599, 1970, 3, 6, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(7527600, 1970, 3, 6, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(7531199, 1970, 3, 6, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(7531200, 1970, 3, 0, 29, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(24465599, 1970, 10, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(24465600, 1970, 10, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(37767599, 1971, 3, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(37767600, 1971, 3, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(37771199, 1971, 3, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(37771200, 1971, 3, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(55915199, 1971, 10, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(55915200, 1971, 10, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(69217199, 1972, 3, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(69217200, 1972, 3, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(69220799, 1972, 3, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(69220800, 1972, 3, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(87969599, 1972, 10, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(87969600, 1972, 10, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(100666799, 1973, 3, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(100666800, 1973, 3, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(100670399, 1973, 3, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(100670400, 1973, 3, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(118209599, 1973, 9, 6, 29, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(118209600, 1973, 9, 0, 30, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(132116399, 1974, 3, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(132116400, 1974, 3, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(132119999, 1974, 3, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(132120000, 1974, 3, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(150868799, 1974, 10, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(150868800, 1974, 10, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(163565999, 1975, 3, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(163566000, 1975, 3, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(163569599, 1975, 3, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(163569600, 1975, 3, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(182318399, 1975, 10, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(182318400, 1975, 10, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(195620399, 1976, 3, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(195620400, 1976, 3, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(195623999, 1976, 3, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(195624000, 1976, 3, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(213767999, 1976, 10, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(213768000, 1976, 10, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(227069999, 1977, 3, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(227070000, 1977, 3, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(227073599, 1977, 3, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(227073600, 1977, 3, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(245217599, 1977, 10, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(245217600, 1977, 10, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(258519599, 1978, 3, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(258519600, 1978, 3, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(258523199, 1978, 3, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(258523200, 1978, 3, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(277271999, 1978, 10, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(277272000, 1978, 10, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(289969199, 1979, 3, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(289969200, 1979, 3, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(289972799, 1979, 3, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(289972800, 1979, 3, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(308721599, 1979, 10, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(308721600, 1979, 10, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(321418799, 1980, 3, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(321418800, 1980, 3, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(321422399, 1980, 3, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(321422400, 1980, 3, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(340171199, 1980, 10, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(340171200, 1980, 10, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(353473199, 1981, 3, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(353473200, 1981, 3, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(353476799, 1981, 3, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(353476800, 1981, 3, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(371620799, 1981, 10, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(371620800, 1981, 10, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(384922799, 1982, 3, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(384922800, 1982, 3, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(384926399, 1982, 3, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(384926400, 1982, 3, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(403070399, 1982, 10, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(403070400, 1982, 10, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(416372399, 1983, 3, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(416372400, 1983, 3, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(416375999, 1983, 3, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(416376000, 1983, 3, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(434519999, 1983, 10, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(434520000, 1983, 10, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(447821999, 1984, 3, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(447822000, 1984, 3, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(447825599, 1984, 3, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(447825600, 1984, 3, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(466574399, 1984, 10, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(466574400, 1984, 10, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(479271599, 1985, 3, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(479271600, 1985, 3, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(479275199, 1985, 3, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(479275200, 1985, 3, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(498023999, 1985, 10, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(498024000, 1985, 10, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(510721199, 1986, 3, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(510721200, 1986, 3, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(510724799, 1986, 3, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(510724800, 1986, 3, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(529473599, 1986, 10, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(529473600, 1986, 10, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(545194799, 1987, 4, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(545194800, 1987, 4, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(545198399, 1987, 4, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(545198400, 1987, 4, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(560923199, 1987, 10, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(560923200, 1987, 10, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(574225199, 1988, 3, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(574225200, 1988, 3, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(574228799, 1988, 3, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(574228800, 1988, 3, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(592372799, 1988, 10, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(592372800, 1988, 10, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(605674799, 1989, 3, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(605674800, 1989, 3, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(605678399, 1989, 3, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(605678400, 1989, 3, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(624427199, 1989, 10, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(624427200, 1989, 10, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(637124399, 1990, 3, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(637124400, 1990, 3, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(637127999, 1990, 3, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(637128000, 1990, 3, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(653457599, 1990, 9, 6, 15, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(653457600, 1990, 9, 0, 16, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(668573999, 1991, 3, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(668574000, 1991, 3, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(668577599, 1991, 3, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(668577600, 1991, 3, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(687326399, 1991, 10, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(687326400, 1991, 10, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(700628399, 1992, 3, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(700628400, 1992, 3, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(700631999, 1992, 3, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(700632000, 1992, 3, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(718775999, 1992, 10, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(718776000, 1992, 10, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(732077999, 1993, 3, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(732078000, 1993, 3, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(732081599, 1993, 3, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(732081600, 1993, 3, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(750225599, 1993, 10, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(750225600, 1993, 10, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(763527599, 1994, 3, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(763527600, 1994, 3, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(763531199, 1994, 3, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(763531200, 1994, 3, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(781675199, 1994, 10, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(781675200, 1994, 10, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(794977199, 1995, 3, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(794977200, 1995, 3, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(794980799, 1995, 3, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(794980800, 1995, 3, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(813729599, 1995, 10, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(813729600, 1995, 10, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(826426799, 1996, 3, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(826426800, 1996, 3, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(826430399, 1996, 3, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(826430400, 1996, 3, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(845179199, 1996, 10, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(845179200, 1996, 10, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(859690799, 1997, 3, 6, 29, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(859690800, 1997, 3, 6, 29, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(859694399, 1997, 3, 6, 29, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(859694400, 1997, 3, 0, 30, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(876628799, 1997, 10, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(876628800, 1997, 10, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(889930799, 1998, 3, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(889930800, 1998, 3, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(889934399, 1998, 3, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(889934400, 1998, 3, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(906868799, 1998, 9, 6, 26, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(906868800, 1998, 9, 0, 27, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(923194799, 1999, 4, 6, 3, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(923194800, 1999, 4, 6, 3, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(923198399, 1999, 4, 6, 3, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(923198400, 1999, 4, 0, 4, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(939527999, 1999, 10, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(939528000, 1999, 10, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(952829999, 2000, 3, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(952830000, 2000, 3, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(952833599, 2000, 3, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(952833600, 2000, 3, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(971582399, 2000, 10, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(971582400, 2000, 10, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(984279599, 2001, 3, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(984279600, 2001, 3, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(984283199, 2001, 3, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(984283200, 2001, 3, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1003031999, 2001, 10, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1003032000, 2001, 10, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1015729199, 2002, 3, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1015729200, 2002, 3, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1015732799, 2002, 3, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1015732800, 2002, 3, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1034481599, 2002, 10, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1034481600, 2002, 10, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1047178799, 2003, 3, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1047178800, 2003, 3, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1047182399, 2003, 3, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1047182400, 2003, 3, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1065931199, 2003, 10, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1065931200, 2003, 10, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1079233199, 2004, 3, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1079233200, 2004, 3, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1079236799, 2004, 3, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1079236800, 2004, 3, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1097380799, 2004, 10, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1097380800, 2004, 10, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1110682799, 2005, 3, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1110682800, 2005, 3, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1110686399, 2005, 3, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1110686400, 2005, 3, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1128830399, 2005, 10, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1128830400, 2005, 10, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1142132399, 2006, 3, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1142132400, 2006, 3, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1142135999, 2006, 3, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1142136000, 2006, 3, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1160884799, 2006, 10, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1160884800, 2006, 10, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1173581999, 2007, 3, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1173582000, 2007, 3, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1173585599, 2007, 3, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1173585600, 2007, 3, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1192334399, 2007, 10, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1192334400, 2007, 10, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1206845999, 2008, 3, 6, 29, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1206846000, 2008, 3, 6, 29, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1206849599, 2008, 3, 6, 29, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1206849600, 2008, 3, 0, 30, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1223783999, 2008, 10, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1223784000, 2008, 10, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1237085999, 2009, 3, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1237086000, 2009, 3, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1237089599, 2009, 3, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1237089600, 2009, 3, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1255233599, 2009, 10, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1255233600, 2009, 10, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1270349999, 2010, 4, 6, 3, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1270350000, 2010, 4, 6, 3, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1270353599, 2010, 4, 6, 3, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1270353600, 2010, 4, 0, 4, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1286683199, 2010, 10, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1286683200, 2010, 10, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1304823599, 2011, 5, 6, 7, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1304823600, 2011, 5, 6, 7, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1304827199, 2011, 5, 6, 7, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1304827200, 2011, 5, 0, 8, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1313899199, 2011, 8, 6, 20, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1313899200, 2011, 8, 0, 21, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1335668399, 2012, 4, 6, 28, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1335668400, 2012, 4, 6, 28, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1335671999, 2012, 4, 6, 28, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1335672000, 2012, 4, 0, 29, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1346558399, 2012, 9, 6, 1, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1346558400, 2012, 9, 0, 2, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1367117999, 2013, 4, 6, 27, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1367118000, 2013, 4, 6, 27, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1367121599, 2013, 4, 6, 27, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1367121600, 2013, 4, 0, 28, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1378612799, 2013, 9, 6, 7, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1378612800, 2013, 9, 0, 8, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1398567599, 2014, 4, 6, 26, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1398567600, 2014, 4, 6, 26, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1398571199, 2014, 4, 6, 26, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1398571200, 2014, 4, 0, 27, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1410062399, 2014, 9, 6, 6, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1410062400, 2014, 9, 0, 7, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1463281199, 2016, 5, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1463281200, 2016, 5, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1463284799, 2016, 5, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1463284800, 2016, 5, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1471147199, 2016, 8, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1471147200, 2016, 8, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1494730799, 2017, 5, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1494730800, 2017, 5, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1494734399, 2017, 5, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1494734400, 2017, 5, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1502596799, 2017, 8, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1502596800, 2017, 8, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1526180399, 2018, 5, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1526180400, 2018, 5, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1526183999, 2018, 5, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1526184000, 2018, 5, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1534046399, 2018, 8, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1534046400, 2018, 8, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1557629999, 2019, 5, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1557630000, 2019, 5, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1557633599, 2019, 5, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1557633600, 2019, 5, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1565495999, 2019, 8, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1565496000, 2019, 8, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1589079599, 2020, 5, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1589079600, 2020, 5, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1589083199, 2020, 5, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1589083200, 2020, 5, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1596945599, 2020, 8, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1596945600, 2020, 8, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1620529199, 2021, 5, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1620529200, 2021, 5, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1620532799, 2021, 5, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1620532800, 2021, 5, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1628999999, 2021, 8, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1629000000, 2021, 8, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1652583599, 2022, 5, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1652583600, 2022, 5, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1652587199, 2022, 5, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1652587200, 2022, 5, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1660449599, 2022, 8, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1660449600, 2022, 8, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1684033199, 2023, 5, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1684033200, 2023, 5, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1684036799, 2023, 5, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1684036800, 2023, 5, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1691899199, 2023, 8, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1691899200, 2023, 8, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1715482799, 2024, 5, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1715482800, 2024, 5, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1715486399, 2024, 5, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1715486400, 2024, 5, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1723348799, 2024, 8, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1723348800, 2024, 8, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1746932399, 2025, 5, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1746932400, 2025, 5, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1746935999, 2025, 5, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1746936000, 2025, 5, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1754798399, 2025, 8, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1754798400, 2025, 8, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1778381999, 2026, 5, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1778382000, 2026, 5, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1778385599, 2026, 5, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1778385600, 2026, 5, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1786247999, 2026, 8, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1786248000, 2026, 8, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1809831599, 2027, 5, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1809831600, 2027, 5, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1809835199, 2027, 5, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1809835200, 2027, 5, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1818302399, 2027, 8, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1818302400, 2027, 8, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1841885999, 2028, 5, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1841886000, 2028, 5, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1841889599, 2028, 5, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1841889600, 2028, 5, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1849751999, 2028, 8, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1849752000, 2028, 8, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1873335599, 2029, 5, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1873335600, 2029, 5, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1873339199, 2029, 5, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1873339200, 2029, 5, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1881201599, 2029, 8, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1881201600, 2029, 8, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1904785199, 2030, 5, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1904785200, 2030, 5, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1904788799, 2030, 5, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1904788800, 2030, 5, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1912651199, 2030, 8, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1912651200, 2030, 8, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1936234799, 2031, 5, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1936234800, 2031, 5, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1936238399, 2031, 5, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1936238400, 2031, 5, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1944100799, 2031, 8, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1944100800, 2031, 8, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1967684399, 2032, 5, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1967684400, 2032, 5, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1967687999, 2032, 5, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1967688000, 2032, 5, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(1976155199, 2032, 8, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(1976155200, 2032, 8, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(1999738799, 2033, 5, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(1999738800, 2033, 5, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(1999742399, 2033, 5, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(1999742400, 2033, 5, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2007604799, 2033, 8, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2007604800, 2033, 8, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2031188399, 2034, 5, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2031188400, 2034, 5, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2031191999, 2034, 5, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2031192000, 2034, 5, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2039054399, 2034, 8, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2039054400, 2034, 8, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2062637999, 2035, 5, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2062638000, 2035, 5, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2062641599, 2035, 5, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2062641600, 2035, 5, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2070503999, 2035, 8, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2070504000, 2035, 8, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2094087599, 2036, 5, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2094087600, 2036, 5, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2094091199, 2036, 5, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2094091200, 2036, 5, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2101953599, 2036, 8, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2101953600, 2036, 8, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2125537199, 2037, 5, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2125537200, 2037, 5, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2125540799, 2037, 5, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2125540800, 2037, 5, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2133403199, 2037, 8, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2133403200, 2037, 8, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2156986799, 2038, 5, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2156986800, 2038, 5, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2156990399, 2038, 5, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2156990400, 2038, 5, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2165457599, 2038, 8, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2165457600, 2038, 8, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2189041199, 2039, 5, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2189041200, 2039, 5, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2189044799, 2039, 5, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2189044800, 2039, 5, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2196907199, 2039, 8, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2196907200, 2039, 8, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2220490799, 2040, 5, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2220490800, 2040, 5, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2220494399, 2040, 5, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2220494400, 2040, 5, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2228356799, 2040, 8, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2228356800, 2040, 8, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2251940399, 2041, 5, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2251940400, 2041, 5, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2251943999, 2041, 5, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2251944000, 2041, 5, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2259806399, 2041, 8, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2259806400, 2041, 8, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2283389999, 2042, 5, 6, 10, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2283390000, 2042, 5, 6, 10, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2283393599, 2042, 5, 6, 10, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2283393600, 2042, 5, 0, 11, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2291255999, 2042, 8, 6, 9, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2291256000, 2042, 8, 0, 10, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2314839599, 2043, 5, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2314839600, 2043, 5, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2314843199, 2043, 5, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2314843200, 2043, 5, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2322705599, 2043, 8, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2322705600, 2043, 8, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2346893999, 2044, 5, 6, 14, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2346894000, 2044, 5, 6, 14, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2346897599, 2044, 5, 6, 14, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2346897600, 2044, 5, 0, 15, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2354759999, 2044, 8, 6, 13, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2354760000, 2044, 8, 0, 14, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2378343599, 2045, 5, 6, 13, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2378343600, 2045, 5, 6, 13, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2378347199, 2045, 5, 6, 13, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2378347200, 2045, 5, 0, 14, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2386209599, 2045, 8, 6, 12, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2386209600, 2045, 8, 0, 13, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2409793199, 2046, 5, 6, 12, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2409793200, 2046, 5, 6, 12, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2409796799, 2046, 5, 6, 12, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2409796800, 2046, 5, 0, 13, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2417659199, 2046, 8, 6, 11, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2417659200, 2046, 8, 0, 12, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2441242799, 2047, 5, 6, 11, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2441242800, 2047, 5, 6, 11, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2441246399, 2047, 5, 6, 11, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2441246400, 2047, 5, 0, 12, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2449108799, 2047, 8, 6, 10, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2449108800, 2047, 8, 0, 11, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2472692399, 2048, 5, 6, 9, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2472692400, 2048, 5, 6, 9, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2472695999, 2048, 5, 6, 9, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2472696000, 2048, 5, 0, 10, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2480558399, 2048, 8, 6, 8, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2480558400, 2048, 8, 0, 9, 1, 0, 0, 1, -10800, "-03", BIDI)
TESTVECTOR(2504141999, 2049, 5, 6, 8, 23, 59, 59, 1, -10800, "-03", BIDI)
TESTVECTOR(2504142000, 2049, 5, 6, 8, 23, 0, 0, 0, -14400, "-04", HINT)
TESTVECTOR(2504145599, 2049, 5, 6, 8, 23, 59, 59, 0, -14400, "-04", HINT)
TESTVECTOR(2504145600, 2049, 5, 0, 9, 0, 0, 0, 0, -14400, "-04", BIDI)
TESTVECTOR(2512612799, 2049, 8, 6, 14, 23, 59, 59, 0, -14400, "-04", BIDI)
TESTVECTOR(2512612800, 2049, 8, 0, 15, 1, 0, 0, 1, -10800, "-03", BIDI)
TEST_END

TEST_BEGIN(tripoli, "Africa/Tripoli")
TESTVECTOR(-1577926365, 1919, 12, 3, 31, 23, 59, 59, 0, 3164, "LMT", BIDI)
TESTVECTOR(-1577926364, 1920, 1, 4, 1, 0, 7, 16, 0, 3600, "CET", BIDI)
TESTVECTOR(-574902001, 1951, 10, 0, 14, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-574902000, 1951, 10, 0, 14, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-568087201, 1951, 12, 1, 31, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-568087200, 1951, 12, 1, 31, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-568083601, 1951, 12, 1, 31, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-568083600, 1952, 1, 2, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-512175601, 1953, 10, 5, 9, 1, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-512175600, 1953, 10, 5, 9, 3, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-504928801, 1953, 12, 4, 31, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-504928800, 1953, 12, 4, 31, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-504925201, 1953, 12, 4, 31, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-504925200, 1954, 1, 5, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-449888401, 1955, 9, 4, 29, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-449888400, 1955, 9, 5, 30, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(-441856801, 1955, 12, 6, 31, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(-441856800, 1955, 12, 6, 31, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(-441853201, 1955, 12, 6, 31, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(-441853200, 1956, 1, 0, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(-347158801, 1958, 12, 3, 31, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(-347158800, 1959, 1, 4, 1, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(378683999, 1981, 12, 4, 31, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(378684000, 1981, 12, 4, 31, 23, 0, 0, 0, 3600, "CET", UNI)
TESTVECTOR(378687599, 1981, 12, 4, 31, 23, 59, 59, 0, 3600, "CET", UNI)
TESTVECTOR(378687600, 1982, 1, 5, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(386463599, 1982, 3, 3, 31, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(386463600, 1982, 4, 4, 1, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(402271199, 1982, 9, 4, 30, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(402271200, 1982, 9, 4, 30, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(402274799, 1982, 9, 4, 30, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(402274800, 1982, 10, 5, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(417999599, 1983, 3, 4, 31, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(417999600, 1983, 4, 5, 1, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(433807199, 1983, 9, 5, 30, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(433807200, 1983, 9, 5, 30, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(433810799, 1983, 9, 5, 30, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(433810800, 1983, 10, 6, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(449621999, 1984, 3, 6, 31, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(449622000, 1984, 4, 0, 1, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(465429599, 1984, 9, 0, 30, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(465429600, 1984, 9, 0, 30, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(465433199, 1984, 9, 0, 30, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(465433200, 1984, 10, 1, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(481589999, 1985, 4, 5, 5, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(481590000, 1985, 4, 6, 6, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(496965599, 1985, 9, 1, 30, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(496965600, 1985, 9, 1, 30, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(496969199, 1985, 9, 1, 30, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(496969200, 1985, 10, 2, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(512953199, 1986, 4, 4, 3, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(512953200, 1986, 4, 5, 4, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(528674399, 1986, 10, 4, 2, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(528674400, 1986, 10, 4, 2, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(528677999, 1986, 10, 4, 2, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(528678000, 1986, 10, 5, 3, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(544229999, 1987, 3, 2, 31, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(544230000, 1987, 4, 3, 1, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(560037599, 1987, 9, 3, 30, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(560037600, 1987, 9, 3, 30, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(560041199, 1987, 9, 3, 30, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(560041200, 1987, 10, 4, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(575852399, 1988, 3, 4, 31, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(575852400, 1988, 4, 5, 1, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(591659999, 1988, 9, 5, 30, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(591660000, 1988, 9, 5, 30, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(591663599, 1988, 9, 5, 30, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(591663600, 1988, 10, 6, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(607388399, 1989, 3, 5, 31, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(607388400, 1989, 4, 6, 1, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(623195999, 1989, 9, 6, 30, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(623196000, 1989, 9, 6, 30, 23, 0, 0, 0, 3600, "CET", HINT)
TESTVECTOR(623199599, 1989, 9, 6, 30, 23, 59, 59, 0, 3600, "CET", HINT)
TESTVECTOR(623199600, 1989, 10, 0, 1, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(641775599, 1990, 5, 4, 3, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(641775600, 1990, 5, 5, 4, 1, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(844034399, 1996, 9, 0, 29, 23, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(844034400, 1996, 9, 0, 29, 23, 0, 0, 0, 3600, "CET", UNI)
TESTVECTOR(844037999, 1996, 9, 0, 29, 23, 59, 59, 0, 3600, "CET", UNI)
TESTVECTOR(844038000, 1996, 9, 1, 30, 0, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(860108399, 1997, 4, 4, 3, 23, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(860108400, 1997, 4, 5, 4, 1, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(875915999, 1997, 10, 5, 3, 23, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(875916000, 1997, 10, 6, 4, 0, 0, 0, 0, 7200, "EET", BIDI)
TESTVECTOR(1352505599, 2012, 11, 6, 10, 1, 59, 59, 0, 7200, "EET", BIDI)
TESTVECTOR(1352505600, 2012, 11, 6, 10, 1, 0, 0, 0, 3600, "CET", UNI)
TESTVECTOR(1352509199, 2012, 11, 6, 10, 1, 59, 59, 0, 3600, "CET", UNI)
TESTVECTOR(1352509200, 2012, 11, 6, 10, 2, 0, 0, 0, 3600, "CET", BIDI)
TESTVECTOR(1364515199, 2013, 3, 5, 29, 0, 59, 59, 0, 3600, "CET", BIDI)
TESTVECTOR(1364515200, 2013, 3, 5, 29, 2, 0, 0, 1, 7200, "CEST", BIDI)
TESTVECTOR(1382659199, 2013, 10, 5, 25, 1, 59, 59, 1, 7200, "CEST", BIDI)
TESTVECTOR(1382659200, 2013, 10, 5, 25, 2, 0, 0, 0, 7200, "EET", BIDI)
TEST_END
